package ch.dlcm.service;

import java.util.ArrayList;
import java.util.List;

import javax.annotation.PostConstruct;

import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.jms.annotation.JmsListener;
import org.springframework.stereotype.Component;
import org.springframework.transaction.annotation.Transactional;

import ch.dlcm.assembling.AssemblingService;
import ch.dlcm.message.OrderMessage;
import ch.dlcm.model.Order;
import ch.dlcm.model.Status.StatusValue;
import ch.dlcm.model.oais.DisseminationInfoPackage;
import ch.dlcm.repository.OrderRepository;

@Component
public class OrderMgmt extends DLCMService {
  private static final Logger log = Logger.getLogger(OrderMgmt.class);

  @Autowired
  OrderRepository orderRepo;

  @Autowired
  AssemblingService assemblingSvc;

  @PostConstruct
  @Override
  protected void loadConfig() {
  }

  @Transactional
  @JmsListener(destination = "${dlcm.queue.order}")
  public void processOrder(OrderMessage order) {
    if (order.getDlcmId() != null) {
      log.info("Reading message " + order.toString());
      processOrder(order.getDlcmId());
    }else {
      checkOrderProgress();
    }

  }
  
  private void checkOrderProgress() {
    List<StatusValue> statusList=new ArrayList<>();
    statusList.add(StatusValue.InDisseminationPreparation);
    for(Order o:orderRepo.findByStatusValue(statusList)) {
      if (o.getStatus().getValue() == StatusValue.InDisseminationPreparation) {
        o = orderRepo.findOneWithDip(o.getDlcmId());
        if (o.isCompleted()) {
          o.setOrderStatus(StatusValue.Ready);
        }else if (o.isDipInError()){
          o.setErrorStatus(messageService.get("access.order.error.dip"));
        }
        if (orderRepo.save(o) == null) {
          log.error("Order cannot be saved: dlcmId=" + o.getDlcmId() + " status=" + o.getStatus().getValue().toString());
        }
        log.info("Order processed: dlcmId=" + o.getDlcmId() + " status=" + o.getStatus().getValue().toString());
      }
    }
  }

  private void processOrder(String orderId) {
    Order o = this.getOne(orderRepo, orderId);
    if (o == null) {
      log.error("Cannot find order (" + orderId + ")");
      return;
    }
    while (o.getStatus().getValue() != StatusValue.InError 
        && o.getStatus().getValue() != StatusValue.Ready
        && o.getStatus().getValue() != StatusValue.InDisseminationPreparation) {
      try {
        log.info("Order: " + o.getName() + " (" + o.getDlcmId() + "): status=" + o.getStatus().getValue().toString());
        switch (o.getStatus().getValue()) {
          case Submitted:
            o = orderRepo.findOneWithAip(orderId);
            if (o.getAipPackages().isEmpty()) {
              o.setErrorStatus(messageService.get("access.order.datafile.notready"));
            } else if (assemblingSvc.preparePackage(o.getDlcmId(), o.getAipPackages()) != null) {
              o.setOrderStatus(StatusValue.InPreparation);
            } else {
              o.setErrorStatus(messageService.get("access.order.error.preparation"));
            }
            break;
          case InPreparation:
            DisseminationInfoPackage dip;
            if (assemblingSvc.getStatus(o.getDlcmId()) == StatusValue.InProgress) {
              if (assemblingSvc.submitPackage(o.getDlcmId(), o.getName()) == null) {
                o.setErrorStatus(messageService.get("access.order.error.submission"));
              }
            } else {
              if ((dip = assemblingSvc.createPackage(o.getDlcmId(), o.getName(), o.getAipPackages())) != null) {
                o = orderRepo.findOneWithDip(orderId);
                o.getDipPackages().add(dip);
                o.setOrderStatus(StatusValue.InDisseminationPreparation);
              } else {
                o.setErrorStatus(messageService.get("access.order.error.sending"));
              }
            }
            break;
          default:
            break;
        }
      } catch (Exception e) {
        log.error("Error in processing order (" + orderId + ")", e);
        o.setErrorStatus(e.getMessage());
      }
      try {
        if (orderRepo.save(o) == null) {
          log.error("Order cannot be saved: dlcmId=" + o.getDlcmId() + " status=" + o.getStatus().getValue().toString());
        }  
      }
      catch (Exception e) {
        log.error("Error in saving order", e);
      }
    }
    log.info("Order processed: dlcmId=" + o.getDlcmId() + " status=" + o.getStatus().getValue().toString());
  }
  
}
