/*
 * Copyright 2014-2015 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ch.dlcm;

import static ch.dlcm.test.executor.DLCMExecutorTestDoc.doDocumentationFieldsCycle;
import static ch.dlcm.test.executor.DLCMExecutorTestDoc.doDocumentationLinksCycle;
import static ch.dlcm.test.executor.DLCMExecutorTestDoc.doPerformRequestCycle;
import static ch.dlcm.test.executor.DLCMExecutorTestDoc.doTestsExpectedCycle;
import static ch.dlcm.test.executor.DLCMTestDocHelper.APPLICATION;
import static ch.dlcm.test.executor.DLCMTestDocHelper.HAL_JSON_MEDIA_TYPE;
import static ch.dlcm.test.executor.DLCMTestDocHelper.HTML_MEDIA_TYPE;
import static ch.dlcm.test.executor.DLCMTestDocHelper.JSON_MEDIA_TYPE;
import static org.springframework.restdocs.mockmvc.MockMvcRestDocumentation.documentationConfiguration;

import org.junit.Before;
import org.junit.Rule;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.http.HttpStatus;
import org.springframework.restdocs.JUnitRestDocumentation;
import org.springframework.test.context.ActiveProfiles;
import org.springframework.test.context.junit4.SpringJUnit4ClassRunner;
import org.springframework.test.context.web.WebAppConfiguration;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.test.web.servlet.MvcResult;
import org.springframework.test.web.servlet.ResultActions;
import org.springframework.test.web.servlet.setup.MockMvcBuilders;
import org.springframework.web.context.WebApplicationContext;

import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;

import ch.dlcm.rest.RESTAction;
import ch.dlcm.rest.RESTName;
import ch.dlcm.test.doc.matcher.accession.AccessDocMatcher;
import ch.dlcm.test.doc.matcher.accession.AccessionDocMatcher;
import ch.dlcm.test.executor.DLCMTestDocHelper;

@ActiveProfiles({"test","arc-file","dat-elasticsearch"})
@RunWith(SpringJUnit4ClassRunner.class)
@SpringBootTest
@WebAppConfiguration
public class AccessionRestTestsAndDocumentation {

	private MockMvc mockMvc;
	
	@Rule
	public final JUnitRestDocumentation restDocumentation = new JUnitRestDocumentation("target/generated-snippets");

	@Autowired
	private WebApplicationContext context;

	@Before
	public void setUp() {
		this.mockMvc = MockMvcBuilders.webAppContextSetup(this.context)
				.apply(documentationConfiguration(this.restDocumentation)).build();
	}

	@Test
	public void initAccessionTestAndDocs() throws Exception {

		// Execute acession "root"
		MvcResult result = doPerformRequestCycle(mockMvc, DLCMTestDocHelper.APPLICATION_ROOT, HAL_JSON_MEDIA_TYPE, HttpStatus.OK).andReturn();
		JsonNode links = new ObjectMapper().readTree(result.getResponse().getContentAsString()).get("_links");

		// Execute acession "_links"
		DLCMTestDocHelper.getAccessionRootLinksActions().forEach(action -> {
			try {
				if (links.has(action)) {
					String url = links.get(action).get("href").asText();
					switch (action) {
						case RESTAction.SELF:
							generateAccessionDoc(url);
							break;
						case RESTName.ACCESS:
							generateAccessDoc(url);
							generateAccesssDipDoc(url + "/" + RESTName.DIP);
							generateAccesssDipStatusDoc(url + "/" + RESTName.DIP + "/" + RESTName.STATUS);
							generateAccesssAipDoc(url + "/" + RESTName.AIP);
							generateAccesssOrdersDoc(url + "/" + RESTName.ORDER);
							break;
						case RESTAction.DASHBOARD:
							generateDashboardDoc(url);
							break;
					}
				}
			} catch (Exception e) {
				throw new RuntimeException(e);
			}
		});
	}

	private MvcResult generateAccesssAipDoc(String url) throws Exception {
	  ResultActions request = doPerformRequestCycle(mockMvc, url, HAL_JSON_MEDIA_TYPE, HttpStatus.OK);
    doTestsExpectedCycle(request, AccessDocMatcher.getAccessAipExpectedMatchers());
    doDocumentationLinksCycle(request, RESTName.ACCESS, AccessDocMatcher.getAccessAipMapLinksDescription());
    return request.andReturn();
  }

  private MvcResult generateAccesssDipStatusDoc(String url) throws Exception {
		ResultActions request = doPerformRequestCycle(mockMvc, url, JSON_MEDIA_TYPE, HttpStatus.OK);
		doTestsExpectedCycle(request, AccessDocMatcher.getAccessDipStatusExpectedMatchers());
		doDocumentationFieldsCycle(request, RESTName.ACCESS + "/" + RESTName.DIP + "/" + RESTName.STATUS, AccessDocMatcher.getAccessDipStatusMapFieldDescription());
		return request.andReturn();
	}

	private MvcResult generateAccesssOrdersDoc(String url) throws Exception {
		ResultActions request = doPerformRequestCycle(mockMvc, url, HAL_JSON_MEDIA_TYPE, HttpStatus.OK);
		doTestsExpectedCycle(request, AccessDocMatcher.getAccessOrdersExpectedMatchers());
		doDocumentationLinksCycle(request, RESTName.ACCESS + "/" + RESTName.ORDER, AccessDocMatcher.getAccessOrdersMapLinksDescription());
		return request.andReturn();
	}

	private MvcResult generateAccesssDipDoc(String url) throws Exception {
		ResultActions request = doPerformRequestCycle(mockMvc, url, HAL_JSON_MEDIA_TYPE, HttpStatus.OK);
		doTestsExpectedCycle(request, AccessDocMatcher.getAccessDipExpectedMatchers());
		doDocumentationLinksCycle(request, RESTName.ACCESS, AccessDocMatcher.getAccessDipMapLinksDescription());
		return request.andReturn();
	}

	private MvcResult generateAccessDoc(String url) throws Exception {
		ResultActions request = doPerformRequestCycle(mockMvc, url, HAL_JSON_MEDIA_TYPE, HttpStatus.OK);
		doTestsExpectedCycle(request, AccessDocMatcher.getAccessExpectedMatchers());
		doDocumentationLinksCycle(request, RESTName.ACCESS, AccessDocMatcher.getAccessMapLinksDescription());
		return request.andReturn();
	}

	private void generateDashboardDoc(String url) throws Exception{
		doPerformRequestCycle(mockMvc, url, HTML_MEDIA_TYPE, HttpStatus.OK);
	}

	private MvcResult generateAccessionDoc(String url) throws Exception{
		ResultActions request = doPerformRequestCycle(mockMvc, url, HAL_JSON_MEDIA_TYPE, HttpStatus.OK);
		doTestsExpectedCycle(request, AccessionDocMatcher.getAccessionRootExpectedMatchers());
		doDocumentationLinksCycle(request, APPLICATION, AccessionDocMatcher.getAccessionRootMapLinksDescription());
		return request.andReturn();

	}


}
