package ch.dlcm.service;

import javax.annotation.PostConstruct;

import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.jms.annotation.JmsListener;
import org.springframework.stereotype.Component;
import org.springframework.transaction.annotation.Transactional;

import ch.dlcm.message.AipMessage;
import ch.dlcm.model.Status.StatusValue;
import ch.dlcm.model.oais.ArchivalInfoPackage;
import ch.dlcm.repository.ArchivalInfoPackageRepository;
import ch.dlcm.storage.StorageService;

@Component
public class AipMgmt extends DLCMService {
  private static final Logger log = Logger.getLogger(AipMgmt.class);

  @Autowired
  ArchivalInfoPackageRepository repo;

  @Autowired
  StorageService storageSvc;

  @PostConstruct
  @Override
  protected void loadConfig() {
  }

  @Transactional
  @JmsListener(destination = "${dlcm.queue.aip}")
  public void processAip(AipMessage aip) {
    log.info("Reading message " + aip.toString());
    ArchivalInfoPackage a = this.getOne(repo, aip.getDlcmId());
    if (a == null) {
      log.error("Cannot find AIP (" + aip.getDlcmId() + ")");
      return;
    }
    while (a.getInfo().getStatus().getValue() != StatusValue.InError
            && a.getInfo().getStatus().getValue() != StatusValue.Completed) {
      try {
        switch (a.getInfo().getStatus().getValue()) {
          case InPreparation:
            if (a.isAllDataFileReady()) {
              a.setPackageStatus(StatusValue.Ready);
            }
            break;
          case Ready:
            if (storageSvc.checkPackage(a.getDlcmId())) {
              a.setPackageStatus(StatusValue.Checked);
            } else {
              inError(a, messageService.get("archival.aip.error.check"));
            }
            break;
          case Checked:
            if (storageSvc.getStatus(a.getDlcmId()) == StatusValue.Ready) {
              if (storageSvc.submitPackage(a.getDlcmId(), a.getInfo().getName()) != null) {
                a.setPackageStatus(StatusValue.InProgress);
              } else {
                inError(a, messageService.get("archival.aip.error.submission"));
              }
            } else if (storageSvc.getStatus(a.getDlcmId()) == StatusValue.Stored) {
              a.setPackageStatus(StatusValue.Stored);
            }
            break;
          case InProgress:
            String archiveId;
            if (storageSvc.getStatus(a.getDlcmId()) == StatusValue.Stored) {
              if ((archiveId = storageSvc.sendPackage(a.getDlcmId(),
                      a.getInfo().getName())) != null) {
                a.setArchiveId(archiveId);
                a.setPackageStatus(StatusValue.Stored);
              } else {
                inError(a, messageService.get("archival.aip.error.storage"));
              }
            }
            break;
          case Reindexing:
            if (!storageSvc.deleteIndex(a.getDlcmId())) {
              inError(a, messageService.get("archival.aip.error.reindexing"));
            }
            break;
          case Stored:
            if (storageSvc.checkIndexing(a.getDlcmId())) {
              a.setPackageStatus(StatusValue.Completed);
            } else {
              if (storageSvc.indexPackage(a.getDlcmId())) {
                a.setPackageStatus(StatusValue.Indexing);
              } else {
                inError(a, messageService.get("archival.aip.error.indexing"));
              }
            }
            break;
          case Indexing:
            if (checkIndexing(a.getDlcmId())) {
              a.setPackageStatus(StatusValue.Completed);
            } else {
              inError(a, messageService.get("archival.aip.error.noindex"));
            }
            break;
          default:
            break;
        }
      } catch (Exception e) {
        log.error("Error in processing AIP (" + aip.getDlcmId() + ")", e);
        inError(a, e.getMessage());
      }
      if (repo.save(a) == null) {
        log.error("AIP cannot be saved: dlcmId=" + a.getDlcmId() + " status="
                + a.getInfo().getStatus().getValue().toString());
      }
    }
    log.info("AIP processed: dlcmId=" + a.getDlcmId() + " status="
            + a.getInfo().getStatus().getValue().toString());

  }

  private boolean checkIndexing(String dlcmId) {
    int tries = 0;
    while (tries < dlcmConfig.getMaxTries()) {
      
      try {
        Thread.sleep(dlcmConfig.getWaitTime());
      } catch (InterruptedException e) {
        // Some useful details
        // http://www.yegor256.com/2015/10/20/interrupted-exception.html
        log.error(e);
        Thread.currentThread().interrupt();
        return false;
      }
      if (storageSvc.checkIndexing(dlcmId)) {
        return true;
      }
      tries++;
    }
    return false;
  }

  private void inError(ArchivalInfoPackage aip, String msg) {
    aip.setErrorStatus(msg);
    log.error(msg + " for AIP " + aip.getDlcmId());
  }
}
