package ch.dlcm.service;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.lang.reflect.Array;
import java.net.URI;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import ch.dlcm.tool.RestClientTool;
import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.core.ParameterizedTypeReference;
import org.springframework.core.env.Environment;
import org.springframework.core.io.Resource;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpHeaders;
import org.springframework.http.HttpMethod;
import org.springframework.http.MediaType;
import org.springframework.http.ResponseEntity;
import org.springframework.util.LinkedMultiValueMap;
import org.springframework.util.MultiValueMap;
import org.springframework.web.client.HttpClientErrorException;
import org.springframework.web.client.ResponseExtractor;
import org.springframework.web.client.RestTemplate;

import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.databind.JavaType;
import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.datatype.jsr310.JavaTimeModule;

import ch.dlcm.rest.RESTName;
import ch.dlcm.rest.RESTResource;

public abstract class ResourceServiceREST<T extends RESTResource<T>> implements ResourceService<T> {
  private static final Logger logger        = Logger.getLogger(ResourceServiceREST.class);
  private static final int    BUFFER_SIZE   = 16384;                                      // 2 ^ 14
  protected static final String DLCM_ID_LABEL = "dlcmId";

  @Autowired
  private Environment         env;

  private String              module       = null;

  protected final String      resourcePath;
  private final Class<T>      resourceClass;
  private final ObjectMapper  mapper       = new ObjectMapper();

  public ResourceServiceREST(String resourcePath, Class<T> resourceClass, String module) {
    this.resourcePath = resourcePath;
    this.resourceClass = resourceClass;
    this.module = module;
    mapper.configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);
    mapper.registerModule(new JavaTimeModule());
  }

  protected String getBaseUrl(String module) {
    switch (module) {
      case RESTName.ADMIN:
        return env.getProperty("dlcm.module.admin");
      case RESTName.PREINGEST:
        return env.getProperty("dlcm.module.preingest");
      case RESTName.INGEST:
        return env.getProperty("dlcm.module.ingest");
      case RESTName.ARCHIVALSTORAGE:
        return env.getProperty("dlcm.module.archival-storage");
      case RESTName.ACCESS:
        return env.getProperty("dlcm.module.access");
    }

    return env.getProperty("dlcm.module.admin");
  }

  @Override
  public T findOne(String dlcmId) {
    T result = null;
    final String url = this.getBaseUrl() + resourcePath + "/" + "{dlcmId}";
    RestTemplate restTemplate = RestClientTool.getClient(getBaseUrl());
    Map<String, String> params = new HashMap<>();
    params.put(DLCM_ID_LABEL, dlcmId);
    try {
      String jsonString = restTemplate.getForObject(url, String.class, params);
      result = mapper.readValue(jsonString, resourceClass);
    } catch (IOException e) {
      logger.error(e);
    } catch (HttpClientErrorException e) {
      if (e.getRawStatusCode() == 404) {
        // No item found return null
        return null;
      } else {
        throw e;
      }
    }
    return result;
  }

  @Override
  public T create(T resource) {
    final String url = this.getBaseUrl() + resourcePath;
    RestTemplate restTemplate = RestClientTool.getClient(getBaseUrl());
    return restTemplate.postForObject(url, resource, resourceClass);
  }

  @Override
  public T update(String dlcmId, T resource) {
    final String url = this.getBaseUrl() + resourcePath + "/" + "{dlcmId}";
    RestTemplate restTemplate = RestClientTool.getClient(getBaseUrl());
    Map<String, String> params = new HashMap<>();
    params.put(DLCM_ID_LABEL, dlcmId);
    return restTemplate.exchange(url, HttpMethod.PUT, new HttpEntity<>(resource), resourceClass, params).getBody();
  }

  @Override
  public void delete(String dlcmId) {
    final String url = this.getBaseUrl() + resourcePath + "/" + "{dlcmId}";
    RestTemplate restTemplate = RestClientTool.getClient(getBaseUrl());
    Map<String, String> params = new HashMap<>();
    params.put(DLCM_ID_LABEL, dlcmId);
    restTemplate.delete(url, params);
  }

  @Override
  public void searchByName(String name) {

  }

  protected String getBaseUrl() {
    return getBaseUrl(this.module);
  }

  private String getAssociationUrl(String resourceLabel){
    return this.getBaseUrl() + resourcePath + "/" + "{dlcmId}" + "/" + resourceLabel;
  }

  protected T doAction(String dlcmId, String action) {
    T result = null;
    final String url = this.getAssociationUrl(action);
    RestTemplate restTemplate = RestClientTool.getClient(getBaseUrl());
    Map<String, String> params = new HashMap<>();
    params.put(DLCM_ID_LABEL, dlcmId);
    try {
      String jsonString = restTemplate.getForObject(url, String.class, params);
      result = mapper.readValue(jsonString, resourceClass);
    } catch (IOException e) {
      logger.error(e);
    } catch (HttpClientErrorException e) {
      if (e.getRawStatusCode() == 404) {
        // No item found return null
        return null;
      } else {
        throw e;
      }
    }
    return result;
  }

  protected List<T> findAll(Class<T> itemClass) {
    final String url = this.getBaseUrl() + resourcePath + "?size=" + Integer.MAX_VALUE;
    RestTemplate restTemplate = RestClientTool.getClient(getBaseUrl());
    String jsonString = restTemplate.getForObject(url, String.class);
    return getList(jsonString, itemClass);
  }

  protected <U extends RESTResource<U>> void addItem(Map<String,?> params, String resourceLabel, Class<U> resourceClass, Map<String,?> resourceIds) {
    RestTemplate restTemplate = RestClientTool.getClient(getBaseUrl());
    restTemplate.postForObject(getAssociationUrl(resourceLabel), resourceIds, Array.newInstance(resourceClass, 0).getClass(), params);
  }

  protected <U extends RESTResource<U>> void addItem(String dlcmId, String itemTypeLabel, String itemId, Class<U> itemClass) {
    @SuppressWarnings("unchecked")
    Class<? extends RESTResource<U>[]> arrayType = (Class<? extends RESTResource<U>[]>) Array.newInstance(itemClass, 0).getClass();
    Map<String, String> params = new HashMap<>();
    params.put(DLCM_ID_LABEL, dlcmId);
    RestTemplate restTemplate = RestClientTool.getClient(getBaseUrl());
    restTemplate.postForObject(getAssociationUrl(itemTypeLabel), new String[] { itemId }, arrayType, params);
  }

  protected <U extends RESTResource<U>> U addItem(String dlcmId, String itemTypeLabel, U item, Class<U> itemClass) {
    RestTemplate restTemplate = RestClientTool.getClient(getBaseUrl());
    return restTemplate.postForObject(getAssociationUrl(itemTypeLabel), item, itemClass, dlcmId);
  }

  protected void removeItem(String parentId, Map<String, String> subResources) {
    final StringBuilder paramUri = new StringBuilder();
    subResources.forEach((param, subResourcesContent) -> {
      paramUri.append("/" + param + "/{" +  param + "}");
    });
    String url = this.getBaseUrl() + resourcePath + "/" + parentId + paramUri.toString();
    RestTemplate restTemplate = RestClientTool.getClient(getBaseUrl());
    restTemplate.delete(url, subResources);
  }

  protected void removeItem(String parentId, String itemTypeLabel, String itemId) {
    final String url = this.getBaseUrl() + resourcePath + "/" + parentId + "/" + itemTypeLabel + "/" + itemId;
    RestTemplate restTemplate = RestClientTool.getClient(getBaseUrl());
    restTemplate.delete(url);
  }

  protected <U> List<U> findAllLinkedResources(String dlcmId, Class<U> itemClass) {
    final String url = this.getBaseUrl() + resourcePath + "/" + "{dlcmId}" + "/" + RESTName.getRESTName(itemClass) + "?size=" + Integer.MAX_VALUE;
    Map<String, String> params = new HashMap<>();
    params.put(DLCM_ID_LABEL, dlcmId);
    RestTemplate restTemplate = RestClientTool.getClient(getBaseUrl());
    String jsonString = restTemplate.getForObject(url, String.class, params);
    return getList(jsonString, itemClass);
  }

  protected <U> U[] uploadFiles(String dlcmId, ParameterizedTypeReference<U[]> responseType, List<Resource> resources) {
    final String url = this.getBaseUrl() + resourcePath + "/" + dlcmId + "/upload";
    MultiValueMap<String, Object> map = new LinkedMultiValueMap<>();
    for (Resource r : resources) {
      map.add("file", r);
    }
    HttpHeaders headers = new HttpHeaders();
    headers.setContentType(MediaType.MULTIPART_FORM_DATA);
    HttpEntity<MultiValueMap<String, Object>> requestEntity = new HttpEntity<>(map, headers);
    RestTemplate restTemplate = RestClientTool.getClient(getBaseUrl());
    ResponseEntity<U[]> res = restTemplate.exchange(url, HttpMethod.POST, requestEntity, responseType);
    return res.getBody();
  }

  protected ByteArrayOutputStream download(String dlcmId, String itemId) {
    final String url = this.getBaseUrl() + resourcePath + "/" + dlcmId + "/" + RESTName.DATAFILE + "/" + itemId + "/download";
    ResponseExtractor<ByteArrayOutputStream> responseExtractor = response -> {
      InputStream inputStream = response.getBody();
      ByteArrayOutputStream buffer = new ByteArrayOutputStream();
      int nRead;
      byte[] data = new byte[BUFFER_SIZE];
      try {
        while ((nRead = inputStream.read(data, 0, data.length)) != -1) {
          buffer.write(data, 0, nRead);
        }
        buffer.flush();
      } catch (IOException e) {
        logger.error("Error in downloading file", e);
      }
      return buffer;
    };
    return RestClientTool.downloadContent(URI.create(url), responseExtractor);
  }

  private <U> List<U> getList(String jsonString, Class<U> itemClass) {
    List<U> result = new ArrayList<>();
    try {
      JsonNode jsonPostResult = mapper.readTree(jsonString);
      String resourceArrayString = jsonPostResult.get("_data").toString();
      JavaType listItemType = mapper.getTypeFactory().constructParametricType(List.class, itemClass);
      result = mapper.readValue(resourceArrayString, listItemType);
    } catch (IOException e) {
      logger.error("Error in building list", e);
    }
    return result;
  }
}
