package ch.dlcm.test.access;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;

import java.util.ArrayList;
import java.util.List;

import org.junit.Test;
import org.junit.runner.RunWith;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.test.context.junit4.SpringJUnit4ClassRunner;

import ch.dlcm.model.Order;
import ch.dlcm.model.QueryType;
import ch.dlcm.model.Status;
import ch.dlcm.model.Status.StatusValue;
import ch.dlcm.model.oais.ArchivalInfoPackage;
import ch.dlcm.service.access.OrderService;
import ch.dlcm.service.archivalstorage.AIPService;
import ch.dlcm.test.DLCMTestConstants;

@RunWith(SpringJUnit4ClassRunner.class)
@SpringBootTest
public class OrderTest {

  @Autowired
  private OrderService orderService;
  @Autowired
  private AIPService aipService;

  @Test
  public void creationTest() {
    // Create a deposit
    Order order1 = new Order();
    order1.setName(DLCMTestConstants.getLabel("Order "));
    Order order2 = orderService.create(order1);

    // Test the creation return
    assertNotNull(order2);
    assertEquals(order1.getName(), order2.getName());
    assertEquals(order1.getCreationTime(), order2.getCreationTime());

    // Test creation
    Order order3 = orderService.findOne(order2.getDlcmId());
    assertNotNull(order3);
    assertEquals(order1.getName(), order3.getName());
    assertEquals(order1.getCreationTime(), order3.getCreationTime());

    // Check if pending ( InProgress, inValidation, inError )
    assertEquals(Status.StatusValue.InProgress, order2.getStatus().getValue());

  }

  @Test
  public void directQueryTest() {
    // With 1 AIP
    executeDirectQuery(1);
    // With 3 AIPs
    executeDirectQuery(3);
  }

  @Test
  public void simpleQueryTest() {
    Order order = orderService.create(buildSimpleQuery("Faust"));
    assertNotNull(order);
    executeQuery(order);
  }

  private void executeDirectQuery(int num) {
    Order order = orderService.create(buildDirectQuery(num));
    assertNotNull(order);
    executeQuery(order);
  }

  private void executeQuery(Order order) {
    // Save AIP
    order = orderService.save(order.getDlcmId());
    assertNotNull(order);

    // Submit query
    Order order2 = new Order();
    order2.getStatus().setValue(StatusValue.Submitted);
    Order order3 = orderService.update(order.getDlcmId(), order2);
    assertNotNull(order);

    // Check status
    int count = 0;

    while (order3.getStatus().getValue() != Status.StatusValue.Ready
            && order3.getStatus().getValue() != Status.StatusValue.InError) {
      DLCMTestConstants.checkTries(count);
      DLCMTestConstants.waitInSeconds();
      order3 = orderService.findOne(order.getDlcmId());
      count++;
    }

    // Check if ready
    assertEquals(Status.StatusValue.Ready, order3.getStatus().getValue());
  }

  private Order buildSimpleQuery(String q) {
    // Create a deposit
    Order order = new Order();
    order.setName(DLCMTestConstants.getLabel("Simple Query with '" + q + "' "));
    order.setQueryType(QueryType.SIMPLE);
    order.setQuery(q);
    return order;
  }

  private Order buildDirectQuery(int num) {
    // Get AIP id
    List<String> aipList = getAIP(num);
    assertEquals(aipList.size(), num);

    // Create a deposit
    Order order = new Order();
    order.setName(DLCMTestConstants.getLabel("Direct Query with " + num + " item(s) "));
    order.setQueryType(QueryType.DIRECT);
    String aips = aipList.get(0);
    if (aipList.size() > 1) {
      for (int i = 1; i < aipList.size(); i++) {
        aips += ";" + aipList.get(i);
      }
    }
    order.setQuery(aips);
    return order;
  }

  private List<String> getAIP(int max) {
    List<String> list = new ArrayList<>();
    int count = 0;

    List<ArchivalInfoPackage> aipList = aipService.findAll();
    for (ArchivalInfoPackage aip : aipList) {
      if (aip.getInfo().getStatus().getValue() == StatusValue.Completed) {
        count++;
        list.add(aip.getDlcmId());
        if (count == max) {
          break;
        }
      }
    }
    return list;
  }

  @Test
  public void updateTest() {
    // Create a deposit
    Order order = new Order();
    order.setName(DLCMTestConstants.getLabel("A"));
    order = orderService.create(order);

    // Update the deposit
    order.setName(DLCMTestConstants.getLabel("B"));
    Order actualOrder = orderService.update(order.getDlcmId(), order);
    Order order4 = orderService.findOne(order.getDlcmId());

    // Test the update
    assertNotNull(actualOrder);
    assertEquals(order.getName(), order4.getName());
    assertEquals(order.getCreationTime(), order4.getCreationTime());

    Status status = new Status();
    status.setValue(Status.StatusValue.Submitted);
    order.setStatus(status);
    order4 = orderService.update(order4.getDlcmId(), order);
    assertEquals(Status.StatusValue.Submitted, order4.getStatus().getValue());
  }

  @Test
  public void deleteTest() {
    // Create a new deposit
    Order order = new Order();
    order.setName(DLCMTestConstants.getLabel("Order "));
    order = orderService.create(order);

    // Delete the deposit
    String dlcmId = order.getDlcmId();
    orderService.delete(dlcmId);
    order = orderService.findOne(dlcmId);

    // Test deletion
    assertNull(order);
  }
}