package ch.dlcm.test.admin;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;

import java.util.List;

import org.junit.Test;
import org.junit.runner.RunWith;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.test.context.junit4.SpringJUnit4ClassRunner;

import ch.dlcm.model.FundingAgency;
import ch.dlcm.model.OrganizationalUnit;
import ch.dlcm.test.DLCMTestConstants;

@RunWith(SpringJUnit4ClassRunner.class)
@SpringBootTest
public class FundingAgencyTest extends DLCMAdminTest {

  @Test
  public void testCreate() {

    String agencyAcronym = DLCMTestConstants.getLabel("FA");
    String agencyName    = DLCMTestConstants.getLabel("My FundingAgency");
    
    /*
     * Create a new FundingAgency
     */
    FundingAgency new_agency = new FundingAgency();
    new_agency.setAcronym(agencyAcronym);
    new_agency.setName(agencyName);
    new_agency = fundingAgencyService.create(new_agency);
    
    /*
     * Fetch it from server and test it has been created
     */
    FundingAgency fetched_agency = fundingAgencyService.findOne(new_agency.getDlcmId());
    assertNotNull(fetched_agency.getDlcmId());
    assertEquals(agencyName, fetched_agency.getName());
  }

  @Test
  public void testUpdate() {
    
    String agencyAcronym1 = DLCMTestConstants.getLabel("FA1");
    String agencyName1    = DLCMTestConstants.getLabel("Funding Agency unit");
    String agencyAcronym2 = DLCMTestConstants.getLabel("FA2");
    String agencyName2    = DLCMTestConstants.getLabel("Funding Agency updated");
    
    FundingAgency new_agency = new FundingAgency();
    new_agency.setAcronym(agencyAcronym1);
    new_agency.setName(agencyName1);
    new_agency = fundingAgencyService.create(new_agency);
    
    /*
     * Checks saving succeeded
     */
    FundingAgency fetched_agency = fundingAgencyService.findOne(new_agency.getDlcmId());
    assertEquals("names are equal", fetched_agency.getName(), agencyName1);
    
    /*
     * Does the update
     */
    fetched_agency.setAcronym(agencyAcronym2);
    fetched_agency.setName(agencyName2);
    fundingAgencyService.update(fetched_agency.getDlcmId(), fetched_agency);
    
    /*
     * Checks update succeeded
     */
    FundingAgency refetched_agency = fundingAgencyService.findOne(fetched_agency.getDlcmId());
    assertEquals("names are equal", refetched_agency.getName(), agencyName2);
    assertEquals("acronyms are equal", refetched_agency.getAcronym(), agencyAcronym2);
  }

  @Test
  public void testDelete() {
    
    String agencyAcronym = DLCMTestConstants.getLabel("FA");
    String agencyName    = DLCMTestConstants.getLabel("My FundingAgency");
    
    /*
     * Create a new FundingAgency
     */
    FundingAgency new_agency = new FundingAgency();
    new_agency.setAcronym(agencyAcronym);
    new_agency.setName(agencyName);
    new_agency = fundingAgencyService.create(new_agency);
    
    /*
     * Check it has been created
     */
    FundingAgency fetched_agency = fundingAgencyService.findOne(new_agency.getDlcmId());
    assertNotNull(fetched_agency.getDlcmId());
    
    /*
     * Delete it 
     */
    fundingAgencyService.delete(fetched_agency.getDlcmId());
    
    /*
     * Fetching again returns nothing
     */
    FundingAgency re_fetched_agency = fundingAgencyService.findOne(new_agency.getDlcmId());
    assertNull(re_fetched_agency);
  }
  
  @Test
  public void testAddAndRemoveOrganizationalUnit() {
    
    String fundingAgencyAcronym   = DLCMTestConstants.getLabel("FA");
    String fundingAgencyName      = DLCMTestConstants.getLabel("My FundingAgency with Unit");
    String organizationalUnitName = DLCMTestConstants.getLabel("My OrganizationalUnit");
    
    /*
     * Create a new FundingAgency
     */
    FundingAgency new_agency = new FundingAgency();
    new_agency.setAcronym(fundingAgencyAcronym);
    new_agency.setName(fundingAgencyName);
    new_agency = fundingAgencyService.create(new_agency);
    
    /*
     * Create a new OrganizationalUnit
     */
    OrganizationalUnit new_unit = new OrganizationalUnit();
    new_unit.setName(organizationalUnitName);
    new_unit = organizationalUnitService.create(new_unit);
    
    /*
     * Add the OrganizationalUnit to the FundingAgency
     */
    fundingAgencyService.addOrganizationalUnit(new_agency.getDlcmId(), new_unit.getDlcmId());
    
    /*
     * Check it has been added
     */
    List<OrganizationalUnit> units = fundingAgencyService.getOrganizationalUnits(new_agency.getDlcmId());
    assertNotNull(units);
    assertEquals(1, units.size());
    OrganizationalUnit u = units.get(0);
    assertEquals(organizationalUnitName, u.getName());
    
    /*
     * Remove the OrganizationalUnit
     */
    fundingAgencyService.removeOrganizationalUnit(new_agency.getDlcmId(), new_unit.getDlcmId());
    
    /*
     * Check it has been removed
     */
    units = fundingAgencyService.getOrganizationalUnits(new_agency.getDlcmId());
    assertNotNull(units);
    assertEquals(0, units.size());
    
    /*
     * Create 3 OrganizationalUnits and link them to the FundingAgency
     */
    String unitName1    = DLCMTestConstants.getLabel("OrganizationalUnit 1");
    String unitName2    = DLCMTestConstants.getLabel("OrganizationalUnit 2");
    String unitName3    = DLCMTestConstants.getLabel("OrganizationalUnit 3");
    String[] unit_names = new String[] {unitName1, unitName2, unitName3};
    
    String unitIdToDelete = null;
    
    int i = 0;
    for (String unit_name : unit_names) {
      OrganizationalUnit indexed_unit = new OrganizationalUnit();
      indexed_unit.setName(unit_name);
      indexed_unit = organizationalUnitService.create(indexed_unit);
      
      fundingAgencyService.addOrganizationalUnit(new_agency.getDlcmId(), indexed_unit.getDlcmId());
      
      if (i==1) {
        unitIdToDelete = indexed_unit.getDlcmId();
      }
      
      i++;
    }
    
    /*
     * Check they are all linked correctly
     */
    units = fundingAgencyService.getOrganizationalUnits(new_agency.getDlcmId());
    assertNotNull(units);
    assertEquals(3, units.size());
    
    /*
     * Delete one specifically and check it is not in the list anymore
     */
    fundingAgencyService.removeOrganizationalUnit(new_agency.getDlcmId(), unitIdToDelete);
    units = fundingAgencyService.getOrganizationalUnits(new_agency.getDlcmId());
    assertNotNull(units);
    assertEquals(2, units.size());
    for (OrganizationalUnit fetched_u : units) {
      assertNotEquals(unitIdToDelete, fetched_u.getDlcmId());
    }
  }
  
  /*************************************************/
  
  protected void deleteFixtures()
  {
    this.clearFundingAgencyFixtures();
    this.clearOrganizationalUnitFixtures();
  }
}
