package ch.dlcm.test.admin;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import java.io.File;
import java.net.URI;
import java.net.URL;
import java.util.Arrays;
import java.util.List;
import java.util.concurrent.ThreadLocalRandom;

import org.json.JSONObject;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.test.context.junit4.SpringJUnit4ClassRunner;

import com.fasterxml.jackson.databind.ObjectMapper;

import ch.dlcm.model.License;
import ch.dlcm.service.admin.LicenseService;

@RunWith(SpringJUnit4ClassRunner.class)
@SpringBootTest
public class LicenseTest {

    private static final String CKAN_IMPORT_FILE_NAME = "import/ckan.json";
    private static final String ALL_IMPORT_FILE_NAME = "import/all.json";
    private static final String DLCM_IMPORT_FILE_NAME = "import/dlcm.json";


    @Autowired private LicenseService licenseService;

    @Test
    public void creationTest() {
       //  Create a license
        License expectedLicense = new License();
        expectedLicense.setTitle("GNU Free Documentation License ");
        String gnuLicense = "GNU"+ ThreadLocalRandom.current().nextInt();
        expectedLicense.setOpenLicenseId(gnuLicense);
        expectedLicense = licenseService.create(expectedLicense);
        License actualLicense = licenseService.findOne(expectedLicense.getDlcmId());

        // Test the creation
        assertNotNull(expectedLicense.getDlcmId());
        assertsLicense(expectedLicense, actualLicense);

        // Test unique openLicenseId
        License sameLicense = new License();
        expectedLicense.setTitle("GNU Free Documentation License");
        expectedLicense.setOpenLicenseId(gnuLicense);
        try {
            licenseService.create(sameLicense);
            fail("Should raise exception: Unique constraint openLicenseId" );
        }catch(Exception e){
            assertEquals("500", e.getMessage().trim());
        }
    }

    @Test
    public void updateTest() throws Exception {
        //Create a licence
        License expectedLicense = new License();
        expectedLicense.setTitle("A"+ ThreadLocalRandom.current().nextInt());
        expectedLicense.setOpenLicenseId("A"+ ThreadLocalRandom.current().nextInt());
        expectedLicense = licenseService.create(expectedLicense);

        // Update the licence
        expectedLicense.setTitle("B"+ ThreadLocalRandom.current().nextInt());
        String openLicenseId = "B"+ ThreadLocalRandom.current().nextInt();
        expectedLicense.setOpenLicenseId(openLicenseId);
        expectedLicense.setDomainContent(true);
        expectedLicense.setDomainData(true);
        expectedLicense.setDomainSoftware(true);
        expectedLicense.setFamily("a family");
        expectedLicense.setIsGeneric(true);
        expectedLicense.setMaintainer("the GNU");
        expectedLicense.setOdConformance(License.ConformanceStatus.APPROVED);
        expectedLicense.setOsdConformance(License.ConformanceStatus.APPROVED);
        expectedLicense.setStatus(License.LicenseStatus.ACTIVE);
        expectedLicense.setUrl(new URI("http://dlcm"));
        License actualLicense = licenseService.update(expectedLicense.getDlcmId(), expectedLicense);

        // Test the update
        assertNotNull(actualLicense);
        assertsLicense(expectedLicense, actualLicense);

        // Test unique openLicenseId
        License sameLicense = new License();
        sameLicense.setOpenLicenseId("C"+ ThreadLocalRandom.current().nextInt());
        sameLicense = licenseService.create(sameLicense);
        sameLicense.setOpenLicenseId(openLicenseId);
        try {
            licenseService.update(sameLicense.getDlcmId(), sameLicense);
            fail("Should raise exception: Unique constraint openLicenseId" );
        }catch(Exception e){
            assertEquals("500", e.getMessage().trim());
        }
    }

    private void assertsLicense(License expectedLicense, License actualLicense){
        assertEquals(expectedLicense.getTitle(), actualLicense.getTitle());
        assertEquals(expectedLicense.getCreationTime(), actualLicense.getCreationTime());
        assertEquals(expectedLicense.getOpenLicenseId(), actualLicense.getOpenLicenseId());
        assertEquals(expectedLicense.getDomainContent(), actualLicense.getDomainContent());
        assertEquals(expectedLicense.getDomainData(), actualLicense.getDomainData());
        assertEquals(expectedLicense.getDomainSoftware(), actualLicense.getDomainSoftware());
        assertEquals(expectedLicense.getFamily(), actualLicense.getFamily());
        assertEquals(expectedLicense.getIsGeneric(), actualLicense.getIsGeneric());
        assertEquals(expectedLicense.getMaintainer(), actualLicense.getMaintainer());
        assertEquals(expectedLicense.getOdConformance(), actualLicense.getOsdConformance());
        assertEquals(expectedLicense.getOsdConformance(), actualLicense.getOsdConformance());
        assertEquals(expectedLicense.getStatus(), actualLicense.getStatus());
        assertEquals(expectedLicense.getUrl(), actualLicense.getUrl());
    }

    @Test
    public void deleteTest(){
        // Create a new licence
        License expectedLicense = new License();
        expectedLicense.setTitle("License  "+ ThreadLocalRandom.current().nextInt());
        expectedLicense.setOpenLicenseId("License  "+ ThreadLocalRandom.current().nextInt());
        expectedLicense = licenseService.create(expectedLicense);

        // Delete
        String dlcmId = expectedLicense.getDlcmId();
        licenseService.delete(dlcmId);
        License actualLicense = licenseService.findOne(dlcmId);

        //Test deletion
        assertNull(actualLicense);
    }

    @Test
    public void findAllTest(){
        // Create a new licence
        License expectedLicense = new License();
        expectedLicense.setTitle("License  "+ ThreadLocalRandom.current().nextInt());
        expectedLicense.setOpenLicenseId("License  "+ ThreadLocalRandom.current().nextInt());

        licenseService.create(expectedLicense);

        // Find all
        List<License> actualLicenses = licenseService.findAll();

        //Test at least
        assertFalse(actualLicenses.isEmpty());
        assertTrue(actualLicenses.size() >= 1);
    }

    @Test
    public void importsLicenseTest() throws Exception {
        // Test import as list
        String expectedPayload = new JSONObject().put("imported", 15).toString();
        URL url = this.getClass().getClassLoader().getResource(DLCM_IMPORT_FILE_NAME);
        ObjectMapper mapper = new ObjectMapper();
        License[] licences = mapper.readValue(new File(url.getPath()), License[].class);
        String actualPayload = licenseService.createLicenses(Arrays.asList(licences)).getBody();
        assertEquals(expectedPayload, actualPayload);
        // Test import as file dlcm format
        expectedPayload = new JSONObject().put("imported", 15).toString();
        actualPayload = licenseService.importLicenseFile(DLCM_IMPORT_FILE_NAME).getBody();
        assertEquals(expectedPayload, actualPayload);
        // Test import as openLicense file format
        expectedPayload = new JSONObject().put("imported", 109).toString();
        actualPayload = licenseService.importOpenLicenseFile(ALL_IMPORT_FILE_NAME).getBody();
        assertEquals(expectedPayload, actualPayload);
        // Test import as ckan file format
        expectedPayload = new JSONObject().put("imported", 15).toString();
        actualPayload = licenseService.importOpenLicenseFile(CKAN_IMPORT_FILE_NAME).getBody();
        assertEquals(expectedPayload, actualPayload);
    }

}
