package ch.dlcm.test.admin;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;

import java.util.List;
import java.util.concurrent.ThreadLocalRandom;

import org.junit.Test;
import org.junit.runner.RunWith;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.test.context.junit4.SpringJUnit4ClassRunner;

import ch.dlcm.model.FundingAgency;
import ch.dlcm.model.Institution;
import ch.dlcm.model.OrganizationalUnit;
import ch.dlcm.model.Person;
import ch.dlcm.service.admin.FundingAgencyService;
import ch.dlcm.service.admin.InstitutionService;
import ch.dlcm.service.admin.OrganizationalUnitService;
import ch.dlcm.service.admin.PersonService;
import ch.dlcm.test.DLCMTestConstants;

@RunWith(SpringJUnit4ClassRunner.class)
@SpringBootTest
public class OrganizationalUnitTest extends DLCMAdminTest {

  @Autowired
  private OrganizationalUnitService organizationalUnitService;
  @Autowired
  private PersonService personService;
  @Autowired
  private InstitutionService institutionService;
  @Autowired
  private FundingAgencyService fundingAgencyService;

  /*************************************************/
  
  @Test
  public void creationTest() {
    // Create a project
    OrganizationalUnit project1 = new OrganizationalUnit();
    project1.setName(DLCMTestConstants.getLabel("DLCM project"));
    project1.setDescription("Swiss Research Data Preservation #CREATE");
    project1 = organizationalUnitService.create(project1);
    OrganizationalUnit project2 = organizationalUnitService.findOne(project1.getDlcmId());

    // Test the creation
    assertEquals(project1.getName(), project2.getName());
    assertEquals(project1.getDescription(), project2.getDescription());
    assertEquals(project1.getCreationTime(), project2.getCreationTime());
  }

  @Test
  public void updateTest() {
    // Create a project
    OrganizationalUnit project1 = new OrganizationalUnit();
    project1.setName(DLCMTestConstants.getLabel("DLCM project"));
    project1.setDescription("Swiss Research Data Preservation #UPDATE");
    project1 = organizationalUnitService.create(project1);

    // Update the person
    OrganizationalUnit project2 = new OrganizationalUnit();
    project2.setName(DLCMTestConstants.getLabel("DLCM project 2.0"));
    project2.setDescription("Swiss Research Data Preservation (v2) #UPDATE");
    OrganizationalUnit project3 = organizationalUnitService.update(project1.getDlcmId(), project2);

    // Test the update
    assertEquals(project1.getDlcmId(), project3.getDlcmId());
    assertEquals(project2.getName(), project3.getName());
    assertEquals(project2.getDescription(), project3.getDescription());
  }

  @Test
  public void deleteTest() {
    // Create a new deposit
    OrganizationalUnit project = new OrganizationalUnit();
    project.setName(DLCMTestConstants.getLabel("DLCM project"));
    project.setDescription("Swiss Research Data Preservation");
    project = organizationalUnitService.create(project);

    // Delete the deposit
    String dlcmId = project.getDlcmId();
    organizationalUnitService.delete(dlcmId);
    project = organizationalUnitService.findOne(dlcmId);

    //Test deletion
    assertNull(project);
  }

  @Test
  public void institutionTest() {
    // Create a project
    OrganizationalUnit project = new OrganizationalUnit();
    project.setName(DLCMTestConstants.getLabel("DLCM project"));
    project.setDescription("Swiss Research Data Preservation #INSTITUTION");
    project = organizationalUnitService.create(project);

    // Create an institution
    Institution institution1 = new Institution();
    institution1.setName(DLCMTestConstants.getLabel("institution "));
    institution1.setDescription("description" + ThreadLocalRandom.current().nextInt());
    institution1 = institutionService.create(institution1);

    // Add the institution to the project
    organizationalUnitService.addInstitution(project.getDlcmId(), institution1.getDlcmId());
    List<Institution> list = organizationalUnitService.getInstitutions(project.getDlcmId());
    Institution institution2 = list.get(0);

    // Test the institution
    assertEquals(institution1.getDlcmId(), institution2.getDlcmId());
    assertEquals(institution1.getName(), institution2.getName());
    assertEquals(institution1.getDescription(), institution2.getDescription());
  }

  @Test
  public void addPersonToTheProjectTest() {
    // Create a project
    OrganizationalUnit project = new OrganizationalUnit();
    project.setName(DLCMTestConstants.getLabel("DLCM project"));
    project.setDescription("Swiss Research Data Preservation #PEOPLE");
    project = organizationalUnitService.create(project);

    // Create a person
    Person person = new Person();
    person.setFirstName(DLCMTestConstants.getLabel("Any First name"));
    person.setLastName(DLCMTestConstants.getLabel("Any Last name"));
    person = personService.create(person);

    // Add the person to the project with default role
    organizationalUnitService.addPerson(project.getDlcmId(), person.getDlcmId());
    List<Person> people = organizationalUnitService.getPeople(project.getDlcmId());

    assertEquals(1, people.size());

    // Add the person to the project with other roles
    organizationalUnitService.addPerson(project.getDlcmId(), person.getDlcmId(), "Approver", "Creator");
    people = organizationalUnitService.getPeople(project.getDlcmId());

    assertEquals(1, people.size());
    Person person2 = people.get(0);

    // Test added person on the project
    assertEquals(person.getDlcmId(), person2.getDlcmId());
    assertEquals(person.getFirstName(), person2.getFirstName());
    assertEquals(person.getLastName(), person2.getLastName());

    // Test remove a role from the person in project
    organizationalUnitService.removePerson(project.getDlcmId(), person2.getDlcmId(), "Approver");
    assertEquals(1, people.size());
    // Test remove a person from project
    organizationalUnitService.removePerson(project.getDlcmId(), person2.getDlcmId());
    people = organizationalUnitService.getPeople(project.getDlcmId());
    assertEquals(true, people.isEmpty());

  }

  @Test
  public void testAddAndRemoveFundingAgency() {
    
    String organizationalUnitName = DLCMTestConstants.getLabel("My OrganizationalUnit with FundingAgency");
    String fundingAgencyAcronym   = DLCMTestConstants.getLabel("FA");
    String fundingAgencyName      = DLCMTestConstants.getLabel("My FundingAgency");
    
    /*
     * Create a new OrganizationalUnit
     */
    OrganizationalUnit new_unit = new OrganizationalUnit();
    new_unit.setName(organizationalUnitName);
    new_unit = organizationalUnitService.create(new_unit);
    
    /*
     * Create a new FundingAgency
     */
    FundingAgency new_agency = new FundingAgency();
    new_agency.setAcronym(fundingAgencyAcronym);
    new_agency.setName(fundingAgencyName);
    new_agency = fundingAgencyService.create(new_agency);
    
    /*
     * Add the FundingAgency to the OrganizationalUnit 
     */
    organizationalUnitService.addFundingAgency(new_unit.getDlcmId(), new_agency.getDlcmId());
    
    /*
     * Check it has been added
     */
    List<FundingAgency> fundingAgencies = organizationalUnitService.getFundingAgencies(new_unit.getDlcmId());
    assertNotNull(fundingAgencies);
    assertEquals(1, fundingAgencies.size());
    FundingAgency f = fundingAgencies.get(0);
    assertEquals(fundingAgencyName, f.getName());
    
    /*
     * Remove the FundingAgency
     */
    organizationalUnitService.removeFundingAgency(new_unit.getDlcmId(), new_agency.getDlcmId());
    
    /*
     * Check it has been removed
     */
    fundingAgencies = organizationalUnitService.getFundingAgencies(new_unit.getDlcmId());
    assertNotNull(fundingAgencies);
    assertEquals(0, fundingAgencies.size());
    
    /*
     * Create 3 FundingAgencies and link them to the OrganizationalUnit
     */
    FundingAgency subFA1 = new FundingAgency();
    subFA1.setAcronym(DLCMTestConstants.getLabel("Sub FA 1"));
    subFA1.setName(DLCMTestConstants.getLabel("Sub FundingAgency 1"));
    
    FundingAgency subFA2 = new FundingAgency();
    subFA2.setAcronym(DLCMTestConstants.getLabel("Sub FA 2"));
    subFA2.setName(DLCMTestConstants.getLabel("Sub FundingAgency 2"));
    
    FundingAgency subFA3 = new FundingAgency();
    subFA3.setAcronym(DLCMTestConstants.getLabel("Sub FA 3"));
    subFA3.setName(DLCMTestConstants.getLabel("Sub FundingAgency 3"));
    
    FundingAgency[] subAgencies = new FundingAgency[] {subFA1, subFA2, subFA3};
    
    String agencyIdToDelete = null;
    
    int i = 0;
    for (FundingAgency subAgency : subAgencies) {
      subAgency = fundingAgencyService.create(subAgency);
      organizationalUnitService.addFundingAgency(new_unit.getDlcmId(), subAgency.getDlcmId());
      
      if (i == 1) {
        agencyIdToDelete = subAgency.getDlcmId();
      }
      
      i++;
    }
    
    /*
     * Check they are all linked correctly
     */
    fundingAgencies = organizationalUnitService.getFundingAgencies(new_unit.getDlcmId());
    assertNotNull(fundingAgencies);
    assertEquals(3, fundingAgencies.size());
    
    /*
     * Delete one specifically and check it is not in the list anymore
     */
    organizationalUnitService.removeFundingAgency(new_unit.getDlcmId(), agencyIdToDelete);
    fundingAgencies = organizationalUnitService.getFundingAgencies(new_unit.getDlcmId());
    assertNotNull(fundingAgencies);
    assertEquals(2, fundingAgencies.size());
    for (FundingAgency fetched_agency : fundingAgencies) {
      assertNotEquals(agencyIdToDelete, fetched_agency.getDlcmId());
    }
    
  }
  
  /*************************************************/
  
  protected void deleteFixtures()
  {
    this.clearOrganizationalUnitFixtures();
    this.clearFundingAgencyFixtures();
  }
}
