package ch.dlcm.test.preingest;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;

import java.io.ByteArrayOutputStream;
import java.net.URI;
import java.util.List;

import org.junit.Test;
import org.junit.runner.RunWith;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.core.io.ClassPathResource;
import org.springframework.test.context.junit4.SpringJUnit4ClassRunner;

import ch.dlcm.model.DataCategory;
import ch.dlcm.model.DataFile;
import ch.dlcm.model.Deposit;
import ch.dlcm.model.DepositDataFile;
import ch.dlcm.model.Status;
import ch.dlcm.model.Status.StatusValue;
import ch.dlcm.service.preingest.DepositService;
import ch.dlcm.test.DLCMTestConstants;

@RunWith(SpringJUnit4ClassRunner.class)
@SpringBootTest
public class DepositDataFileTest {

  private final static String REFERENCE_FILE = "http://www.unige.ch/rectorat/static/budget-2013.pdf";
  private final static int REFERENCE_FILE_SIZE = 917380;

  @Autowired
  private DepositService depositService;

  @Test
  public void creationTest() {
    // Create a local deposit with data file
    Deposit localDeposit = createLocalDeposit();

    // Create remote deposit with data file
    Deposit remoteDeposit = createRemoteDeposit(localDeposit);

    // Fetch the deposit data file list from the server
    List<DepositDataFile> listDepositDataFile = depositService
            .getDepositDataFile(remoteDeposit.getDlcmId());

    // Test the creation
    assertEquals(localDeposit.getDataFiles().get(0).getDlcmId(),
            listDepositDataFile.get(0).getDlcmId());
    assertEquals(localDeposit.getDataFiles().get(0).getDataFile().getFormat(), ".pdf");
  }

  @Test
  public void uploadTest() {
    String fileToUpload = "bootstrap.yml";
    // Create a deposit
    Deposit deposit = new Deposit();
    deposit.setTitle(DLCMTestConstants.getLabel("Upload "));
    deposit = depositService.create(deposit);

    // Upload file
    DepositDataFile depositDataFile = depositService.upload(deposit.getDlcmId(),
            new ClassPathResource(fileToUpload));

    List<DepositDataFile> list = depositService.getDepositDataFile(deposit.getDlcmId());

    // Test the upload
    assertTrue(list.size() == 1);
    assertEquals(list.get(0).getDlcmId(), depositDataFile.getDlcmId());
    assertTrue(list.get(0).getDataFile().getSourceData().toString().endsWith(fileToUpload));
    assertEquals(list.get(0).getDataFile().getFileName(), fileToUpload);
  }

  @Test
  public void downloadTest() {
    // Create a local deposit with data file
    Deposit localDeposit = createLocalDeposit();

    // Create remote deposit with data file
    Deposit remoteDeposit = createRemoteDeposit(localDeposit);

    // Wait that the file is processed
    waitDataFileIsReady(remoteDeposit.getDlcmId(), 1);

    // Download the file
    ByteArrayOutputStream buffer = depositService.downloadDataFile(remoteDeposit.getDlcmId(),
            localDeposit.getDataFiles().get(0).getDlcmId());
    int fileLength = buffer.toByteArray().length;
    assertEquals(fileLength, REFERENCE_FILE_SIZE);
  }

  @Test
  public void updateTest() {

  }

  @Test
  public void deleteTest() {
    // Create a local deposit with data file
    Deposit localDeposit = createLocalDeposit();

    // Create remote deposit with data file
    Deposit remoteDeposit = createRemoteDeposit(localDeposit);

    // Test if data file is present
    int depositListSize = depositService.getDepositDataFile(remoteDeposit.getDlcmId()).size();
    assertEquals(depositListSize, 1);

    // Remove the data file
    String depositDataFileId = localDeposit.getDataFiles().get(0).getDlcmId();

    // Delete the data file
    depositService.removeDepositDataFile(remoteDeposit.getDlcmId(), depositDataFileId);

    // Test if data file is absent
    depositListSize = depositService.getDepositDataFile(remoteDeposit.getDlcmId()).size();
    assertEquals(depositListSize, 0);

  }

  @Test
  public void deleteWhenReadyTest() {
    // Create a local deposit with data file
    Deposit localDeposit = createLocalDeposit();

    // Create remote deposit with data file
    Deposit remoteDeposit = createRemoteDeposit(localDeposit);

    // Wait until data file downloaded
    waitDataFileIsReady(remoteDeposit.getDlcmId(), 1);

    // Remove the data file
    String depositDataFileId = localDeposit.getDataFiles().get(0).getDlcmId();

    // Delete the data file
    depositService.removeDepositDataFile(remoteDeposit.getDlcmId(), depositDataFileId);

    // Test if data file is absent
    int depositListSize = depositService.getDepositDataFile(remoteDeposit.getDlcmId()).size();
    assertEquals(depositListSize, 0);

  }

  @Test
  public void reuploadSameFileTest(){
    // Test Bug DLCM-267

    // Create a local deposit with data file
    Deposit localDeposit = createLocalDeposit();

    // Create remote deposit with data file
    Deposit remoteDeposit = createRemoteDeposit(localDeposit);

    // Wait until data file downloaded
    waitDataFileIsReady(remoteDeposit.getDlcmId(), 1);

    // Remove the data file
    String depositDataFileId = localDeposit.getDataFiles().get(0).getDlcmId();

    // Delete the data file
    depositService.removeDepositDataFile(remoteDeposit.getDlcmId(), depositDataFileId);

    // Test if data file is absent
    int depositListSize = depositService.getDepositDataFile(remoteDeposit.getDlcmId()).size();
    assertEquals(depositListSize, 0);

    //Reupload the same deposit
    remoteDeposit = createRemoteDeposit(localDeposit);

    // Wait until data file downloaded
    waitDataFileIsReady(remoteDeposit.getDlcmId(), 1);

  }

  @Test
  public void approveTest() {
    String[] filesToUpload = { "bootstrap.yml", "dlcm.xml" };
    // Create a deposit
    Deposit deposit = new Deposit();
    deposit.setTitle(DLCMTestConstants.getLabel("Submission "));
    deposit = depositService.create(deposit);

    // Upload files
    for (String fileToUpload : filesToUpload) {
      assertNotNull(
              depositService.upload(deposit.getDlcmId(), new ClassPathResource(fileToUpload)));
    }

    // Test the uploads
    List<DepositDataFile> list = depositService.getDepositDataFile(deposit.getDlcmId());
    assertTrue(list.size() == filesToUpload.length);

    // Wait until data file downloaded
    waitDataFileIsReady(deposit.getDlcmId(), 2);

    // Approve deposit
    Deposit deposit1 = depositService.findOne(deposit.getDlcmId());
    deposit1.getStatus().setValue(StatusValue.Approved);
    Deposit deposit2 = depositService.update(deposit.getDlcmId(), deposit1);

    // Check status
    int count=0;
    
    while(deposit2.getStatus().getValue() != Status.StatusValue.Completed
            && deposit2.getStatus().getValue() != Status.StatusValue.InError) {
      DLCMTestConstants.checkTries(count);
      DLCMTestConstants.waitInSeconds();
      deposit2 = depositService.findOne(deposit.getDlcmId());
      count++;
    }
    
    assertEquals(Status.StatusValue.Completed, deposit2.getStatus().getValue());
  }

  private void waitDataFileIsReady(String depositId, int dataFileNumber) {
    List<DepositDataFile> dataFileList = depositService.getDepositDataFile(depositId);
    int count = 0;

    // Test if data file is present
    assertEquals(dataFileList.size(), dataFileNumber);

    while (!checkDataFileListIsCompleted(dataFileList)) {
      DLCMTestConstants.checkTries(count);
      DLCMTestConstants.waitInSeconds();
      dataFileList = depositService.getDepositDataFile(depositId);
      count++;
    }
    return;
  }

  private boolean checkDataFileListIsCompleted(List<DepositDataFile> dataFileList) {
    for (DepositDataFile df : dataFileList) {
      if (df.getDataFile().getStatus().getValue() != StatusValue.Ready) {
        return false;
      }
    }
    return true;
  }

  private Deposit createLocalDeposit() {
    // Create a deposit object
    Deposit deposit = new Deposit();
    deposit.setTitle(DLCMTestConstants.getLabel("Deposit "));

    // Create a data file object
    DataFile dataFile = new DataFile();
    dataFile.setSourceData(URI.create(REFERENCE_FILE));
    dataFile.setDataCategory(DataCategory.Secondary);
    dataFile.setDataType(DataCategory.Documentation);
    URI format = URI.create(REFERENCE_FILE);
    dataFile.setFormat(format.getPath().substring(format.getPath().lastIndexOf(".")));

    // Create a deposit data file object
    DepositDataFile depositDataFile = new DepositDataFile();
    depositDataFile.setDataFile(dataFile);

    // Add the deposit data file to the deposit object
    deposit.addItem(depositDataFile);
    return deposit;
  }

  private Deposit createRemoteDeposit(Deposit deposit) {
    // Create the deposit on the server
    Deposit remoteDeposit = depositService.create(deposit);

    // Create the deposit data file on the server
    depositService.addDepositDataFile(remoteDeposit.getDlcmId(), deposit.getDataFiles().get(0));
    return remoteDeposit;
  }
}
