package ch.dlcm.controller;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.nio.file.Files;
import java.nio.file.Paths;

import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.core.io.InputStreamResource;
import org.springframework.data.domain.Pageable;
import org.springframework.hateoas.mvc.ControllerLinkBuilder;
import org.springframework.http.HttpHeaders;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.web.multipart.MultipartFile;

import ch.dlcm.config.DLCMConfig;
import ch.dlcm.exception.DLCMException;
import ch.dlcm.model.DataFileInterface;
import ch.dlcm.model.Status.StatusValue;
import ch.dlcm.model.oais.InfoPackageInterface;
import ch.dlcm.rest.RESTAction;
import ch.dlcm.rest.RESTCollection;
import ch.dlcm.rest.RESTName;
import ch.dlcm.rest.RESTRepository;
import ch.dlcm.rest.RESTResult;
import ch.dlcm.rest.RESTResult.ActionStatus;
import ch.dlcm.rest.RESTTool;
import ch.dlcm.service.MessageService;
import ch.dlcm.tool.FileTool;

public abstract class DLCMController {
  private static final Logger log = Logger.getLogger(DLCMController.class);

  @Autowired
  protected DLCMConfig dlcmConfig;

  @Autowired
  protected MessageService messageService;

  protected ResponseEntity<InputStreamResource> download(String clazz, String id, URI uri,
          String target) {
    try {
      File file = FileTool.getFileFromUri(uri);
      if (target == null || target.isEmpty()) {
        target = file.getName();
      }

      HttpHeaders respHeaders = new HttpHeaders();
      respHeaders.setContentType(FileTool.getContentType(uri));
      respHeaders.setContentLength(file.length());
      respHeaders.setContentDispositionFormData("attachment", target);

      InputStreamResource inStreamRes = new InputStreamResource(new FileInputStream(file));
      return new ResponseEntity<>(inStreamRes, respHeaders, HttpStatus.OK);
    } catch (UnsupportedOperationException e) {
      log.error("Download not supported for " + clazz + " (" + id + ")", e);
      return new ResponseEntity<>(HttpStatus.NOT_IMPLEMENTED);
    } catch (IOException e) {
      log.error(e);
      return new ResponseEntity<>(HttpStatus.SERVICE_UNAVAILABLE);
    } catch (URISyntaxException e) {
      log.error(e);
      return new ResponseEntity<>(HttpStatus.NOT_FOUND);
    }
  }

  protected <T, V extends DataFileInterface> RESTResult resumeDataFile(String parentid, String id,
          RESTRepository<T> repo, Class<T> mainResourceClass, Class<V> subResourceClass) {
    T parentitem = repo.findOne(parentid);
    if (parentitem == null)
      return null;
    V item = repo.findByParentIdAndChildId(parentid, mainResourceClass, subResourceClass, id);
    if (item == null) {
      return null;
    }
    RESTResult res = new RESTResult(item.getDlcmId());
    if (item.getDataFile().getStatus().getValue() == StatusValue.Ready) {
      res.setMesssage(messageService.get("datafile.completed", new Object[] { id, parentid }));
    } else if (item.getDataFile().getStatus().getValue() == StatusValue.ToProcess) {
      res.setMesssage(messageService.get("datafile.inprogress", new Object[] { id, parentid }));
    } else {
      res.setStatus(ActionStatus.NotExecuted);
      if (item.getDataFile().getFinalData() != null) {
        item.setFileStatus(StatusValue.Processed);
      } else {
        item.setFileStatus(StatusValue.Received);
      }
      T parentitem2 = repo.save(parentitem);
      if (parentitem2 == null) {
        res.setMesssage(messageService.get("datafile.cannotupdate", new Object[] { id, parentid }));
      } else {
        res.setStatus(ActionStatus.Executed);
        res.setMesssage(messageService.get("datafile.resume", new Object[] { id, parentid }));
      }
    }
    return res;
  }

  protected <T extends InfoPackageInterface> RESTResult resumePackage(String id,
          RESTRepository<T> repo) {
    T item = repo.findOne(id);
    if (item == null) {
      return null;
    }
    RESTResult res = new RESTResult(item.getDlcmId());
    if (item.getInfo().getStatus().getValue() == StatusValue.Completed) {
      res.setMesssage(messageService.get("package.completed", new Object[] { id }));
    } else if (item.getInfo().getStatus().getValue() != StatusValue.InPreparation
            && item.getInfo().getStatus().getValue() != StatusValue.InError) {
      res.setMesssage(messageService.get("package.inprogress", new Object[] { id }));
    } else {
      res.setStatus(ActionStatus.NotExecuted);
      if (!item.isAllDataFileReady()) {
        res.setMesssage(messageService.get("package.datafile.notready", new Object[] { id }));
      } else {
        item.setPackageStatus(StatusValue.InPreparation);
        T item2 = repo.save(item);
        if (item2 == null) {
          res.setMesssage(messageService.get("package.cannotupdate", new Object[] { id }));
        } else {
          res.setStatus(ActionStatus.Executed);
          res.setMesssage(messageService.get("package.resume", new Object[] { id }));

        }
      }
    }
    return res;
  }

  protected <T extends InfoPackageInterface> RESTResult reindexPackage(String id,
          RESTRepository<T> repo) {
    T item = repo.findOne(id);
    if (item == null) {
      return null;
    }
    RESTResult res = new RESTResult(item.getDlcmId());
    if (item.getInfo().getStatus().getValue() != StatusValue.Stored
            && item.getInfo().getStatus().getValue() != StatusValue.Completed) {
      res.setMesssage(messageService.get("package.inprogress", new Object[] { id }));
    } else {
      res.setStatus(ActionStatus.NotExecuted);
      if (!item.isAllDataFileReady()) {
        res.setMesssage(messageService.get("package.datafile.notready", new Object[] { id }));
      } else {
        item.setPackageStatus(StatusValue.Reindexing);
        T item2 = repo.save(item);
        if (item2 == null) {
          res.setMesssage(messageService.get("package.cannotupdate", new Object[] { id }));
        } else {
          res.setStatus(ActionStatus.Executed);
          res.setMesssage(messageService.get("package.reindex", new Object[] { id }));

        }
      }
    }
    return res;
  }

  protected boolean uploadData(MultipartFile file, String destination) throws DLCMException {
    try {
      if (file.isEmpty()) {
        throw new DLCMException("Failed to save empty file " + file.getOriginalFilename());
      }
      Files.copy(file.getInputStream(), Paths.get(destination).resolve(file.getOriginalFilename()));
    } catch (IOException e) {
      String message = "Error uploading file";
      log.error(message, e);
      throw new DLCMException(message, e);
    }
    return true;
  }

  protected <T> void addSortLinks(ControllerLinkBuilder clb, RESTCollection<T> collection) {
    collection.add(RESTTool.sort(clb.toUriComponentsBuilder(), RESTName.CREATION)
            .withRel(RESTAction.LASTCREATED));
    collection.add(RESTTool.sort(clb.toUriComponentsBuilder(), RESTName.UPDATE)
            .withRel(RESTAction.LASTUPDATE));
  }

  protected <T> void addPageLinks(ControllerLinkBuilder clb, RESTCollection<T> collection,
          Pageable pageable) {
    if (collection.get_page().hasPrevious()) {
      collection.add(RESTTool.pageLink(clb.toUriComponentsBuilder(),
              pageable.previousOrFirst().getPageSize(), pageable.previousOrFirst().getPageNumber())
              .withRel(RESTAction.PREVIOUS));
    }
    if (collection.get_page().hasNext()) {
      collection.add(RESTTool.pageLink(clb.toUriComponentsBuilder(), pageable.next().getPageSize(),
              pageable.next().getPageNumber()).withRel(RESTAction.NEXT));
    }
  }

}
