package ch.dlcm.controller;

import static org.springframework.hateoas.mvc.ControllerLinkBuilder.linkTo;
import static org.springframework.hateoas.mvc.ControllerLinkBuilder.methodOn;

import java.io.File;
import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.Pageable;
import org.springframework.data.jpa.domain.Specification;
import org.springframework.hateoas.ResourceSupport;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.web.bind.annotation.ModelAttribute;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestMethod;
import org.springframework.web.multipart.MultipartFile;

import ch.dlcm.exception.DLCMException;
import ch.dlcm.model.DataCategory;
import ch.dlcm.model.DataFile;
import ch.dlcm.model.DataFileInterface;
import ch.dlcm.model.StatusHistory;
import ch.dlcm.repository.HistoryRepository;
import ch.dlcm.rest.RESTAction;
import ch.dlcm.rest.RESTCollection;
import ch.dlcm.rest.RESTName;
import ch.dlcm.rest.RESTRepository;
import ch.dlcm.rest.RESTResource;
import ch.dlcm.rest.RESTTool;
import ch.dlcm.tool.FileTool;

public abstract class ResourceController<T extends RESTResource<T>> extends DLCMController implements ResourceControllerInterface {
  private static final Logger log = Logger.getLogger(ResourceController.class);

  @Autowired
  protected RESTRepository<T> itemRepository;

  @Autowired
  HistoryRepository historyRepository;

  public HttpEntity<?> home() {
    return this.listAll(null, null);
  }

  @RequestMapping(method = RequestMethod.GET)
  public HttpEntity<RESTCollection<T>> listAll(@ModelAttribute T search, Pageable pageable) {

    Specification<T> spec = search.getSpecification();
    final Page<T> listItem = itemRepository.findAll(spec, pageable);
    for (final T t : listItem) {
      addLinks(t);
    }
    RESTCollection<T> collection = new RESTCollection<>(listItem, pageable);
    collection.add(linkTo(this.getClass()).withSelfRel());
    collection.add(RESTTool.parentLink((linkTo(this.getClass())).toUriComponentsBuilder()).withRel(RESTAction.MODULE));
    this.addSortLinks(linkTo(this.getClass()), collection);
    this.addPageLinks(linkTo(this.getClass()), collection, pageable);
    addOthersLinks(collection);
    return new ResponseEntity<>(collection, HttpStatus.OK);
  }

  @RequestMapping(method = RequestMethod.POST)
  public HttpEntity<T> create(@RequestBody T t) throws DLCMException {
    T item = itemRepository.save(t);
    if (item != null) {
      this.addLinks(item);
      return new ResponseEntity<>(item, HttpStatus.CREATED);
    }
    return new ResponseEntity<>(HttpStatus.INTERNAL_SERVER_ERROR);
  }

  @RequestMapping(value = "/{id}", method = RequestMethod.GET)
  public HttpEntity<T> list(@PathVariable String id) {
    T t = itemRepository.findOne(id);
    if (t != null) {
      this.addLinks(t);
      return new ResponseEntity<>(t, HttpStatus.OK);
    }
    return new ResponseEntity<>(HttpStatus.NOT_FOUND);
  }

  @RequestMapping(value = "/{id}", method = { RequestMethod.PUT, RequestMethod.PATCH })
  public HttpEntity<T> update(@PathVariable String id, @RequestBody T t2) {
    T t1 = itemRepository.findOne(id);
    if (t1 == null)
      return new ResponseEntity<>(HttpStatus.NOT_FOUND);
    T item = null;
    if (t1.update(t2))
      item = itemRepository.save(t1);
    if (item != null) {
      this.addLinks(item);
      return new ResponseEntity<>(item, HttpStatus.OK);
    }
    return new ResponseEntity<>(HttpStatus.NOT_MODIFIED);
  }

  @RequestMapping(value = "/{id}", method = RequestMethod.DELETE)
  public HttpStatus delete(@PathVariable String id) {
    if (itemRepository.findOne(id) == null)
      return HttpStatus.NOT_FOUND;
    try {
      itemRepository.delete(id);
    } catch (IllegalArgumentException e) {
      log.error("Id is null", e);
      return HttpStatus.GONE;
    }
    return HttpStatus.OK;
  }

  @RequestMapping(value = "/{id}/" + RESTName.HISTORY, method = RequestMethod.GET)
  public HttpEntity<RESTCollection<StatusHistory>> history(@PathVariable String id, Pageable pageable) {
    T item = itemRepository.findOne(id);
    if (item == null) {
      return new ResponseEntity<>(HttpStatus.NOT_FOUND);
    }
    final Page<StatusHistory> listItem = historyRepository.findByDlcmId(id, pageable);
    RESTCollection<StatusHistory> collection = new RESTCollection<>(listItem, pageable);
    collection.add(linkTo(methodOn(((ResourceController) this).getClass()).history(id, pageable)).withSelfRel());
    collection.add(RESTTool.parentLink((linkTo(methodOn(((ResourceController) this).getClass()).history(id, pageable))).toUriComponentsBuilder())
        .withRel(RESTAction.PARENT));
    this.addPageLinks(linkTo(methodOn(((ResourceController) this).getClass()).history(id, pageable)), collection, pageable);
    return new ResponseEntity<>(collection, HttpStatus.OK);
  }

  protected <V extends DataFileInterface> HttpEntity<List<V>> upload(String id, V df, List<MultipartFile> files, List<String> dataCategories,
      List<String> dataTypes, List<String> relLocations) throws DLCMException {
    T item = itemRepository.findOne(id);
    List<V> fileList = new ArrayList<>();
    if (item == null) {
      return new ResponseEntity<>(HttpStatus.NOT_FOUND);
    }
    if (!item.isModifiable()) {
      return new ResponseEntity<>(HttpStatus.NOT_MODIFIED);
    }
    String tempdir = System.getProperty("java.io.tmpdir") + "/" + id;
    if (!FileTool.checkDirectory(new File(tempdir))) {
      return new ResponseEntity<>(HttpStatus.NOT_MODIFIED);
    }
    // Check list
    if (!checkLists(files, dataCategories, dataTypes, relLocations)) {
      log.error("Wrong parameter list size: category or type or folder");
      return new ResponseEntity<>(HttpStatus.BAD_REQUEST);
    }
    int i = 0;
    for (MultipartFile file : files) {
      String tempFile = tempdir;
      if (relLocations != null && !relLocations.isEmpty()) {
        tempFile += "/" + relLocations.get(i);
        if (!FileTool.checkDirectory(new File(tempFile))) {
          return new ResponseEntity<>(HttpStatus.NOT_MODIFIED);
        }
      }
      if (this.uploadData(file, tempFile)) {
        File ulFile = new File(tempFile + "/" + file.getOriginalFilename());
        DataFile f = new DataFile();
        f.setSourceData(ulFile.toURI());
        if (dataCategories != null) {
          f.setDataCategory(DataCategory.valueOf(dataCategories.get(i)));
        }
        if (dataTypes != null) {
          f.setDataType(DataCategory.valueOf(dataTypes.get(i)));
        }
        if (relLocations != null) {
          f.setRelativeLocation(relLocations.get(i));
        }
        df.setDataFile(f);
        item.getDataFiles().add(df);
        fileList.add(df);
      }
      i++;
    }
    T res = itemRepository.save(item);
    if (res == null) {
      return new ResponseEntity<>(HttpStatus.NOT_MODIFIED);
    }
    return new ResponseEntity<>(fileList, HttpStatus.OK);

  }

  private boolean checkLists(List<MultipartFile> files, List<String> dataCategories, List<String> dataTypes, List<String> relLocations) {
    if ((dataCategories != null && !dataCategories.isEmpty() && dataCategories.size() != files.size())
        || (dataTypes != null && !dataTypes.isEmpty() && dataTypes.size() != files.size())
        || (relLocations != null && !relLocations.isEmpty() && relLocations.size() != files.size())) {
      return false;
    }
    return true;
  }

  protected boolean addLinks(T t) {
    t.removeLinks();
    t.addLinks(linkTo(this.getClass()), true, true);
    return true;
  }

  protected <W extends ResourceSupport> boolean addOthersLinks(W w) {
    return false;
  }

}
