package ch.dlcm.service;

import java.net.URI;
import java.net.UnknownHostException;

import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.HttpStatus;
import org.springframework.web.client.HttpClientErrorException;
import org.springframework.web.client.RestTemplate;

import ch.dlcm.config.DLCMConfig;
import ch.dlcm.exception.DLCMResourceNotFoundException;
import ch.dlcm.exception.DLCMRestException;
import ch.dlcm.model.ArchiveMetadata;
import ch.dlcm.model.DataCategory;
import ch.dlcm.model.DataFile;
import ch.dlcm.model.oais.AipDataFile;
import ch.dlcm.model.oais.ArchivalInfoPackage;
import ch.dlcm.model.oais.DisseminationInfoPackage;
import ch.dlcm.model.oais.RepresentationInfo;
import ch.dlcm.model.oais.SipDataFile;
import ch.dlcm.model.oais.SubmissionInfoPackage;
import ch.dlcm.rest.RESTName;
import ch.dlcm.rest.RESTRepository;

public abstract class DLCMService {
  private static final Logger log = Logger.getLogger(DLCMService.class);
  @Autowired
  protected DLCMConfig dlcmConfig;
  @Autowired
  MessageService messageService;

  protected abstract void loadConfig() throws UnknownHostException;

  protected <T> T getOne(RESTRepository<T> repo, String id) {
    T t = null;
    int tries = 0;
    while (t == null && tries < dlcmConfig.getMaxTries()) {
      try {
        Thread.sleep(dlcmConfig.getWaitTime());
      } catch (InterruptedException e) {
        // Some useful details
        // http://www.yegor256.com/2015/10/20/interrupted-exception.html
        log.error(e);
        Thread.currentThread().interrupt();
        return null;
      }
      t = repo.findOne(id);
      tries++;
    }
    return t;
  }

  protected <T> T getResource(String url, Class<T> clazz)
          throws DLCMResourceNotFoundException, DLCMRestException {
    RestTemplate restClt = new RestTemplate();
    try {
      return restClt.getForObject(url, clazz);
    } catch (HttpClientErrorException e) {
      if (e.getStatusCode() == HttpStatus.NOT_FOUND) {
        throw new DLCMResourceNotFoundException(url);
      }
      String message = "Error in getting resource (" + url + ")";
      log.error(message, e);
      throw new DLCMRestException(message, e);
    }
  }

  public String sendSIP(String id, String name) {
    RestTemplate restClt = new RestTemplate();
    // Create SIP
    SubmissionInfoPackage sip = new SubmissionInfoPackage();
    RepresentationInfo info = new RepresentationInfo();
    info.setName(messageService.get("preingest.sip.name", new Object[] { name }));
    info.setDescription(messageService.get("preingest.sip.description"));
    sip.setInfo(info);
    sip.setDepositId(id);
    try {
      // Send SIP object
      SubmissionInfoPackage resSip = restClt.postForObject(
              this.dlcmConfig.getIngestUrl() + "/" + RESTName.SIP, sip,
              SubmissionInfoPackage.class);
      if (resSip.getDlcmId() != null) {
        log.info("SIP created: dlcmId=" + resSip.getDlcmId());
        // Send SIP package
        SipDataFile sipPackage = new SipDataFile();
        DataFile df = new DataFile();
        df.setSourceData(URI.create(this.dlcmConfig.getPreingestUrl() + "/" + RESTName.DEPOSIT + "/"
                + id + "/" + RESTName.SIP));
        df.setDataCategory(DataCategory.Package);
        df.setDataType(DataCategory.InformationPackage);
        sipPackage.setDataFile(df);
        SipDataFile resPackage = restClt
                .postForObject(
                        this.dlcmConfig.getIngestUrl() + "/" + RESTName.SIP + "/"
                                + resSip.getDlcmId() + "/" + RESTName.DATAFILE,
                        sipPackage, SipDataFile.class);
        if (resPackage.getDlcmId() != null) {
          log.info("SIP package created: dlcmId=" + resSip.getDlcmId());
          return resSip.getDlcmId();
        } else {
          log.error("Cannot create the package info for SIP (" + resSip.getDlcmId() + ")");
        }
      } else {
        log.error("Cannot create SIP for deposit (" + id + ")");
      }
    } catch (Exception e) {
      log.error("Error in sending SIP (" + id + ")", e);
    }
    return null;
  }

  public String sendAIP(String id, String name) {
    RestTemplate restClt = new RestTemplate();
    // Create AIP
    ArchivalInfoPackage aip = new ArchivalInfoPackage();
    RepresentationInfo info = new RepresentationInfo();
    info.setName(messageService.get("ingest.aip.name", new Object[] { name }));
    info.setDescription(messageService.get("ingest.aip.description"));
    aip.setInfo(info);
    aip.addSip(id);
    try {
      // Send AIP object
      ArchivalInfoPackage resAip = restClt.postForObject(
              this.dlcmConfig.getArchivalStorageUrl() + "/" + RESTName.AIP, aip,
              ArchivalInfoPackage.class);
      if (resAip.getDlcmId() != null) {
        log.info("AIP created: dlcmId=" + resAip.getDlcmId());
        // Send AIP package
        AipDataFile aipPackage = new AipDataFile();
        DataFile df = new DataFile();
        df.setSourceData(URI.create(this.dlcmConfig.getIngestUrl() + "/" + RESTName.SIP + "/" + id
                + "/" + RESTName.AIP));
        df.setDataCategory(DataCategory.Package);
        df.setDataType(DataCategory.InformationPackage);
        aipPackage.setDataFile(df);
        AipDataFile resPackage = restClt.postForObject(
                this.dlcmConfig.getArchivalStorageUrl() + "/" + RESTName.AIP + "/"
                        + resAip.getDlcmId() + "/" + RESTName.DATAFILE,
                aipPackage, AipDataFile.class);
        if (resPackage.getDlcmId() != null) {
          log.info("AIP package created: dlcmId=" + resAip.getDlcmId());
          return resAip.getDlcmId();
        } else {
          log.error("Cannot create the package info for AIP (" + resAip.getDlcmId() + ")");
        }
      } else {
        log.error("Cannot create AIP for deposit (" + id + ")");
      }
    } catch (Exception e) {
      log.error("Error in sending AIP (" + id + ")", e);
    }
    return null;
  }

  public DisseminationInfoPackage getDIP(String id)
          throws DLCMRestException, DLCMResourceNotFoundException {
    String url = this.dlcmConfig.getAccessUrl() + "/" + RESTName.DIP + "/" + id;
    return getResource(url, DisseminationInfoPackage.class);
  }

  public ArchivalInfoPackage getAIP(String id)
          throws DLCMRestException, DLCMResourceNotFoundException {
    String url = this.dlcmConfig.getArchivalStorageUrl() + "/" + RESTName.AIP + "/" + id;
    return getResource(url, ArchivalInfoPackage.class);
  }

  public ArchiveMetadata getMetatadata(String id)
          throws DLCMRestException, DLCMResourceNotFoundException {
    String url = this.dlcmConfig.getDataMgmtUrl() + "/" + RESTName.DI + "/" + id;
    return getResource(url, ArchiveMetadata.class);
  }

  public String getMetatadataList(String query, int page, int size) {
    RestTemplate restClt = new RestTemplate();
    return restClt.getForObject(this.dlcmConfig.getDataMgmtUrl() + "/" + RESTName.DI + "/"
            + RESTName.SEARCH + "?query=" + query + "&page=" + page + "&size=" + size,
            String.class);
  }

  public boolean indexMetadata(String id, ArchiveMetadata md) {
    RestTemplate restClt = new RestTemplate();
    try {
      ArchiveMetadata res = restClt.postForObject(
              this.dlcmConfig.getDataMgmtUrl() + "/" + RESTName.DI, md, ArchiveMetadata.class);
      if (res == null) {
        return false;
      }

    } catch (HttpClientErrorException e) {
      log.error("Error in indexing metadata (" + id + ")", e);
      return false;
    }
    return true;
  }

  protected boolean checkMetadata(String id) {
    RestTemplate restClt = new RestTemplate();
    try {
      ArchiveMetadata res = restClt.getForObject(
              this.dlcmConfig.getDataMgmtUrl() + "/" + RESTName.DI + "/" + id,
              ArchiveMetadata.class);
      if (res == null) {
        return false;
      }
    } catch (HttpClientErrorException e) {
      if (e.getStatusCode() == HttpStatus.NOT_FOUND) {
        return false;
      }
      log.error("Error in getting metadata (" + id + ")", e);
      return false;
    }
    return true;
  }

  public boolean deleteIndex(String id) {
    RestTemplate restClt = new RestTemplate();
    try {
      restClt.delete(this.dlcmConfig.getDataMgmtUrl() + "/" + RESTName.DI + "/" + id);
    } catch (HttpClientErrorException e) {
      log.error("Error in deleting metadata index (" + id + ")", e);
      return false;
    }
    return true;
  }

  public String getTenant() {
    return this.dlcmConfig.getIndexName();
  }

}
