package ch.dlcm.service;

import java.io.IOException;
import java.net.URISyntaxException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.security.NoSuchAlgorithmException;
import java.util.List;

import javax.annotation.PostConstruct;

import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.jms.annotation.JmsListener;
import org.springframework.stereotype.Component;

import ch.dlcm.config.ContextHelper;
import ch.dlcm.message.AipDownloadMessage;
import ch.dlcm.message.AipMessage;
import ch.dlcm.message.DataFileMessage;
import ch.dlcm.message.DipMessage;
import ch.dlcm.message.SipMessage;
import ch.dlcm.model.ComplianceLevel;
import ch.dlcm.model.DataFileChecksum;
import ch.dlcm.model.DataFileChecksum.ChecksumAlgorithm;
import ch.dlcm.model.DataFileChecksum.ChecksumOrigin;
import ch.dlcm.model.DataFileChecksum.ChecksumType;
import ch.dlcm.model.DataFileInterface;
import ch.dlcm.model.DepositDataFile;
import ch.dlcm.model.Status.StatusValue;
import ch.dlcm.model.oais.AipDataFile;
import ch.dlcm.model.oais.DipDataFile;
import ch.dlcm.model.oais.SipDataFile;
import ch.dlcm.repository.AipDataFileRepository;
import ch.dlcm.repository.DepositDataFileRepository;
import ch.dlcm.repository.DipDataFileRepository;
import ch.dlcm.repository.SipDataFileRepository;
import ch.dlcm.rest.RESTRepository;
import ch.dlcm.tool.FileTool;
import org.springframework.transaction.annotation.Transactional;

@Component
public class DataFileService extends DLCMService {
	private static final Logger log = Logger.getLogger(DataFileService.class);

	private String[] checksumList;

	@Autowired
	protected DepositDataFileRepository depositDataFileRepository;
	@Autowired
	protected SipDataFileRepository sipDataFileRepository;
	@Autowired
	protected AipDataFileRepository aipDataFileRepository;
	@Autowired
	protected DipDataFileRepository dipDataFileRepository;
	@Autowired
	protected MetadataService metadataService;
	@Autowired
	protected FileFormatService fileFormatService;
	
	@PostConstruct
	@Override
	protected void loadConfig() {
		checksumList = this.dlcmConfig.getChecksums();
	}

	protected boolean checkPackageDirectory(String location, String packageId) {
		Path p = Paths.get(location, packageId);
		if (!p.toFile().exists()) {
			try {
				Files.createDirectories(p);
			} catch (IOException e) {
				log.error("Error checking package directory", e);
				return false;
			}
		}
		return true;
	}

	@JmsListener(destination = "${dlcm.queue.datafile}")
	@Transactional
	public void processDataFile(DataFileMessage dataFile) {
		log.info("Reading Data File message " + dataFile.getDlcmId() + " ("
				+ dataFile.getDataFileClass().getSimpleName() + ")");
		if (dataFile.getDataFileClass().equals(DepositDataFile.class)) {
			this.processDataFile(this.dlcmConfig.getPreingestPreparationLocation(), this.depositDataFileRepository,
					dataFile.getDlcmId());
		} else if (dataFile.getDataFileClass().equals(SipDataFile.class)) {
			SipDataFile sip = this.processDataFile(this.dlcmConfig.getIngestLocation(), this.sipDataFileRepository,
					dataFile.getDlcmId());
			if (sip.getDataFile().getStatus().getValue() == StatusValue.Ready) {
				ContextHelper.getPublisher().publishEvent(new SipMessage(sip.getPackageId()));
			}
		} else if (dataFile.getDataFileClass().equals(AipDataFile.class)) {
			AipDataFile aip = this.processDataFile(this.dlcmConfig.getArchivalLocation(), this.aipDataFileRepository,
					dataFile.getDlcmId());
			if (aip.getDataFile().getStatus().getValue() == StatusValue.Ready) {
				ContextHelper.getPublisher().publishEvent(new AipMessage(aip.getPackageId()));
				ContextHelper.getPublisher().publishEvent(new AipDownloadMessage(aip.getPackageId())); //TODO Optimize for knowing if AIP creation or download
			}
		} else if (dataFile.getDataFileClass().equals(DipDataFile.class)) {
			DipDataFile dip = this.processDataFile(this.dlcmConfig.getAccessLocation(), this.dipDataFileRepository,
					dataFile.getDlcmId());
			if (dip.getDataFile().getStatus().getValue() == StatusValue.Ready) {
				ContextHelper.getPublisher().publishEvent(new DipMessage(dip.getPackageId()));
			}
		}
	}

	private <T extends DataFileInterface> T processDataFile(String location, RESTRepository<T> repo, String id) {
		T df = this.getOne(repo, id);
		if (df == null) {
			log.error("Cannot find Data File (" + id + ")");
			return null;
		}

		while (df.getDataFile().getStatus().getValue() != StatusValue.InError
				&& df.getDataFile().getStatus().getValue() != StatusValue.Ready) {
			switch (df.getDataFile().getStatus().getValue()) {
			case Received:
				if (df.getDataFile().checkSource()) {
					df.setFileStatus(StatusValue.ToProcess);
				} else {
					inError(df, messageService.get("datafile.error.check"));
				}
				break;
			case ToProcess:
				if (checkPackageDirectory(location, df.getPackageId())) {
					if (df.getDataFile().getData(location + "/" + df.getPackageId())) {
						df.setFileStatus(StatusValue.Processed);
					} else {
						inError(df, messageService.get("datafile.error.download"));
					}
				} else {
					log.error("Problem with package directory (" + df.getPackageId() + ") in " + location);
				}
				break;
			case Processed:
				df.setFileStatus(StatusValue.Ready);
				try {
					df.getDataFile().setComplianceLevel(ComplianceLevel.NO_COMPLIANCE);
					if (!verifyChecksum(df)) {
						inError(df, messageService.get("datafile.error.checksum"));
					} else {
						df.getDataFile().setComplianceLevel(ComplianceLevel.WEAK_COMPLIANCE);
						if (!verifyDataFile(df)) {
							inError(df, messageService.get("datafile.error.technicalvalidation"));
						}
					}
				} catch (NoSuchAlgorithmException | IOException | URISyntaxException e) {
					inError(df, messageService.get("datafile.error", new Object[] { e.getMessage() }));
					log.error(e.getMessage(), e);
				}

				break;
			default:
				break;

			}
			df = repo.save(df);
		}
		log.info("Data file processed: type=" + df.getClass().getSimpleName() + "dlcmId=" + df.getDlcmId() + " status="
				+ df.getDataFile().getStatus().getValue().toString());
		return df;
	}

	private <T extends DataFileInterface> boolean verifyDataFile(T df) throws URISyntaxException, IOException {
		switch (df.getDataFile().getDataCategory()) {
		case Package:
			switch (df.getDataFile().getDataType()) {
			case Metadata:
				return metadataService.checkDlcmMetadata(df);
			case ExternalMetadata:
				return metadataService.checkExternalMetadata(df);
			default:
				break;
			}
			return true;
		case Primary:
		case Secondary:
			return fileFormatService.checkFormat(df);
		default:
			break;
		}
		return true;
	}

	private <T extends DataFileInterface> void inError(T df, String msg) {
		df.setErrorStatus(msg);
		log.error(msg + " for " + df.getDataFile().getSourceData());
	}

	private <T extends DataFileInterface> boolean verifyChecksum(T df)
			throws NoSuchAlgorithmException, IOException, URISyntaxException {
		// Checksum computation
		if (this.checksumList.length == 0) {
			return true;
		}

		// Compute all checksums
		String[] checksums = FileTool.computeChecksum(df.getDataFile().getFinalData(), this.checksumList);
		List<DataFileChecksum> existingChecksums = df.getChecksum();

		// Check if exist => compare else add it
		for (int i = 0; i < this.checksumList.length; i++) {
			if (existingChecksums == null || existingChecksums.isEmpty()) {
				df.getChecksum().add(new DataFileChecksum(ChecksumAlgorithm.valueOf(this.checksumList[i]),
						ChecksumType.Complete, ChecksumOrigin.DLCM, checksums[i]));
			} else {
				boolean found = false;
				for (DataFileChecksum checksum : existingChecksums) {
					if (checksum.getChecksumAlgo() == ChecksumAlgorithm.valueOf(this.checksumList[i])
							&& checksum.getChecksumType() == ChecksumType.Complete) {
						found = true;
						if (!checksum.getChecksum().equals(checksums[i])) {
							return false;
						}
					}
				}
				if (!found) {
					df.getChecksum().add(new DataFileChecksum(ChecksumAlgorithm.valueOf(this.checksumList[i]),
							ChecksumType.Complete, ChecksumOrigin.DLCM, checksums[i]));
				}
			}
		}
		return true;
	}

}
