package ch.dlcm.indexing.elasticsearch;

import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.List;

import javax.annotation.PostConstruct;

import org.apache.log4j.Logger;
import org.elasticsearch.action.delete.DeleteResponse;
import org.elasticsearch.action.index.IndexResponse;
import org.elasticsearch.action.search.SearchResponse;
import org.elasticsearch.client.transport.TransportClient;
import org.elasticsearch.common.settings.Settings;
import org.elasticsearch.common.transport.InetSocketTransportAddress;
import org.elasticsearch.index.query.QueryBuilders;
import org.elasticsearch.rest.RestStatus;
import org.elasticsearch.search.SearchHit;
import org.elasticsearch.search.SearchHits;
import org.elasticsearch.transport.client.PreBuiltTransportClient;
import org.springframework.context.annotation.Profile;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.PageImpl;
import org.springframework.data.domain.Pageable;
import org.springframework.stereotype.Service;

import ch.dlcm.exception.DLCMException;
import ch.dlcm.indexing.IndexingService;
import ch.dlcm.model.ArchiveMetadata;
import ch.dlcm.rest.RESTPage;
import ch.dlcm.service.DLCMService;

@Profile("dat-elasticsearch")
@Service
public class ElasticsearchService extends DLCMService implements IndexingService {
	private static final Logger log = Logger.getLogger(ElasticsearchService.class);
	
	public static final int TIMEOUT= 60000;

	private TransportClient esClt;
  private String indexName;

	@PostConstruct
	@Override
	protected void loadConfig() throws UnknownHostException {
		esClt = new PreBuiltTransportClient(Settings.EMPTY).addTransportAddress(new InetSocketTransportAddress(
				InetAddress.getByName(this.dlcmConfig.getIndexingHost()), this.dlcmConfig.getIndexingPort()));
		indexName = this.dlcmConfig.getIndexName();
	}

	private boolean checkIndex(String index, String type) {
	  if (!index.startsWith(this.indexName)) {
	    return false;
	  }
		esClt.admin().indices().prepareGetSettings(index).get();
		return true;
	}

	@Override
	public Page<ArchiveMetadata> findAll(Pageable pageable) {
		SearchResponse res = esClt.prepareSearch(indexName)
								  .setSize(pageable.getPageSize())
								  .setFrom(pageable.getPageSize()*pageable.getPageNumber())
								  .get();
		return getPage(res.getHits(), pageable);
	}
	
	
	private Page<ArchiveMetadata> getPage(SearchHits searchHits, Pageable pageable) {
		List<ArchiveMetadata> list = new ArrayList<>();
		for(SearchHit h : searchHits.getHits()){
			list.add(this.convertHit(h));
		}
		return new PageImpl<>(list, pageable, searchHits.getTotalHits());

	}

	private ArchiveMetadata convertHit(SearchHit h) {
		ArchiveMetadata am = new ArchiveMetadata();
		am.setDlcmId(h.getId());
		am.setIndex(h.getIndex());
		am.setType(h.getType());
		am.setMetadata(h.getSourceAsString());
		return am;
	}
	
	@Override
	public Page<ArchiveMetadata> search(String query, Pageable pageable) {
		SearchResponse res = esClt.prepareSearch(indexName)
								  .setSize(pageable.getPageSize())
								  .setFrom(pageable.getPageSize()*pageable.getPageNumber())
								  .setQuery(QueryBuilders.queryStringQuery(query))
								  .get();
		return getPage(res.getHits(), pageable);
	}

	@Override
	public ArchiveMetadata save(ArchiveMetadata arc) throws DLCMException {
		if (!arc.check()) {
			log.error("Missing fields in " + arc.toString());
			throw new DLCMException("Missing fields in " + arc.toString());
		}
		if (!this.checkIndex(arc.getIndex(), arc.getType())) {
			log.error("Problem in index configuration (" + arc.getIndex() + "," + arc.getType() + ")");
			throw new DLCMException("Problem in index configuration (" + arc.getIndex() + "," + arc.getType() + ")");
		}
		IndexResponse res = esClt.prepareIndex(arc.getIndex(), arc.getType(), arc.getDlcmId()).setSource(arc.getMetadata()).get();
		if (res.status() == RestStatus.CREATED){
			return arc;
		}
		return null;
	}

	@Override
	public boolean delete(ArchiveMetadata a) {
		DeleteResponse res = esClt.prepareDelete(a.getIndex(), a.getType(), a.getDlcmId()).get();
		if (res.status() == RestStatus.OK){
			return true;
		}
		return false;
	}

	@Override
	public ArchiveMetadata findOne(String id)  {
		SearchResponse res = esClt.prepareSearch(indexName).setQuery(QueryBuilders.termQuery("_id", id)).get();
		if (res.getHits().getHits().length != 1){
			return null;
		}
		return convertHit(res.getHits().getHits()[0]);
	}

	@Override
	public List<ArchiveMetadata> findAll() {
		List<ArchiveMetadata> list = new ArrayList<>();
		SearchResponse res = esClt.prepareSearch(indexName)
								  .setSize(RESTPage.DEFAULT_SIZE_PAGE)
			  					  .setFrom(0)
			  					  .get();
		for(SearchHit h : res.getHits().getHits()){
			list.add(this.convertHit(h));
		}
		return list;
	}

}
