package ch.dlcm.settings.elasticsearch;

import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.List;

import javax.annotation.PostConstruct;

import org.apache.log4j.Logger;
import org.elasticsearch.action.admin.indices.create.CreateIndexResponse;
import org.elasticsearch.action.admin.indices.delete.DeleteIndexResponse;
import org.elasticsearch.action.admin.indices.settings.get.GetSettingsResponse;
import org.elasticsearch.client.transport.TransportClient;
import org.elasticsearch.common.settings.Settings;
import org.elasticsearch.common.transport.InetSocketTransportAddress;
import org.elasticsearch.index.IndexNotFoundException;
import org.elasticsearch.transport.client.PreBuiltTransportClient;
import org.springframework.context.annotation.Profile;
import org.springframework.stereotype.Service;

import com.carrotsearch.hppc.cursors.ObjectObjectCursor;

import ch.dlcm.exception.DLCMInitialDataException;
import ch.dlcm.model.ArchiveIndex;
import ch.dlcm.service.DLCMService;
import ch.dlcm.settings.SettingsService;

@Profile("dat-elasticsearch")
@Service
public class ElasticsearchSettingsService extends DLCMService implements SettingsService {
  private static final Logger log = Logger.getLogger(ElasticsearchSettingsService.class);

  private TransportClient esClt;
  private String indexPrefix;

  private static final String MAPPING_NAME = "_default_";
  private static final String MAPPING_DEFAULT = "{\"_all\": {\"enabled\": false}}";

  @PostConstruct
  @Override
  protected void loadConfig() throws UnknownHostException {
    esClt = new PreBuiltTransportClient(Settings.EMPTY).addTransportAddress(
            new InetSocketTransportAddress(InetAddress.getByName(this.dlcmConfig.getIndexingHost()),
                    this.dlcmConfig.getIndexingPort()));
    indexPrefix = this.dlcmConfig.getIndexName();
  }

  protected boolean checkIndexName(String indexName) {
    return indexName.startsWith(this.indexPrefix);
  }

  @Override
  public List<ArchiveIndex> findAll() {
    List<ArchiveIndex> list = new ArrayList<>();
    GetSettingsResponse res = esClt.admin().indices().prepareGetSettings().get();
    for (ObjectObjectCursor<String, Settings> cursor : res.getIndexToSettings()) {
      if (this.checkIndexName(cursor.key)) {
        ArchiveIndex ai = new ArchiveIndex();
        ai.setDlcmId(cursor.key);
        ai.setSettings(cursor.value.getAsStructuredMap());
        list.add(ai);
      }
    }
    return list;
  }

  @Override
  public ArchiveIndex save(ArchiveIndex index) {
    if (index.get_settings().isEmpty()) {
      index.setSettings(MAPPING_DEFAULT);
    }
    if (!this.checkIndexName(index.getDlcmId())){
      log.error("Wrong index name (" + index.getDlcmId() + ")");
      return null;
    }
    CreateIndexResponse res = esClt.admin().indices().prepareCreate(index.getDlcmId())
            .addMapping(MAPPING_NAME, index.get_settings()).get();
    if (res.isAcknowledged()) {
      return index;
    }
    log.error("Cannot create index (" + index + ")");
    return null;

  }

  @Override
  public boolean delete(ArchiveIndex ai) {
    if (!this.checkIndexName(ai.getDlcmId())){
      return false;
    }
    DeleteIndexResponse res = esClt.admin().indices().prepareDelete(ai.getDlcmId()).get();
    if (res.isAcknowledged()) {
      return true;
    }
    return false;
  }

  @Override
  public ArchiveIndex findOne(String index) {
    if (!this.checkIndexName(index)){
      log.error("Wrong index name (" + index + ")");
      return null;
    }
    ArchiveIndex ai = new ArchiveIndex();
    ai.setDlcmId(index);
    GetSettingsResponse resS=null;
    try {
      resS = esClt.admin().indices().prepareGetSettings(index).get();
    }catch(IndexNotFoundException e) {
      log.error("Index '"+index+"' does not exist",e);
      return null;
    }
    if (resS.getIndexToSettings().values().isEmpty()) {
      return null;
    }
    ai.setSettings(resS.getIndexToSettings().get(index).getAsStructuredMap());
    return ai;
  }

  @Override
  public void init() throws Exception {
    if (this.findOne(this.indexPrefix) == null) {
      ArchiveIndex index = new ArchiveIndex();
      index.setDlcmId(indexPrefix);
      if (this.save(index) != null) {
        log.info("Index '" + this.indexPrefix + "' created");
      }else {
        String message = "Cannot create '" + this.indexPrefix + "' index";
        log.error(message);
        throw new DLCMInitialDataException(message);
      }
    }
  }

}
