package ch.dlcm.model;

import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Objects;

import javax.persistence.Embeddable;
import javax.persistence.Embedded;
import javax.persistence.EnumType;
import javax.persistence.Enumerated;
import javax.validation.constraints.NotNull;

import ch.dlcm.exception.DLCMException;
import ch.dlcm.tool.RestClientTool;
import org.apache.commons.validator.routines.UrlValidator;
import org.apache.log4j.Logger;
import org.springframework.web.client.ResponseExtractor;

import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonProperty;

import ch.dlcm.model.Status.StatusType;
import ch.dlcm.model.Status.StatusValue;
import ch.dlcm.tool.FileTool;
import ch.dlcm.tool.StringTool;

@Embeddable
public class DataFile implements JPAEmbeddableEntityInterface<DataFile> {
  private static final Logger log = Logger.getLogger(DataFile.class);

  @NotNull
  protected URI sourceData;

  private String relativeLocation;

  private URI finalData;

  private String contentType;

  private long size;

  // TODO Refers to issue DLCM-222
  private String format;

  @Embedded
  private Status status = new Status();

  @JsonProperty(required = true)
  @Enumerated(EnumType.STRING)
  private DataCategory dataCategory;

  @JsonProperty(required = true)
  @Enumerated(EnumType.STRING)
  private DataCategory dataType;

  @Enumerated(EnumType.ORDINAL)
  private ComplianceLevel complianceLevel;

  @Override
  public boolean equals(final Object other) {
    if (other == null) {
      return false;
    }
    if (!getClass().equals(other.getClass())) {
      return false;
    }
    DataFile castOther = (DataFile) other;
    return Objects.equals(getSourceData(), castOther.getSourceData()) && Objects.equals(getRelativeLocation(), castOther.getRelativeLocation())
        && Objects.equals(getFinalData(), castOther.getFinalData()) && Objects.equals(getContentType(), castOther.getContentType())
        && Objects.equals(getSize(), castOther.getSize()) && Objects.equals(getStatus(), castOther.getStatus())
        && Objects.equals(getDataCategory(), castOther.getDataCategory()) && Objects.equals(getDataType(), castOther.getDataType())
        && Objects.equals(getComplianceLevel(), castOther.getComplianceLevel());
  }

  @Override
  public int hashCode() {
    int result = super.hashCode();
    result = 31 * result + (getSourceData() != null ? getSourceData().hashCode() : 0);
    result = 31 * result + (getRelativeLocation() != null ? getRelativeLocation().hashCode() : 0);
    result = 31 * result + (getFinalData() != null ? getFinalData().hashCode() : 0);
    result = 31 * result + (getContentType() != null ? getContentType().hashCode() : 0);
    result = 31 * result + (new Integer((int) (long) getSize()).hashCode());
    result = 31 * result + (getStatus() != null ? getStatus().hashCode() : 0);
    result = 31 * result + (getDataCategory() != null ? getDataCategory().hashCode() : 0);
    result = 31 * result + (getDataType() != null ? getDataType().hashCode() : 0);
    result = 31 * result + (getComplianceLevel() != null ? getComplianceLevel().hashCode() : 0);
    return result;
  }

  public ComplianceLevel getComplianceLevel() {
    return complianceLevel;
  }

  public void setComplianceLevel(ComplianceLevel complianceLevel) {
    this.complianceLevel = complianceLevel;
  }

  public DataFile() {
    this.status.setType(StatusType.DataFile);
  }

  public void init() {
    if (status.getValue() == null) {
      status.setValue(StatusValue.Received);
    }
    if (relativeLocation == null) {
      relativeLocation = "";
    } else {
      checkRelativeLocation();
    }
    if (dataCategory == null) {
      dataCategory = DataCategory.Others;
    }
    if (dataType == null) {
      dataType = DataCategory.Unknown;
    }
    complianceLevel = ComplianceLevel.NOT_ASSESSED;
  }

  private void checkRelativeLocation() {
    if (!relativeLocation.startsWith("/")) {
      setRelativeLocation("/" + relativeLocation);
    }
  }

  /* **************************************
  * Getters and Setters
  * ************************************** */

  public String getFormat() {
    return format;
  }

  public void setFormat(String format) {
    this.format = format;
  }

  public URI getSourceData() {
    return sourceData;
  }

  public String getFileName() {
    return sourceData.toString().substring(sourceData.toString().lastIndexOf('/') + 1);
  }

  public void setSourceData(URI sourceData) {
    this.sourceData = sourceData;
  }

  public String getContentType() {
    return contentType;
  }

  public void setContentType(String contentType) {
    this.contentType = contentType;
  }

  public Status getStatus() {
    return status;
  }

  public void setStatus(Status status) {
    this.status = status;
  }

  public URI getFinalData() {
    return finalData;
  }

  public void setFinalData(URI finalData) {
    this.finalData = finalData;
  }

  public DataCategory getDataCategory() {
    return dataCategory;
  }

  public void setDataCategory(DataCategory dataCat) {
    this.dataCategory = dataCat;
  }

  public DataCategory getDataType() {
    return dataType;
  }

  public void setDataType(DataCategory dataType) {
    this.dataType = dataType;
  }

  public long getSize() {
    return size;
  }

  public String getRelativeLocation() {
    return relativeLocation;
  }

  public void setRelativeLocation(String relativeLocation) {
    this.relativeLocation = relativeLocation;
  }

  public String getSmartSize() {
    return StringTool.formatSmartSize(size);
  }

  public void setSize(long size) {
    this.size = size;
  }

  /* *************************************
  *  Functional methods
  * **************************************/

  public boolean checkSource() {
    String[] schemes = { "http", "https", "ftp", "ftps", "file" };
    UrlValidator urlValidator = new UrlValidator(schemes, UrlValidator.ALLOW_LOCAL_URLS);
    return urlValidator.isValid(sourceData.toString());
  }

  @JsonIgnore
  public boolean getData(String preparationPath) {
    try {
      URI srcUri = this.getSourceData();
      File destData = this.defineTargetFile(preparationPath);
      switch (srcUri.getScheme()) {
        case "http":
        case "https":
          // Streams the response instead of loading it all in memory
          ResponseExtractor<Void> responseExtractor = response -> {
            // Here I write the response to a file but do what you like
            //List<String> tt = response.getHeaders().getValuesAsList("attachment");
            Path path = Paths.get(destData.getAbsolutePath());
            Files.copy(response.getBody(), path);
            return null;
          };
          RestClientTool.downloadContent(this.getSourceData(), responseExtractor);
          return this.setFinalDataDetail(destData.toURI());
        case "file":
          File src = new File(srcUri);
          if (FileTool.moveFile(src, destData)) {
            return this.setFinalDataDetail(destData.toURI());
          }
          break;
        case "ftp":
        case "ftps":
          // TODO ftp client
          break;
      }
    } catch (IOException | DLCMException e) {
      log.error(e);
    }
    return false;
  }

  private File defineTargetFile(String preparationPath) throws DLCMException{
    String target = preparationPath;
    switch (this.dataCategory) {
      case Others:
        target += Package.MISC.getName() + getRelativeLocation() + "/" + this.getFileName();
        break;
      case Package:
        switch (this.dataType) {
          case ExternalMetadata:
            target += Package.METADATA.getName() + "/" + this.getFileName();
            break;
          case Metadata:
            target += "/" + Package.METADATAFILE.getName();
            break;
          case InformationPackage:
            target += "/" + this.getFileName();
            break;
          default:
            target += Package.MISC.getName() + getRelativeLocation() + "/" + this.getFileName();
            break;
        }
        break;
      case Primary:
        target += Package.DATA.getName() + getRelativeLocation() + "/" + this.getFileName();
        break;
      case Documentation:
        target += Package.DOC.getName() + getRelativeLocation() + "/" + this.getFileName();
        break;
      case Software:
        target += Package.SOFTWARE.getName() + getRelativeLocation() + "/" + this.getFileName();
        break;
      default:
        target += Package.MISC.getName() + getRelativeLocation() + "/" + this.getFileName();
        break;
    }
    File t = new File(target);
    if (FileTool.checkDirectory(t.getParentFile())) {
      return t;
    } else {
      throw new DLCMException("Impossible to create directory");
    }
  }

  @JsonIgnore
  private boolean setFinalDataDetail(URI dfUri) {
    this.setFinalData(dfUri);
    try {
      if (this.getContentType() == null) {
        this.setContentType(FileTool.getContentType(this.getFinalData()).toString());
      }
      long size = FileTool.getSize(this.getFinalData());
      if (size == 0) {
        log.error("Cannot get file size for " + this.getFinalData());
        return false;
      }
      if (this.getSize() == 0) {
        this.setSize(size);
      } else {
        if (this.getSize() != size) {
          log.error("Wrong file size for " + this.getFinalData());
          return false;
        }
      }
    } catch (URISyntaxException e) {
      log.error("Wrong URI for " + this.getFinalData(), e);
      return false;
    }
    return true;
  }

  @Override
  public boolean update(DataFile df) {
    boolean changed = false;
    if (StringTool.toUpdate(this.getSourceData().toString(), df.getSourceData().toString())) {
      this.setSourceData(df.getSourceData());
      this.status.setValue(StatusValue.Received);
      changed = true;
    }
    if (df.getStatus() != null && df.getStatus().getValue() != null
        && (this.getStatus().getValue() == StatusValue.Received || this.getStatus().getValue() == StatusValue.InError)) {
      this.setStatus(df.getStatus());
      changed = true;
    }

    if (StringTool.toUpdate(this.getRelativeLocation(), df.getRelativeLocation())) {
      this.setRelativeLocation(df.getRelativeLocation());
      this.checkRelativeLocation();
      this.checkStatus();
      changed = true;
    }
    if (df.getDataCategory() != null && this.getDataCategory() != df.getDataCategory()) {
      this.setDataCategory(df.getDataCategory());
      this.checkStatus();
      changed = true;
    }
    if (df.getDataType() != null && this.getDataType() != df.getDataType()) {
      this.setDataType(df.getDataType());
      this.checkStatus();
      changed = true;
    }
    return changed;
  }

  private void checkStatus() {
    if (this.getStatus().getValue() == StatusValue.Processed) {
      this.setSourceData(this.getFinalData());
      this.getStatus().setValue(StatusValue.Received);
    }
  }

  public boolean isAvailable() {
    return this.getStatus().getValue() == StatusValue.Processed || this.getStatus().getValue() == StatusValue.Ready;
  }

}