package ch.dlcm.model;

import java.net.URI;
import java.time.OffsetDateTime;
import java.util.ArrayList;
import java.util.List;

import javax.persistence.CascadeType;
import javax.persistence.Convert;
import javax.persistence.Embedded;
import javax.persistence.Entity;
import javax.persistence.EntityListeners;
import javax.persistence.EnumType;
import javax.persistence.Enumerated;
import javax.persistence.JoinColumn;
import javax.persistence.JoinTable;
import javax.persistence.ManyToMany;
import javax.persistence.OneToMany;
import javax.persistence.OneToOne;
import javax.persistence.PrePersist;
import javax.persistence.PreUpdate;
import javax.persistence.Transient;
import javax.persistence.UniqueConstraint;
import javax.validation.constraints.NotNull;
import javax.validation.constraints.Size;

import org.springframework.data.jpa.domain.Specification;
import org.springframework.hateoas.mvc.ControllerLinkBuilder;

import com.fasterxml.jackson.annotation.JsonIdentityInfo;
import com.fasterxml.jackson.annotation.JsonIdentityReference;
import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonProperty;
import com.fasterxml.jackson.annotation.ObjectIdGenerators;

import ch.dlcm.converter.KeywordsConverter;
import ch.dlcm.criteria.DepositSpecification;
import ch.dlcm.exception.DLCMException;
import ch.dlcm.listener.HistoryListener;
import ch.dlcm.listener.PackageListener;
import ch.dlcm.model.Status.StatusType;
import ch.dlcm.model.Status.StatusValue;
import ch.dlcm.model.validator.AssertThatAnotherFieldHasValue;
import ch.dlcm.rest.RESTName;
import ch.dlcm.rest.RESTResource;
import ch.dlcm.rest.RESTSubResourceInterface;
import ch.dlcm.tool.StringTool;

@Entity
@EntityListeners({HistoryListener.class, PackageListener.class})
@AssertThatAnotherFieldHasValue.List({
		@AssertThatAnotherFieldHasValue(fieldName = "agreement", fieldValue = "true", dependFieldName = "submissionAgreement"),
		@AssertThatAnotherFieldHasValue(fieldName = "access", fieldValue = "Public", dependFieldName = "license")
})
public class Deposit extends RESTResource<Deposit> implements RESTSubResourceInterface{

	@NotNull
	@Size(min = 1, max = 256)
	private String title;

	@Embedded
	private Status status = new Status();

	@JsonIgnore
	@ManyToMany
	@JoinTable(name = "depositContributors", 
			   joinColumns = { @JoinColumn(name = "parentId") }, 
			   inverseJoinColumns = { @JoinColumn(name = "childId") }, 
			   uniqueConstraints = { @UniqueConstraint(columnNames = { "parentId", "childId" }) })
	private List<Person> contributors = new ArrayList<>();

	@OneToMany(cascade = CascadeType.ALL, orphanRemoval = true)
	@JoinColumn(name = "packageId", referencedColumnName = "dlcmId")
	@JsonIgnore
	private List<DepositDataFile> dataFiles = new ArrayList<>();

	private String sipId;
	private String preparationId;
	private String doi;
	private Boolean agreement;
	private Integer year;

	@Enumerated(EnumType.STRING)
 	private Access access;

	@OneToOne (targetEntity = License.class)
	@JoinColumn(name = "licenseId", referencedColumnName = "dlcmId")
	@JsonIdentityInfo(generator=ObjectIdGenerators.PropertyGenerator.class, property="dlcmId")
	@JsonIdentityReference(alwaysAsId=true)
	private License license;

	@OneToOne (targetEntity = Language.class)
	@JoinColumn(name = "languageId", referencedColumnName = "dlcmId")
	private Language language;

	@Size(max = 1024)
	private String description;

	@Convert(converter = KeywordsConverter.class)
	private List<String> keywords;

	@OneToOne (targetEntity = OrganizationalUnit.class)
	@JoinColumn(name = "organisationId", referencedColumnName = "dlcmId")
	private String organisationId;

    @OneToOne (targetEntity = SubmissionAgreement.class)
    @JoinColumn(name = "submissionAgreementId", referencedColumnName = "dlcmId")
    private SubmissionAgreement submissionAgreement;

	@JsonIgnore
	private String sip;

	public Deposit() {
		this.status.setType(StatusType.Deposit);
	}

	@PrePersist
	@PreUpdate
	@Override
	public void init() {
		if (status.getValue() == null) {
			status.setValue(StatusValue.InProgress);
		}
		if (year == null) {
			year = getUpdateTime().getYear();
		}
		if (access == null){
			access = Access.Restricted;
		}
		if (agreement == null){
			agreement = false;
		}
	}

	public String getTitle() {
		return title;
	}

	public void setTitle(String title) {
		this.title = title;
		updateTime();
	}


	public Status getStatus() {
		return status;
	}

	public void setStatus(Status status) {
		this.status= status;
		updateTime();
	}
	
	public void setDepositStatus(StatusValue status) {
		this.status.setValue(status);
		this.status.setInfo(null);
		updateTime();
	}
	
	public void setErrorStatus(String desc) {
		this.status.setValue(StatusValue.InError);
		this.status.setInfo(desc);
		updateTime();
	}

	public Integer getYear() {
		return year;
	}

	public void setYear(Integer year) {
		this.year = year;
		updateTime();
	}

	public int getDataFileNumber(){
		return this.getDataFiles().size();
	}
	
	public List<Person> getContributors() {
		return contributors;
	}

	public void setContributors(List<Person> contributors) {
		this.contributors = contributors;
		updateTime();
	}

	private boolean addAuthor(Person p) {
		updateTime();
		return this.contributors.add(p);
	}

	@SuppressWarnings("unchecked")
	public List<DepositDataFile> getDataFiles() {
		return dataFiles;
	}

	public void setDataFiles(List<DepositDataFile> files) {
		               this.dataFiles = files;
		               updateTime();
	}

	private boolean addDataFile(DepositDataFile f) {
		updateTime();
		return this.dataFiles.add(f);
	}

	public String getSip() {
		return sip;
	}

	public void setSip(String sip) {
		this.sip = sip;
		updateTime();
	}

	public String getSipId() {
		return sipId;
	}

	public void setSipId(String sipId) {
		this.sipId = sipId;
		updateTime();
	}

	public String getPreparationId() {
		return preparationId;
	}

	public void setPreparationId(String preparationId) {
		this.preparationId = preparationId;
		updateTime();
	}

	@Override
	public void addLinks(ControllerLinkBuilder clb, boolean mainRes, boolean subResOnly) {
		super.addLinks(clb, mainRes, subResOnly);
		if (mainRes) {
			this.add(clb.slash(getDlcmId()).slash(RESTName.PERSON).withRel(RESTName.PERSON));
			this.add(clb.slash(getDlcmId()).slash(RESTName.DATAFILE).withRel(RESTName.DATAFILE));
			this.add(clb.slash(getDlcmId()).slash(RESTName.SUBMISSION_AGREEMENT).withRel(RESTName.SUBMISSION_AGREEMENT));
			this.add(clb.slash(getDlcmId()).slash(RESTName.LICENSE).withRel(RESTName.LICENSE));
			this.add(clb.slash(getDlcmId()).slash(RESTName.LANGUAGE).withRel(RESTName.LANGUAGE));
			if (this.getStatus().getValue() == StatusValue.Completed) {
				this.add(clb.slash(getDlcmId()).slash(RESTName.SIP).withRel(RESTName.SIP));
			}
			this.add(clb.slash(getDlcmId()).slash(RESTName.UL).withRel(RESTName.UL));
			this.add(clb.slash(getDlcmId()).slash(RESTName.DL).withRel(RESTName.DL));
			this.add(clb.slash(getDlcmId()).slash(RESTName.HISTORY).withRel(RESTName.HISTORY));
		}
	}

	@Override
	public <T> boolean addItem(T t) {
		if (t instanceof Person)
			return this.addAuthor((Person) t);
		else if (t instanceof DepositDataFile)
			return this.addDataFile((DepositDataFile) t);
		return false;
	}

	@Override
	public <T> boolean removeItem(T t) {
		if (t instanceof Person)
			return this.removeAuthor((Person) t);
		else if (t instanceof DepositDataFile)
			return this.removeDataFile((DepositDataFile) t);
		return false;
	}

	private boolean removeDataFile(DepositDataFile t) {
		updateTime();
		return this.dataFiles.remove(t);
	}

	private boolean removeAuthor(Person t) {
		updateTime();
		return this.contributors.remove(t);
	}

	@Override
	public boolean update(RESTResource<Deposit> item) {
		Deposit deposit = (Deposit)item;
		this.getStatus().setType(StatusType.Deposit);
		OffsetDateTime lastUpdate = this.getUpdateTime();
		// Title
		if (StringTool.toUpdate(this.getTitle(), deposit.getTitle())){
			this.setTitle(deposit.getTitle());
		}
		// Status
		if (deposit.getStatus().getValue() != null &&
			deposit.getStatus().getValue() != this.getStatus().getValue() &&
			this.getStatus().checkValue(deposit.getStatus().getValue())){
				this.setStatus(deposit.getStatus());
		}
		// Year
		if (deposit.getYear() != null && !this.getYear().equals(deposit.getYear())){
			this.setYear(deposit.getYear());
		}
		// Doi
		if (StringTool.toUpdate(this.getDoi(), deposit.getDoi())){
			this.setDoi(deposit.getDoi());
		}
		// Description
		if (StringTool.toUpdate(this.getDescription(), deposit.getDescription())){
			this.setDescription(deposit.getDescription());
		}
		// Keywords
		if (deposit.getKeywords() != null) {
			for (int i = 0; i < this.getKeywords().size(); i++) {
				// lists are already sorted
				if (StringTool.toUpdate(this.getKeywords().get(i), deposit.getKeywords().get(i))){
					this.setKeywords(deposit.getKeywords());
					break;
				}
			}
		}
		// OrganisationId
		if (StringTool.toUpdate(this.getOrganisationId(), deposit.getOrganisationId())){
			this.setOrganisationId(deposit.getOrganisationId());
		}
		// SubmissionAgreement
		if (deposit.getSubmissionAgreement() != null && this.getSubmissionAgreement() != null &&
						!deposit.getSubmissionAgreement().equals(this.getSubmissionAgreement()) &&
						!this.getSubmissionAgreement().getDlcmId().equals(deposit.getSubmissionAgreement().getDlcmId())) {
				this.setSubmissionAgreement(deposit.getSubmissionAgreement());
			}

		// Agreement
		if (deposit.getAgreement() != null && !deposit.getAgreement().equals(this.getAgreement())){
			this.setAgreement(deposit.getAgreement());
		}
		// Access
		if (deposit.getAccess()!= null && !deposit.getAccess().equals(this.getAccess())){
			this.setAccess(deposit.getAccess());
		}
		// License
		if (deposit.getLicense()!= null && !deposit.getLicense().equals(this.getLicense())){
			this.setLicense(deposit.getLicense());
		}

		if (lastUpdate == this.getUpdateTime()){
			return false;
		}
		else{
			return true;
		}
	}

	@Transient
	@JsonIgnore
	@Override
	public Specification<Deposit> getSpecification() {
		return new DepositSpecification(this);
	}

	@JsonIgnore
	public URI getSipUri() throws DLCMException {
		if ("".equals(sip)) {
			throw new DLCMException("SIP package not available");
		}
		return URI.create(this.sip);
	}

	@JsonIgnore
	public boolean isModifiable() {
		if (this.getStatus().getValue() == StatusValue.Approved || this.getStatus().getValue() == StatusValue.Rejected
				|| this.getStatus().getValue() == StatusValue.Completed) {
			return false;
		}
		return true;
	}
	
	public boolean isAllDataFileReady() {
	  if (this.getDataFiles().isEmpty()) {
      return false;
    }
		for (DepositDataFile df : this.getDataFiles()){
			if(df.getDataFile().getStatus().getValue() != StatusValue.Ready){
				return false;
			}
		}
		return true;
	}

	@Override
	public boolean equals(Object o) {
		if (this == o) return true;
		if (!(o instanceof Deposit)) return false;
		if (!super.equals(o)) return false;

		Deposit deposit = (Deposit) o;

		if (getTitle() != null ? !getTitle().equals(deposit.getTitle()) : deposit.getTitle() != null) return false;
		if (getStatus() != null ? !getStatus().equals(deposit.getStatus()) : deposit.getStatus() != null) return false;
		if (getYear() != null ? !getYear().equals(deposit.getYear()) : deposit.getYear() != null) return false;
		if (getContributors() != null ? !getContributors().equals(deposit.getContributors()) : deposit.getContributors() != null)
			return false;
		if (getDataFiles() != null ? !getDataFiles().equals(deposit.getDataFiles()) : deposit.getDataFiles() != null)
			return false;
		if (getSipId() != null ? !getSipId().equals(deposit.getSipId()) : deposit.getSipId() != null) return false;
		if (getPreparationId() != null ? !getPreparationId().equals(deposit.getPreparationId()) : deposit.getPreparationId() != null) return false;
		return getSip() != null ? getSip().equals(deposit.getSip()) : deposit.getSip() == null;
	}

	@Override
	public int hashCode() {
		int result = super.hashCode();
		result = 31 * result + (getTitle() != null ? getTitle().hashCode() : 0);
		result = 31 * result + (getStatus() != null ? getStatus().hashCode() : 0);
		result = 31 * result + (getYear() != null ? getYear().hashCode() : 0);
		result = 31 * result + (getContributors() != null ? getContributors().hashCode() : 0);
		result = 31 * result + (getDataFiles() != null ? getDataFiles().hashCode() : 0);
		result = 31 * result + (getSipId() != null ? getSipId().hashCode() : 0);
		result = 31 * result + (getPreparationId() != null ? getPreparationId().hashCode() : 0);
		result = 31 * result + (getSip() != null ? getSip().hashCode() : 0);
		return result;
	}

	@Override
	public String managedBy() {
		return RESTName.PREINGEST;
	}

	public String getDoi() {
		return doi;
	}

	public void setDoi(String doi) {
		this.doi = doi;
		updateTime();
	}

	public Access getAccess() {
		return access;
	}

	public void setAccess(Access access) {
		this.access = access;
		updateTime();
	}

	public Language getLanguage() {
		return language;
	}

	public void setLanguage(Language language) {
		this.language = language;
		updateTime();
	}

	public String getDescription() {
		return description;
	}

	public void setDescription(String description) {
		this.description = description;
		updateTime();
	}

	public List<String> getKeywords() {
		return keywords;
	}

	public void setKeywords(List<String> keywords) {
		this.keywords = keywords;
		updateTime();
	}

	public String getOrganisationId() {
		return organisationId;
	}

	public void setOrganisationId(String organisationId) {
		this.organisationId = organisationId;
		updateTime();
	}

	@JsonIgnore
    public SubmissionAgreement getSubmissionAgreement() {
        return submissionAgreement;
    }

    @JsonProperty
    public void setSubmissionAgreement(SubmissionAgreement submissionAgreement) {
        this.submissionAgreement = submissionAgreement;
        updateTime();
    }

	public Boolean getAgreement() {
		return agreement;
	}

	public void setAgreement(Boolean agreement) {
		this.agreement = agreement;
		updateTime();
	}

	public License getLicense() {
		return license;
	}

	public void setLicense(License license) {
		this.license = license;
		updateTime();
	}
}
