package ch.dlcm.model;

import java.net.URI;
import java.util.ArrayList;
import java.util.List;

import javax.persistence.CollectionTable;
import javax.persistence.ElementCollection;
import javax.persistence.Entity;
import javax.persistence.EntityListeners;
import javax.persistence.JoinColumn;
import javax.persistence.PostRemove;
import javax.persistence.PrePersist;
import javax.persistence.PreUpdate;
import javax.persistence.UniqueConstraint;

import org.springframework.data.jpa.domain.Specification;
import org.springframework.hateoas.mvc.ControllerLinkBuilder;

import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonUnwrapped;

import ch.dlcm.criteria.DepositDataFileSpecification;
import ch.dlcm.exception.DLCMException;
import ch.dlcm.exception.DLCMFileDeleteException;
import ch.dlcm.listener.DataFileListener;
import ch.dlcm.listener.HistoryListener;
import ch.dlcm.model.Status.StatusValue;
import ch.dlcm.rest.RESTName;
import ch.dlcm.rest.RESTResource;
import ch.dlcm.tool.FileTool;

@Entity
@EntityListeners({ HistoryListener.class, DataFileListener.class })
public class DepositDataFile extends RESTResource<DepositDataFile> implements DataFileInterface {
  @JsonIgnore
  private String packageId;

  @JsonUnwrapped
  private DataFile dataFile = new DataFile();

  @ElementCollection
  @CollectionTable(name = "depositDataFileChecksum", joinColumns = {
          @JoinColumn(name = "fileId") }, uniqueConstraints = @UniqueConstraint(columnNames = {
                  "fileId", "checksumAlgo", "checksumType", "checksumOrigin" }))
  private List<DataFileChecksum> checksums = new ArrayList<>();

  public DepositDataFile() {
    init();
  }

  @Override
  public String getPackageId() {
    return packageId;
  }

  @Override
  public void setPackageId(final String packageId) {
    this.packageId = packageId;
  }

  @Override
  public DataFile getDataFile() {
    return dataFile;
  }

  @Override
  public void setDataFile(final DataFile dataFile) {
    this.dataFile = dataFile;
    updateTime();
  }

  public List<DataFileChecksum> getChecksums() {
    return checksums;
  }

  public void setChecksums(List<DataFileChecksum> checksums) {
    this.checksums = checksums;
    updateTime();
  }

  @Override
  public void addLinks(final ControllerLinkBuilder clb, final boolean mainRes,
          final boolean subResOnly) {
    super.addLinks(clb, mainRes, subResOnly);
    if (this.getDataFile().getStatus().getValue() == StatusValue.Processed
            || this.getDataFile().getStatus().getValue() == StatusValue.Ready) {
      this.add(clb.slash(getDlcmId()).slash(RESTName.DL).withRel(RESTName.DL));
    }
    if (this.getDataFile().getStatus().getValue() != StatusValue.Ready) {
      this.add(clb.slash(getDlcmId()).slash(RESTName.RESUME).withRel(RESTName.RESUME));
    }
    this.add(clb.slash(getDlcmId()).slash(RESTName.HISTORY).withRel(RESTName.HISTORY));
  }

  @Override
  public URI getDownloadUri() throws DLCMException {
    if (this.getDataFile().getStatus().getValue() != StatusValue.Processed
            && this.getDataFile().getStatus().getValue() != StatusValue.Ready) {
      throw new DLCMException("Data file not available");
    }
    if (this.getDataFile().getFinalData() == null) {
      throw new DLCMException("Data file not accessible");
    }
    return this.getDataFile().getFinalData();
  }

  @Override
  public Specification<DepositDataFile> getSpecification() {
    return new DepositDataFileSpecification(this);
  }

  @Override
  @PrePersist
  public void init() {
    dataFile.init();
    check();
  }

  @PreUpdate
  public void checkUpdate() {
    check();
  }

  @PostRemove
  public void removeFile() throws DLCMFileDeleteException {
    URI finalDataURI = this.getDataFile().getFinalData();
    if(finalDataURI == null){
      return; //nothing to delete
    }
    if (!FileTool.deleteFile(finalDataURI)) {
      throw new DLCMFileDeleteException(finalDataURI.toString());
    }
  }

  @Override
  public boolean update(final RESTResource<DepositDataFile> item) {
    DepositDataFile t = (DepositDataFile) item;
    if (this.getDataFile().update(t.getDataFile())) {
      this.updateTime();
      return true;
    }
    return false;
  }

  @Override
  @JsonIgnore
  public List<DataFileChecksum> getChecksum() {
    return checksums;
  }

  @Override
  public boolean equals(Object o) {
    if (this == o)
      return true;
    if (!(o instanceof DepositDataFile))
      return false;
    if (!super.equals(o))
      return false;

    DepositDataFile that = (DepositDataFile) o;

    if (getPackageId() != null ? !getPackageId().equals(that.getPackageId())
            : that.getPackageId() != null)
      return false;
    if (getDataFile() != null ? !getDataFile().equals(that.getDataFile())
            : that.getDataFile() != null)
      return false;
    return getChecksums() != null ? getChecksums().equals(that.getChecksums())
            : that.getChecksums() == null;
  }

  @Override
  public int hashCode() {
    int result = super.hashCode();
    result = 31 * result + (getPackageId() != null ? getPackageId().hashCode() : 0);
    result = 31 * result + (getDataFile() != null ? getDataFile().hashCode() : 0);
    result = 31 * result + (getChecksums() != null ? getChecksums().hashCode() : 0);
    return result;
  }

  @Override
  public String managedBy() {
    return RESTName.PREINGEST;
  }
}
