package ch.dlcm.model;

import java.lang.reflect.Field;
import java.net.URI;
import java.time.OffsetDateTime;
import java.util.HashMap;
import java.util.Map;

import javax.persistence.Column;
import javax.persistence.Convert;
import javax.persistence.Entity;
import javax.persistence.EnumType;
import javax.persistence.Enumerated;
import javax.validation.constraints.NotNull;

import org.springframework.data.jpa.domain.Specification;

import com.fasterxml.jackson.annotation.JsonAnyGetter;
import com.fasterxml.jackson.annotation.JsonAnySetter;
import com.fasterxml.jackson.annotation.JsonProperty;
import com.fasterxml.jackson.annotation.JsonValue;

import ch.dlcm.converter.UriConverter;
import ch.dlcm.criteria.LicenseSpecification;
import ch.dlcm.exception.DLCMException;
import ch.dlcm.rest.RESTName;
import ch.dlcm.rest.RESTResource;
import ch.dlcm.tool.StringTool;

/**
 * Structure based on open licenses. http://licenses.opendefinition.org
 */
@Entity
public class License extends RESTResource<License> {

    @NotNull
    @Column(unique = true, nullable = false)
    private String openLicenseId;
    private String title;
    private Boolean domainContent;
    private Boolean domainData;
    private Boolean domainSoftware;
    private String family;
    private String maintainer;
    private Boolean isGeneric;

    @Convert(converter = UriConverter.class)
    private URI url;

    @Enumerated(EnumType.STRING)
    private LicenseStatus status;

    @Enumerated(EnumType.STRING)
    private ConformanceStatus odConformance;

    @Enumerated(EnumType.STRING)
    private ConformanceStatus osdConformance;

    public String getOpenLicenseId() {
        return openLicenseId;
    }

    public void setOpenLicenseId(String openLicenseId) {
        this.openLicenseId = openLicenseId;
        updateTime();
    }

    public LicenseStatus getStatus() {
        return status;
    }

    public void setStatus(LicenseStatus status) {
        this.status = status;
        updateTime();
    }

    public ConformanceStatus getOdConformance() {
        return odConformance;
    }

    public void setOdConformance(ConformanceStatus odConformance) {
        this.odConformance = odConformance;
        updateTime();
    }

    public ConformanceStatus getOsdConformance() {
        return osdConformance;
    }

    public void setOsdConformance(ConformanceStatus osdConformance) {
        this.osdConformance = osdConformance;
        updateTime();
    }

    public String getTitle() {
        return title;
    }

    public void setTitle(String title) {
        this.title = title;
        updateTime();
    }

    public URI getUrl() {
        return url;
    }

    public void setUrl(URI url) {
        this.url = url;
        updateTime();
    }

    public Boolean getDomainContent() {
        return domainContent;
    }

    public void setDomainContent(Boolean domainContent) {
        this.domainContent = domainContent;
        updateTime();
    }

    public Boolean getDomainData() {
        return domainData;
    }

    public void setDomainData(Boolean domainData) {
        this.domainData = domainData;
        updateTime();
    }

    public Boolean getDomainSoftware() {
        return domainSoftware;
    }

    public void setDomainSoftware(Boolean domainSoftware) {
        this.domainSoftware = domainSoftware;
        updateTime();
    }

    public String getFamily() {
        return family;
    }

    public void setFamily(String family) {
        this.family = family;
        updateTime();
    }

    public String getMaintainer() {
        return maintainer;
    }

    public void setMaintainer(String maintainer) {
        this.maintainer = maintainer;
        updateTime();
    }

    @Override
    public Specification<License> getSpecification() {
        return new LicenseSpecification(this);
    }

    @Override
    public void init() {}

    @Override
    public boolean update(RESTResource<License> resource) {
        License license = (License) resource;
        OffsetDateTime lastUpdate = this.getUpdateTime();

        // Open licence identifier
        if (StringTool.toUpdate(this.getOpenLicenseId(), license.getOpenLicenseId())) {
            this.setOpenLicenseId(license.getOpenLicenseId());
        }
        // Title
        if (StringTool.toUpdate(this.getTitle(), license.getTitle())) {
            this.setTitle(license.getTitle());
        }
        // Url
        if (license.getUrl() != null && !license.getUrl().equals(this.getUrl())){
            this.setUrl(license.getUrl());
        }
        // Maintainer
        if (StringTool.toUpdate(this.getMaintainer(), license.getMaintainer())) {
            this.setMaintainer(license.getMaintainer());
        }
        // Family
        if (StringTool.toUpdate(this.getFamily(), license.getFamily())) {
            this.setFamily(license.getFamily());
        }
        // Domain software
        if (license.getDomainSoftware() != null && !license.getDomainSoftware().equals(this.getDomainSoftware())) {
            this.setDomainSoftware(license.getDomainSoftware());
        }
        // Domain data
        if (license.getDomainData() != null && !license.getDomainData().equals(this.getDomainData())) {
            this.setDomainData(license.getDomainData());
        }
        // Domain content
        if (license.getDomainContent() != null && !license.getDomainContent().equals(this.getDomainContent())) {
            this.setDomainContent(license.getDomainContent());
        }
        // Domain content
        if (license.getStatus() != null && !license.getStatus().equals(this.getStatus())) {
            this.setStatus(license.getStatus());
        }
        // Od Conformance
        if (license.getOdConformance() != null && !license.getOdConformance().equals(this.getOdConformance())) {
            this.setOdConformance(license.getOdConformance());
        }
        // Osd Conformance
        if (license.getOsdConformance() != null && !license.getOsdConformance().equals(this.getOsdConformance())) {
            this.setOsdConformance(license.getOsdConformance());
        }
        // Is Generic
        if (license.getIsGeneric() != null && !license.getIsGeneric().equals(this.getIsGeneric())) {
            this.setIsGeneric(license.getIsGeneric());
        }


        if (lastUpdate == this.getUpdateTime()) {
            return false;
        } else {
            return true;
        }
    }

    @Override
    public String managedBy() {
        return RESTName.ADMIN;
    }

    public Boolean getIsGeneric() {
        return isGeneric;
    }

    public void setIsGeneric(Boolean isGeneric) {
        this.isGeneric = isGeneric;
    }

    // ==============
    // Internals types
    // ==============

    /**
     * Open License proxy structure. Used to import all.json format data available on open license foundation
     */
    public static class OpenLicense extends License {

        private Map<String, License> licences = new HashMap<>();

        @JsonAnyGetter
        public Map<String, License> getLicenses() {
            return licences;
        }

        @JsonAnySetter
        public void setOpenLicenses(String name, OpenLicense openLicense) throws DLCMException{
            licences.put(name, convertToLicense(openLicense));
        }

        private License convertToLicense(OpenLicense openLicense) throws DLCMException{
            License license = new License();
            Field[] fields = License.class.getDeclaredFields();
            for (Field licenseField: fields) {
                licenseField.setAccessible(true);
                try {
                    Object fieldValue = licenseField.get(openLicense);
                    licenseField.set(license, fieldValue);
                } catch(IllegalAccessException e) {
                    throw new DLCMException("Error in license conversion", e);
                }
                licenseField.setAccessible(false);
            }
            return license;
        }

        @Override
        @JsonProperty("id")
        public String getOpenLicenseId() {
            return super.getOpenLicenseId();
        }

        @Override
        @JsonProperty("od_conformance")
        public ConformanceStatus getOdConformance() {
            return super.getOdConformance();
        }

        @Override
        @JsonProperty("osd_conformance")
        public ConformanceStatus getOsdConformance() {
            return super.getOsdConformance();
        }

        @Override
        @JsonProperty("domain_content")
        public Boolean getDomainContent() {
            return super.getDomainContent();
        }

        @Override
        @JsonProperty("domain_data")
        public Boolean getDomainData() {
            return super.getDomainData();
        }

        @Override
        @JsonProperty("domain_software")
        public Boolean getDomainSoftware() {
            return super.getDomainSoftware();
        }

        @Override
        @JsonProperty("is_generic")
        public Boolean getIsGeneric() {
            return super.getIsGeneric();
        }

    }


    public enum LicenseStatus{
        ACTIVE("active"), RETIRED("retired"), SUPERSEDED("superseded");

        private String value;

        LicenseStatus(String value) {
            this.value = value;
        }

        @JsonValue
        public String getValue(){
            return value;
        }
    }

    public enum ConformanceStatus{
        APPROVED("approved"), NOT_REVIEWED("not reviewed"), REJECTED("rejected"), EMPTY("");
        private String value;

        ConformanceStatus(String value) {
            this.value = value;
        }

        @JsonValue
        public String getValue(){
            return value;
        }
    }
    // ======
    // end internals types
    // ======

}
