package ch.dlcm.model;

import java.time.OffsetDateTime;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;

import javax.persistence.Embedded;
import javax.persistence.Entity;
import javax.persistence.EntityListeners;
import javax.persistence.EnumType;
import javax.persistence.Enumerated;
import javax.persistence.JoinColumn;
import javax.persistence.JoinTable;
import javax.persistence.ManyToMany;
import javax.persistence.PrePersist;
import javax.persistence.Table;
import javax.persistence.Transient;
import javax.persistence.UniqueConstraint;
import javax.validation.constraints.NotNull;
import javax.validation.constraints.Size;

import org.springframework.data.jpa.domain.Specification;
import org.springframework.hateoas.mvc.ControllerLinkBuilder;

import com.fasterxml.jackson.annotation.JsonIgnore;

import ch.dlcm.criteria.OrderSpecification;
import ch.dlcm.listener.HistoryListener;
import ch.dlcm.listener.PackageListener;
import ch.dlcm.model.Status.StatusType;
import ch.dlcm.model.Status.StatusValue;
import ch.dlcm.model.oais.ArchivalInfoPackage;
import ch.dlcm.model.oais.DisseminationInfoPackage;
import ch.dlcm.rest.RESTName;
import ch.dlcm.rest.RESTResource;
import ch.dlcm.rest.RESTSubResourceInterface;
import ch.dlcm.tool.StringTool;

@Entity
@Table(name = "orderQuery")
@EntityListeners({ HistoryListener.class, PackageListener.class })
public class Order extends RESTResource<Order> implements RESTSubResourceInterface {

  @NotNull
  @Size(min = 1, max = 256)
  private String name;

  @Embedded
  private Status status = new Status();

  @NotNull
  @Enumerated(EnumType.STRING)
  private QueryType queryType;

  private String query;

  @JsonIgnore
  @ManyToMany
  @JoinTable(name = "orderAip", joinColumns = { @JoinColumn(name = "orderId") }, inverseJoinColumns = {
      @JoinColumn(name = "aipId") }, uniqueConstraints = { @UniqueConstraint(columnNames = { "orderId", "aipId" }) })
  private List<ArchivalInfoPackage> aipPackages = new ArrayList<>();

  @JsonIgnore
  @ManyToMany
  @JoinTable(name = "orderDip", joinColumns = { @JoinColumn(name = "orderId") }, inverseJoinColumns = {
      @JoinColumn(name = "dipId") }, uniqueConstraints = { @UniqueConstraint(columnNames = { "orderId", "dipId" }) })
  private List<DisseminationInfoPackage> dipPackages = new ArrayList<>();

  public Order() {
    this.status.setType(StatusType.Order);
  }

  @PrePersist
  @Override
  public void init() {
    if (this.getStatus().getValue() == null) {
      this.getStatus().setValue(StatusValue.InProgress);
    }
  }

  public List<DisseminationInfoPackage> getDipPackages() {
    return dipPackages;
  }

   public void setDipPackages(List<DisseminationInfoPackage> dipList) {
    this.dipPackages = dipList;
    updateTime();
  }

  public String getName() {
    return name;
  }

  public void setName(String name) {
    this.name = name;
    updateTime();
  }

  public Status getStatus() {
    return status;
  }

  public void setStatus(Status status) {
    this.status = status;
    updateTime();
  }

  public void setOrderStatus(StatusValue status) {
    this.status.setValue(status);
    this.status.setInfo(null);
    updateTime();
  }

  public void setErrorStatus(String desc) {
    this.status.setValue(StatusValue.InError);
    this.status.setInfo(desc);
    updateTime();
  }

  public String getQuery() {
    return query;
  }

  public void setQuery(String query) {
    this.query = query;
    updateTime();
  }
  
  public int getAipNumber(){
    return this.getAipPackages().size();
  }
  
  public int getDipNumber(){
    return this.getDipPackages().size();
  }

  @Override
  public void addLinks(ControllerLinkBuilder clb, boolean mainRes, boolean subResOnly) {
    super.addLinks(clb, mainRes, subResOnly);
    if (mainRes) {
      this.add(clb.slash(getDlcmId()).slash(RESTName.AIP).withRel(RESTName.AIP));
      this.add(clb.slash(getDlcmId()).slash(RESTName.VIEW).withRel(RESTName.VIEW));
      this.add(clb.slash(getDlcmId()).slash(RESTName.SAVE).withRel(RESTName.SAVE));

      if (this.getDipNumber() > 0) {
        this.add(clb.slash(getDlcmId()).slash(RESTName.DIP).withRel(RESTName.DIP));
      }
      if (this.getStatus().getValue() == StatusValue.Ready) {
        this.add(clb.slash(getDlcmId()).slash(RESTName.DL).withRel(RESTName.DL));
      }
      this.add(clb.slash(getDlcmId()).slash(RESTName.HISTORY).withRel(RESTName.HISTORY));
    }
  }

  @Override
  public boolean update(RESTResource<Order> item) {
    Order t = (Order) item;
    this.getStatus().setType(StatusType.Order);
    OffsetDateTime lastUpdate = this.getUpdateTime();
    if (StringTool.toUpdate(this.getName(), t.getName())) {
      this.setName(t.getName());
    }
    if (t.getStatus().getValue() != null && this.getStatus().getValue() != t.getStatus().getValue()) {
      if (this.getStatus().checkValue(t.getStatus().getValue())) {
        this.setStatus(t.getStatus());
      }
    }
    if (StringTool.toUpdate(this.getQuery(), t.getQuery())) {
      this.setQuery(t.getQuery());
    }

    if (lastUpdate == this.getUpdateTime()) {
      return false;
    } else {
      return true;
    }
  }

  @Transient
  @JsonIgnore
  @Override
  public Specification<Order> getSpecification() {
    return new OrderSpecification(this);
  }

  @Override
  public <T> boolean addItem(T t) {
    if (t instanceof DisseminationInfoPackage) {
      return this.addDip((DisseminationInfoPackage) t);
    } else if (t instanceof ArchivalInfoPackage) {
      return this.addAip((ArchivalInfoPackage) t);
    }
    return false;
  }

  private boolean addAip(ArchivalInfoPackage aip) {
    if (!this.exist(this.aipPackages, aip)) {
      updateTime();
      return this.aipPackages.add(aip);
    }
    return false;
  }

  private boolean addDip(DisseminationInfoPackage dip) {
    updateTime();
    return this.dipPackages.add(dip);
  }

  @Override
  public <T> boolean removeItem(T t) {
    if (t instanceof DisseminationInfoPackage) {
      return this.removeDip((DisseminationInfoPackage) t);
    } else if (t instanceof ArchivalInfoPackage) {
      return this.removeAip((ArchivalInfoPackage) t);
    }
    return false;
  }

  private boolean removeAip(ArchivalInfoPackage aip) {
    updateTime();
    return this.aipPackages.remove(aip);
  }

  private boolean removeDip(DisseminationInfoPackage dip) {
    updateTime();
    return this.dipPackages.remove(dip);
  }

  @Override
  public String managedBy() {
    return RESTName.ACCESS;
  }

  @Override
  public boolean equals(final Object other) {
    if (other == null) {
      return false;
    }
    if (!getClass().equals(other.getClass())) {
      return false;
    }
    Order castOther = (Order) other;
    return Objects.equals(getName(), castOther.getName()) && Objects.equals(getStatus(), castOther.getStatus())
        && Objects.equals(getQueryType(), castOther.getQueryType()) && Objects.equals(getQuery(), castOther.getQuery())
        && Objects.equals(getDipPackages(), castOther.getDipPackages());
  }

  public QueryType getQueryType() {
    return queryType;
  }

  public void setQueryType(QueryType queryType) {
    this.queryType = queryType;
  }

  public List<ArchivalInfoPackage> getAipPackages() {
    return aipPackages;
  }

  public void setAipPackages(List<ArchivalInfoPackage> aipPackages) {
    this.aipPackages = aipPackages;
  }

  @Override
  public int hashCode() {
    return Objects.hash(getName(), getStatus(), getQueryType(), getQuery(), getAipPackages(), getDipPackages());
  }

  public boolean isCompleted() {
    if (this.getDipPackages().isEmpty()) {
      return false;
    }
    for(DisseminationInfoPackage dip:this.getDipPackages()) {
      if (dip.getInfo().getStatus().getValue() != StatusValue.Completed) {
        return false;
      }
    }
    return true;
  }
  
  public boolean isDipInError() {
    if (this.getDipPackages().isEmpty()) {
      return false;
    }
    for(DisseminationInfoPackage dip:this.getDipPackages()) {
      if (dip.getInfo().getStatus().getValue() == StatusValue.InError) {
        return true;
      }
    }
    return false;
  }

  public boolean isEmptyQuery() {
    if (this.query != null) {
      if (!this.query.isEmpty()) {
        return false;
      }
    }
    return true;
  }


}
