package ch.dlcm.model;

import java.time.OffsetDateTime;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;
import java.util.function.Predicate;

import javax.persistence.CascadeType;
import javax.persistence.Entity;
import javax.persistence.JoinColumn;
import javax.persistence.JoinTable;
import javax.persistence.ManyToMany;
import javax.persistence.OneToMany;
import javax.persistence.Transient;
import javax.persistence.UniqueConstraint;
import javax.validation.constraints.NotNull;
import javax.validation.constraints.Size;

import org.springframework.data.jpa.domain.Specification;
import org.springframework.hateoas.mvc.ControllerLinkBuilder;

import com.fasterxml.jackson.annotation.JsonIgnore;

import ch.dlcm.criteria.OrganizationalUnitSpecification;
import ch.dlcm.rest.RESTName;
import ch.dlcm.rest.RESTResource;
import ch.dlcm.rest.RESTSubResourceInterface;
import ch.dlcm.tool.StringTool;

@Entity
public class OrganizationalUnit extends RESTResource<OrganizationalUnit> implements RESTSubResourceInterface {

  @NotNull
  @Size(min = 1, max = 256)
  private String name;

  @Size(max = 1024)
  private String description;

  @JsonIgnore
  @ManyToMany
  @JoinTable(name = "organizationalUnitInstitutions",
          joinColumns = {@JoinColumn(name = "parentId")},
          inverseJoinColumns = {@JoinColumn(name = "childId")},
          uniqueConstraints = {@UniqueConstraint(columnNames = {"parentId", "childId"})})
  private List<Institution> institutions;

  @JsonIgnore
  @OneToMany(mappedBy = "dlcmCompositeId.organizationalUnit", cascade = CascadeType.ALL, orphanRemoval = true)
  private List<OrganizationalUnitPersonRole> organizationalUnitPersonRoles = new ArrayList<>();

  @JsonIgnore
  @ManyToMany
  @JoinTable(name = "organizationalUnitFundingAgencies",
          joinColumns = {@JoinColumn(name = "parentId")},
          inverseJoinColumns = {@JoinColumn(name = "childId")},
          uniqueConstraints = {@UniqueConstraint(columnNames = {"parentId", "childId"})})
  private List<FundingAgency> fundingAgencies;
  
  
  
  
  @Override
  public boolean equals(final Object other) {
    if (!(other instanceof OrganizationalUnit)) {
      return false;
    }
    if (!super.equals(other)) {
      return false;
    }
    OrganizationalUnit castOther = (OrganizationalUnit) other;
    return Objects.equals(name, castOther.name) && Objects.equals(description, castOther.description)
            && Objects.equals(institutions, castOther.institutions) && Objects.equals(organizationalUnitPersonRoles, castOther.organizationalUnitPersonRoles);
  }

  @Override
  public int hashCode() {
    return Objects.hash(super.hashCode(), name, description, institutions, organizationalUnitPersonRoles);
  }

  public boolean addInstitution(Institution institution) {
    updateTime();
    boolean result = institutions.add(institution);
    if (!institution.getProjects().contains(this)) {
      institution.addProject(this);
    }
    return result;
  }

  protected boolean removeInstitution(Institution institution) {
    updateTime();
    boolean result = institutions.remove(institution);
    if (institution.getProjects().contains(this)) {
      institution.removeProject(this);
    }
    return result;
  }
  
  public boolean addFundingAgency(FundingAgency fundingAgency) {
    updateTime();
    boolean result = fundingAgencies.add(fundingAgency);
    if (!fundingAgency.getOrganizationalUnits().contains(this)) {
    	fundingAgency.addOrganizationalUnit(this);
    }
    return result;
  }

  public boolean removeFundingAgency(FundingAgency fundingAgency) {
    updateTime();
    boolean result = fundingAgencies.remove(fundingAgency);
    if (fundingAgency.getOrganizationalUnits().contains(this)) {
    	fundingAgency.removeOrganizationalUnit(this);
    }
    return result;
  }
  
  public List<FundingAgency> getFundingAgencies() {
    return this.fundingAgencies;
  }
  
  public boolean addPerson(Person person, List<Role> roles) {
    for (Role role : roles) {
      if (!addPerson(person, role)) {
        return false;
      }
    }
    return true;
  }

  private boolean addPerson(Person person, Role role) {
    updateTime();
    OrganizationalUnitPersonRole organizationalUnitPersonRole = new OrganizationalUnitPersonRole();
    organizationalUnitPersonRole.setRole(role);
    organizationalUnitPersonRole.setOrganizationalUnit(this);
    organizationalUnitPersonRole.setPerson(person);
    return this.organizationalUnitPersonRoles.add(organizationalUnitPersonRole);
  }

  /**
   * @deprecated Use instead addPerson(Person, Role)
   * TODO to be removed
   */
  protected boolean addPerson(Person person) {
    updateTime();
    boolean result = getPeople().add(person);
    if (!person.getProjects().contains(this)) {
      person.addProject(this);
    }
    return result;
  }

  protected boolean removePerson(Person person) {
    updateTime();
    final Person personFinal = person;
    Predicate<OrganizationalUnitPersonRole> organizationalUnitPersonRolePredicate = ouprPredicate -> ouprPredicate.getDlcmCompositeId().getPerson() == personFinal;
    boolean result = this.organizationalUnitPersonRoles.removeIf(organizationalUnitPersonRolePredicate);
    return result;
  }


  @Override
  public void init() {
  }

  @Override
  public void addLinks(ControllerLinkBuilder clb, boolean mainRes, boolean subResOnly) {
    super.addLinks(clb, mainRes, subResOnly);
    if (mainRes) {
      this.add(clb.slash(getDlcmId()).slash(RESTName.INSTITUTION).withRel(RESTName.INSTITUTION));
      this.add(clb.slash(getDlcmId()).slash(RESTName.PERSON).withRel(RESTName.PERSON));
      this.add(clb.slash(getDlcmId()).slash(RESTName.FUNDING_AGENCY).withRel(RESTName.FUNDING_AGENCY));
    }
  }

  /**
   * @deprecated TODO to be redefined
   * Use instead addPerson(Person, Role)
   */
  @Override
  public <T> boolean addItem(T t) {
    if (t instanceof Institution) {
      return this.addInstitution((Institution) t);
    }
    if (t instanceof Person) {
      return this.addPerson((Person) t);
    }
    if (t instanceof FundingAgency) {
      return this.addFundingAgency((FundingAgency)t);
    }
    return false;
  }

  @Override
  public <T> boolean removeItem(T t) {
    
    if (t instanceof Institution) {
      return this.removeInstitution((Institution) t);
    }
    
    if (t instanceof Person) {
      return this.removePerson((Person) t);
    }
    
    if (t instanceof OrganizationalUnitPersonRole) {
      return this.removePersonRole((OrganizationalUnitPersonRole) t);
    }
    
    if (t instanceof FundingAgency) {
      return this.removeFundingAgency((FundingAgency) t);
    }
    
    return false;
  }

  private boolean removePersonRole(OrganizationalUnitPersonRole unitPersonRole) {
    updateTime();
    Predicate<OrganizationalUnitPersonRole> organizationalUnitPersonRolePredicate = ouprPredicate -> ouprPredicate.equals(unitPersonRole);
    boolean result = this.organizationalUnitPersonRoles.removeIf(organizationalUnitPersonRolePredicate);
    return result;
  }

  @Override
  public boolean update(RESTResource<OrganizationalUnit> item) {
    OrganizationalUnit t = (OrganizationalUnit) item;
    OffsetDateTime lastUpdate = this.getUpdateTime();
    if (StringTool.toUpdate(this.getName(), t.getName()))
      this.setName(t.getName());
    if (StringTool.toUpdate(this.getDescription(), t.getDescription()))
      this.setDescription(t.getDescription());
    if (lastUpdate == this.getUpdateTime())
      return false;
    else
      return true;
  }

  public String getName() {
    return name;
  }

  public String getDescription() {
    return description;
  }

  public List<Institution> getInstitutions() {
    return institutions;
  }

  @Transient
  @JsonIgnore
  public final List<Person> getPeople() {
    List<Person> people = new ArrayList<>();
    this.organizationalUnitPersonRoles.forEach(organizationalUnitPersonRole -> {
      final Predicate<List<Person>> filterCriteria = listPrecicate -> listPrecicate.contains(organizationalUnitPersonRole.getPerson());
      if (filterCriteria.negate().test(people)) {
        people.add(organizationalUnitPersonRole.getPerson());
      }
    });
    return people;
  }

  public void setName(String name) {
    this.name = name;
    updateTime();
  }

  public void setDescription(String description) {
    this.description = description;
    updateTime();
  }

  public List<OrganizationalUnitPersonRole> getOrganizationalUnitPersonRoles() {
    return organizationalUnitPersonRoles;
  }

  @Transient
  @JsonIgnore
  @Override
  public Specification<OrganizationalUnit> getSpecification() {
    return new OrganizationalUnitSpecification(this);
  }

  @Override
  public String managedBy() {
    return RESTName.ADMIN;
  }

}
