package ch.dlcm.model;

import ch.dlcm.criteria.PersonSpecification;
import ch.dlcm.rest.RESTName;
import ch.dlcm.rest.RESTResource;
import ch.dlcm.rest.RESTSubResourceInterface;
import ch.dlcm.tool.StringTool;
import com.fasterxml.jackson.annotation.*;
import org.springframework.data.jpa.domain.Specification;
import org.springframework.hateoas.mvc.ControllerLinkBuilder;

import javax.persistence.*;
import javax.validation.constraints.NotNull;
import javax.validation.constraints.Size;
import java.time.OffsetDateTime;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.function.Predicate;

@Entity
public class Person extends RESTResource<Person> implements RESTSubResourceInterface {

  @NotNull
  @Size(min = 1, max = 256)
  private String firstName;

  @NotNull
  @Size(min = 1, max = 256)
  private String lastName;

  @JsonIgnore
  @ManyToMany
  @JoinTable(name = "personInstitutions",
          joinColumns = {@JoinColumn(name = "parentId")},
          inverseJoinColumns = {@JoinColumn(name = "childId")},
          uniqueConstraints = {@UniqueConstraint(columnNames = {"parentId", "childId"})})
  private List<Institution> institutions;

  @JsonIgnore
  @OneToMany(mappedBy = "dlcmCompositeId.person", cascade = CascadeType.ALL, fetch = FetchType.LAZY, orphanRemoval = true)
  protected List<OrganizationalUnitPersonRole> organizationalUnitPersonRoles = new ArrayList<>();

  private String orcid;

  @Transient
  @JsonIgnore
  public List<OrganizationalUnit> getProjects() {
    List<OrganizationalUnit> projects = new ArrayList<>();
    this.organizationalUnitPersonRoles.forEach(organizationalUnitPersonRole -> {
      final Predicate<List<OrganizationalUnit>> filterCriteria = listPrecicate -> listPrecicate.contains(organizationalUnitPersonRole.getOrganizationalUnit());
      if (filterCriteria.negate().test(projects)) {
        projects.add(organizationalUnitPersonRole.getOrganizationalUnit());
      }
    });
    return projects;
  }

  @JsonIgnore
  @Transient
  public List<Map<String, List<Role>>> getProjectRoles() {
    Map<String, List<Role>> projectRoles = new HashMap<>();
    getProjects().stream().forEach(organizationalUnit -> {
      List<Role> roles = new ArrayList<>();
      final Predicate<OrganizationalUnitPersonRole> projectCriteria = listPrecicate -> listPrecicate.getOrganizationalUnit().equals(organizationalUnit);
      this.organizationalUnitPersonRoles.stream().filter(projectCriteria).forEach(organizationalUnitPersonRole1 ->
        roles.add(organizationalUnitPersonRole1.getRole())
      );
      projectRoles.put(organizationalUnit.getDlcmId(), roles);
    });
    List<Map<String, List<Role>>> projectRolesList = new ArrayList<>();
    projectRolesList.add(projectRoles);
    return projectRolesList;
  }

  public String getFirstName() {
    return firstName;
  }

  public void setFirstName(String firstName) {
    this.firstName = firstName;
    updateTime();
  }

  public String getLastName() {
    return lastName;
  }

  public void setLastName(String lastName) {
    this.lastName = lastName;
    updateTime();
  }

  public List<Institution> getInstitutions() {
    return institutions;
  }

  public boolean addInstitution(Institution institution) {
    updateTime();
    boolean result = institutions.add(institution);
    if (!institution.getPeople().contains(this)) {
      institution.addPerson(this);
    }
    return result;
  }

  public boolean removeInstitution(Institution institution) {
    updateTime();
    boolean result = institutions.remove(institution);
    if (institution.getPeople().contains(this)) {
      institution.removePerson(this);
    }
    return result;
  }

  public boolean addProject(OrganizationalUnit project, List<Role> roles) {
    for (Role role : roles) {
      if (!addProject(project, role)) {
        return false;
      }
    }
    return true;
  }

  private boolean addProject(OrganizationalUnit project, Role role) {
    updateTime();
    OrganizationalUnitPersonRole organizationalUnitPersonRole = new OrganizationalUnitPersonRole();
    organizationalUnitPersonRole.setRole(role);
    organizationalUnitPersonRole.setOrganizationalUnit(project);
    organizationalUnitPersonRole.setPerson(this);
    return this.organizationalUnitPersonRoles.add(organizationalUnitPersonRole);
  }


  /**
   * @deprecated Use instead addProject(OrganizationalUnit, Role)
   * TODO to be replaced
   */
  protected boolean addProject(OrganizationalUnit unit) {
    updateTime();
    boolean result = getProjects().add(unit);
    if (!unit.getPeople().contains(this)) {
      unit.addPerson(this);
    }
    return result;
  }

  protected boolean removeProject(OrganizationalUnit unit) {
    updateTime();
    final OrganizationalUnit project = unit;
    Predicate<OrganizationalUnitPersonRole> organizationalUnitPersonRolePredicate = ouprPredicate -> ouprPredicate.getDlcmCompositeId().getOrganizationalUnit() == project;
    boolean result = this.organizationalUnitPersonRoles.removeIf(organizationalUnitPersonRolePredicate);
    return result;
  }

  @Override
  public void init() {
  }

  @Override
  public void addLinks(ControllerLinkBuilder clb, boolean mainRes, boolean subResOnly) {
    super.addLinks(clb, mainRes, subResOnly);
    if (mainRes) {
      this.add(clb.slash(getDlcmId()).slash(RESTName.INSTITUTION).withRel(RESTName.INSTITUTION));
      this.add(clb.slash(getDlcmId()).slash(RESTName.UNIT).withRel(RESTName.UNIT));
    }
  }

  @Override
  public <T> boolean addItem(T t) {
    if (t instanceof Institution)
      return this.addInstitution((Institution) t);
    return false;
  }

  @Override
  public <T> boolean removeItem(T t) {
    if (t instanceof Institution) {
      return this.removeInstitution((Institution) t);
    }
    if (t instanceof OrganizationalUnit) {
      return this.removeProject((OrganizationalUnit) t);
    }
    if (t instanceof OrganizationalUnitPersonRole) {
      return this.removeProjectRole((OrganizationalUnitPersonRole) t);
    }
    return false;
  }

  private boolean removeProjectRole(OrganizationalUnitPersonRole unitPersonRole) {
    updateTime();
    Predicate<OrganizationalUnitPersonRole> organizationalUnitPersonRolePredicate = ouprPredicate -> ouprPredicate.equals(unitPersonRole);
    boolean result = this.organizationalUnitPersonRoles.removeIf(organizationalUnitPersonRolePredicate);
    return result;
  }

  @Override
  public boolean update(RESTResource<Person> personResource) {
    Person person = (Person) personResource;
    OffsetDateTime lastUpdate = this.getUpdateTime();
    if (StringTool.toUpdate(this.getFirstName(), person.getFirstName())) {
      this.setFirstName(person.getFirstName());
    }
    if (StringTool.toUpdate(this.getLastName(), person.getLastName())) {
      this.setLastName(person.getLastName());
    }
    if (StringTool.toUpdate(this.getOrcid(), person.getOrcid())) {
      this.setOrcid(person.getOrcid());
    }

    if (lastUpdate == this.getUpdateTime())
      return false;
    else
      return true;
  }

  public List<OrganizationalUnitPersonRole> getOrganizationalUnitPersonRoles() {
    return organizationalUnitPersonRoles;
  }


  @Transient
  @JsonIgnore
  @Override
  public Specification<Person> getSpecification() {
    return new PersonSpecification(this);
  }

  @Override
  public String managedBy() {
    return RESTName.ADMIN;
  }

  public String getOrcid() {
    return orcid;
  }

  public void setOrcid(String orcid) {
    this.orcid = orcid;
  }
}
