package ch.dlcm.model;

import java.security.InvalidParameterException;
import java.util.Objects;

import javax.persistence.Column;
import javax.persistence.Embeddable;
import javax.persistence.EnumType;
import javax.persistence.Enumerated;
import javax.persistence.Transient;
import javax.validation.constraints.NotNull;

import com.fasterxml.jackson.annotation.JsonIgnore;

@Embeddable
public class Status {
	@NotNull
	@Enumerated(EnumType.STRING)
	@Column(name = "status")
	private StatusValue value;

	@Column(name = "statusInfo")
	private String info;

	@JsonIgnore
	@Transient
	private StatusType type;

	public enum StatusValue {
		InError, Paused, // All
		InProgress, InValidation, Rejected, Approved, Submitted, Completed, // Deposit
		Received, ToProcess, Processed, // DataFile
		InPreparation, Ready, Checked, Stored, Indexing, Reindexing, Downloading, // Package
		InDisseminationPreparation // Order
	}

	public enum StatusType {
		Deposit, DataFile, Package, Order
	}

	public Status() {
	}
	
	@Override
	public boolean equals(final Object other) {
		if (!(other instanceof Status)) {
			return false;
		}
		Status castOther = (Status) other;
		return Objects.equals(value, castOther.value) && Objects.equals(info, castOther.info)
				&& Objects.equals(type, castOther.type);
	}

	@Override
	public int hashCode() {
		return Objects.hash(value, info, type);
	}

	public StatusValue getValue() {
		return value;
	}

	public void setValue(StatusValue value) {
		this.value = value;
	}

	public void setType(StatusType type) {
		this.type = type;
	}

	/**
	 * Different workflows by type
	 *  ----------------------------------------------------------- <br>
	 *  status      | Datafile  | Deposit | Package | Order 	    <br>
	 *  ----------------------------------------------------------- <br>
	 * InError      |  X			X			X		X			<br>
	 * InProgress	|				X			X		X			<br>
	 * InValidation	|				X								<br>
	 * Rejected		|				X								<br>
	 * Approved		|				X								<br>
	 * Submitted	|				X					X			<br>
	 * Completed	|				X			X					<br>
	 * Received		|  X											<br>
	 * ToProcess	|  X											<br>
	 * Processed	|  X											<br>
	 * InPreparation|							X		X			<br>
	 * Ready		|							X		X			<br>
	 * Checked		|							X					<br>
	 * Stored		|							X					<br>
	 * Indexing		|							X					<br>
	 * Reindexing	|							X					<br>
	 */
	public static StatusValue[] listStatusValues(StatusType t) {
		switch (t) {
		case DataFile:
			return new StatusValue[] { StatusValue.Received, StatusValue.ToProcess, StatusValue.Processed, StatusValue.Ready,
					StatusValue.InError };
		case Deposit:
			return new StatusValue[] { StatusValue.InProgress, StatusValue.InValidation, StatusValue.Rejected,
					StatusValue.Approved, StatusValue.Submitted, StatusValue.Completed, StatusValue.InError,
					StatusValue.Paused};
		case Package:
			return new StatusValue[] { StatusValue.InPreparation, StatusValue.Ready, StatusValue.Checked,
					StatusValue.InProgress, StatusValue.Stored, StatusValue.Indexing, StatusValue.Completed, 
					StatusValue.Reindexing, StatusValue.Downloading, StatusValue.InError };
		case Order:
			return new StatusValue[] { StatusValue.InProgress, StatusValue.Submitted, StatusValue.InPreparation, 
			    StatusValue.InDisseminationPreparation, StatusValue.Ready, StatusValue.InError };
		default:
			throw new InvalidParameterException();
		}
	}

	public boolean checkValue(StatusValue value) {
		// Check if value exists
		StatusValue.valueOf(value.toString());
		// check if value is applicable by type
		for(StatusValue v: listStatusValues()){
			if (value == v){
				return true;
			}
		}
		return false;
	}

	private StatusValue[] listStatusValues() {
		return listStatusValues(this.type);
	}

	public String getInfo() {
		return info;
	}

	public void setInfo(String info) {
		this.info = info;
	}

  public void reset() {
    this.value = StatusValue.InProgress;
    this.info=null;    
  }

}
