package ch.dlcm.model.oais;

import java.net.URI;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;

import javax.persistence.CascadeType;
import javax.persistence.CollectionTable;
import javax.persistence.Column;
import javax.persistence.ElementCollection;
import javax.persistence.Entity;
import javax.persistence.EntityListeners;
import javax.persistence.JoinColumn;
import javax.persistence.OneToMany;
import javax.persistence.PrePersist;
import javax.persistence.Table;
import javax.persistence.UniqueConstraint;

import org.springframework.data.jpa.domain.Specification;
import org.springframework.hateoas.mvc.ControllerLinkBuilder;

import com.fasterxml.jackson.annotation.JsonIgnore;

import ch.dlcm.criteria.ArchivalInfoPackageSpecification;
import ch.dlcm.exception.DLCMException;
import ch.dlcm.listener.HistoryListener;
import ch.dlcm.listener.PackageListener;
import ch.dlcm.model.Status.StatusValue;
import ch.dlcm.rest.RESTName;
import ch.dlcm.rest.RESTResource;
import ch.dlcm.rest.RESTSubResourceInterface;

@Entity
@Table(name="aip")
@EntityListeners({HistoryListener.class, PackageListener.class})
public class ArchivalInfoPackage extends RESTResource<ArchivalInfoPackage> implements RESTSubResourceInterface, InfoPackageInterface {
	
	private RepresentationInfo info = new RepresentationInfo();
	
	@JsonIgnore
	@OneToMany(cascade = CascadeType.ALL, orphanRemoval = true)
	@JoinColumn(name="collectionId", referencedColumnName="dlcmId")
	private List<ArchivalInfoPackage> collection = new ArrayList<>();

	@JsonIgnore
	@OneToMany(cascade = CascadeType.ALL, orphanRemoval = true)
	@JoinColumn(name="packageId", referencedColumnName="dlcmId")
	private List<AipDataFile> dataFiles = new ArrayList<>();

	@ElementCollection
	@Column(name="sipId")
	@CollectionTable(name = "aipSip", joinColumns = { @JoinColumn(name = "packageId") },
		 			 uniqueConstraints= @UniqueConstraint(columnNames={"packageId","sipId"}))
	private List<String> sipIds = new ArrayList<>();
	
	private String archiveId;

	@Override
	public boolean equals(final Object other) {
		if (!(other instanceof ArchivalInfoPackage)) {
			return false;
		}
		ArchivalInfoPackage castOther = (ArchivalInfoPackage) other;
		return Objects.equals(info, castOther.info) && Objects.equals(collection, castOther.collection)
				&& Objects.equals(dataFiles, castOther.dataFiles) && Objects.equals(sipIds, castOther.sipIds)
				&& Objects.equals(archiveId, castOther.archiveId);
	}

	@Override
	public int hashCode() {
		return Objects.hash(info, collection, dataFiles, sipIds, archiveId);
	}

	@JsonIgnore
	public boolean isCollection(){
		return this.collection != null;
	}
	
	@JsonIgnore
	public boolean isUnit(){
		return this.dataFiles != null;
	}
	
	public RepresentationInfo getInfo() {
		return info;
	}
	
	public void setInfo(RepresentationInfo info) {
		this.info = info;
		updateTime();
	}
	public List<ArchivalInfoPackage> getCollection() {
		return collection;
	}

	public void setCollection(List<ArchivalInfoPackage> collection) {
		this.collection = collection;
		updateTime();
	}
	
	public List<AipDataFile> getDataFiles() {
		return dataFiles;
	}

	public void setDataFiles(List<AipDataFile> dataFiles) {
		this.dataFiles = dataFiles;
		updateTime();
	}
	
	public String getArchiveId() {
		return archiveId;
	}

	public void setArchiveId(String archiveId) {
		this.archiveId = archiveId;
		updateTime();
	}

	public List<String> getSipIds() {
		return sipIds;
	}

	public void setSipIds(List<String> sipIds) {
		this.sipIds = sipIds;
		updateTime();
	}
	
	public void addSip(String sip){
		if (this.sipIds == null){
			this.sipIds = new ArrayList<>();
		}
		this.sipIds.add(sip);
		updateTime();
	}

	@Override
	public Specification<ArchivalInfoPackage> getSpecification() {
		return new ArchivalInfoPackageSpecification(this);
	}
	
	@Override
	@PrePersist
	public void init() {
		info.init();
	}

	@Override
	public boolean update(RESTResource<ArchivalInfoPackage> item) {
		ArchivalInfoPackage t = (ArchivalInfoPackage)item;
		if(this.getInfo().update(t.getInfo())){
			this.updateTime();
			return true;
		}
		return false;
	}

	@Override
	public <T> boolean addItem(T t) {
		if (t instanceof AipDataFile)
			return this.addDataFile((AipDataFile) t);
		return false;
	}
	
	private boolean addDataFile(AipDataFile t) {
		updateTime();
		return this.dataFiles.add(t);
	}

	@Override
	public <T> boolean removeItem(T t) {
		if (t instanceof AipDataFile)
			return this.removeDataFile((AipDataFile) t);
		return false;
	}

	private boolean removeDataFile(AipDataFile t) {
		updateTime();
		return this.dataFiles.remove(t);
	}

	@Override
	public void addLinks(ControllerLinkBuilder clb, boolean mainRes, boolean subResOnly) {
		super.addLinks(clb, mainRes, subResOnly);
		if (mainRes) {
			this.add(clb.slash(getDlcmId()).slash(RESTName.DATAFILE).withRel(RESTName.DATAFILE));
			if (this.getInfo().getStatus().getValue() != StatusValue.InPreparation) {
				this.add(clb.slash(getDlcmId()).slash(RESTName.DL).withRel(RESTName.DL));
			}
			if (this.getInfo().getStatus().getValue() != StatusValue.Completed) {
				this.add(clb.slash(getDlcmId()).slash(RESTName.RESUME).withRel(RESTName.RESUME));
			}
			if (this.getInfo().getStatus().getValue() == StatusValue.Completed 
					|| this.getInfo().getStatus().getValue() == StatusValue.Stored) {
				this.add(clb.slash(getDlcmId()).slash(RESTName.REINDEX).withRel(RESTName.REINDEX));
			}
			if (this.getInfo().getStatus().getValue() != StatusValue.Completed
					&& this.getInfo().getStatus().getValue() != StatusValue.Stored){
				this.add(clb.slash(getDlcmId()).slash(RESTName.UL).withRel(RESTName.UL));
			}
			this.add(clb.slash(getDlcmId()).slash(RESTName.HISTORY).withRel(RESTName.HISTORY));
		}
	}
	
	public void addLinksForAccess(ControllerLinkBuilder clb, boolean mainRes, boolean subResOnly) {
	  super.addLinks(clb, mainRes, subResOnly);
    if (mainRes) {
      this.add(clb.slash(getDlcmId()).slash(RESTName.DATAFILE).withRel(RESTName.DATAFILE));
      if (this.getInfo().getStatus().getValue() == StatusValue.Completed) {
        this.add(clb.slash(getDlcmId()).slash(RESTName.DL).withRel(RESTName.DL));
      }else {
        this.add(clb.slash(getDlcmId()).slash(RESTName.RESUME).withRel(RESTName.RESUME));
      }
      this.add(clb.slash(getDlcmId()).slash(RESTName.HISTORY).withRel(RESTName.HISTORY));
    }
  }

	public boolean isAllDataFileReady() {
	  if (this.getDataFiles().isEmpty()) {
	    return false;
	  }
		for (AipDataFile df : this.getDataFiles()){
			if(df.getDataFile().getStatus().getValue() != StatusValue.Ready){
				return false;
			}
		}
		return true;
	}
	
	@Override
	public URI getDownloadUri() throws DLCMException {
		if (this.getInfo().getStatus().getValue() == StatusValue.InPreparation) {
			throw new DLCMException("Package not available");
		}
		if (this.getDataFiles().isEmpty()) {
			throw new DLCMException("Package not accessible");
		}
		if (this.getInfo().getStatus().getValue() == StatusValue.Stored) {
			return URI.create(this.getArchiveId());
		}
		return this.getDataFiles().get(0).getDataFile().getFinalData();
	}

	@Override
	public String managedBy() {
		return RESTName.ARCHIVALSTORAGE;
	}

  public void reset() {
    this.getInfo().getStatus().reset();
    this.archiveId = null;
  }
	
}
