package ch.dlcm.model.oais;

import java.net.URI;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;

import javax.persistence.CollectionTable;
import javax.persistence.ElementCollection;
import javax.persistence.Entity;
import javax.persistence.EntityListeners;
import javax.persistence.JoinColumn;
import javax.persistence.PostRemove;
import javax.persistence.PrePersist;
import javax.persistence.PreUpdate;
import javax.persistence.UniqueConstraint;

import org.springframework.data.jpa.domain.Specification;
import org.springframework.hateoas.mvc.ControllerLinkBuilder;

import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonUnwrapped;

import ch.dlcm.exception.DLCMException;
import ch.dlcm.exception.DLCMFileDeleteException;
import ch.dlcm.listener.DataFileListener;
import ch.dlcm.listener.HistoryListener;
import ch.dlcm.model.DataFile;
import ch.dlcm.model.DataFileChecksum;
import ch.dlcm.model.DataFileInterface;
import ch.dlcm.model.Status.StatusValue;
import ch.dlcm.rest.RESTName;
import ch.dlcm.rest.RESTResource;
import ch.dlcm.tool.FileTool;

@Entity
@EntityListeners({ HistoryListener.class, DataFileListener.class })
public class SipDataFile extends RESTResource<SipDataFile> implements DataFileInterface {
	@JsonIgnore
	private String packageId;

	@JsonUnwrapped
	private DataFile dataFile = new DataFile();

	@ElementCollection
	@CollectionTable(name = "sipDataFileChecksum", joinColumns = {
			@JoinColumn(name = "fileId") }, uniqueConstraints = @UniqueConstraint(columnNames = { "fileId",
					"checksumAlgo", "checksumType", "checksumOrigin" }))
	private List<DataFileChecksum> checksums = new ArrayList<>();

	@Override
	public boolean equals(final Object other) {
		if (!(other instanceof SipDataFile)) {
			return false;
		}
		SipDataFile castOther = (SipDataFile) other;
		return Objects.equals(packageId, castOther.packageId) && Objects.equals(dataFile, castOther.dataFile)
				&& Objects.equals(checksums, castOther.checksums);
	}

	@Override
	public int hashCode() {
		return Objects.hash(packageId, dataFile, checksums);
	}

	public SipDataFile() {
		init();
	}

	@Override
	public String getPackageId() {
		return packageId;
	}

	@Override
	public void setPackageId(String packageId) {
		this.packageId = packageId;
	}

	@Override
	public DataFile getDataFile() {
		return dataFile;
	}

	@Override
	public void setDataFile(DataFile dataFile) {
		this.dataFile = dataFile;
	}

	public List<DataFileChecksum> getChecksums() {
		return checksums;
	}

	public void setChecksums(List<DataFileChecksum> checksums) {
		this.checksums = checksums;
	}

	@Override
	public void addLinks(ControllerLinkBuilder clb, boolean mainRes, boolean subResOnly) {
		super.addLinks(clb, mainRes, subResOnly);
		if (this.getDataFile().isAvailable()) {
			this.add(clb.slash(getDlcmId()).slash(RESTName.DL).withRel(RESTName.DL));
		}
		if (this.getDataFile().getStatus().getValue() != StatusValue.Ready) {
			this.add(clb.slash(getDlcmId()).slash(RESTName.RESUME).withRel(RESTName.RESUME));
		}
		this.add(clb.slash(getDlcmId()).slash(RESTName.HISTORY).withRel(RESTName.HISTORY));
	}

	@Override
	public URI getDownloadUri() throws DLCMException {
		if (!this.getDataFile().isAvailable()) {
			throw new DLCMException("Data file not available");
		}
		if (this.getDataFile().getFinalData() == null) {
			throw new DLCMException("Data file not accessible");
		}
		return this.getDataFile().getFinalData();
	}

	@Override
	@PrePersist
	public void init() {
		dataFile.init();
		check();
	}

	@PreUpdate
	public void checkUpdate() {
		check();
	}
	
	@PostRemove
  public void removeFile() throws DLCMFileDeleteException {
		URI finalDataURI = this.getDataFile().getFinalData();
		if(finalDataURI == null){
			return; //nothing to delete
		}
    if (!FileTool.deleteFile(finalDataURI)) {
      throw new DLCMFileDeleteException(finalDataURI.toString());
    }
  }

	@Override
	public boolean update(RESTResource<SipDataFile> item) {
		SipDataFile t = (SipDataFile) item;
		if (this.getDataFile().update(t.getDataFile())) {
			this.updateTime();
			return true;
		}
		return false;
	}

	@Override
	public Specification<SipDataFile> getSpecification() {
		// TODO return new SipDataFileSpecification(this);
		return null;
	}

	@Override
	@JsonIgnore
	public List<DataFileChecksum> getChecksum() {
		if (this.checksums == null) {
			this.checksums = new ArrayList<>();
		}
		return this.checksums;
	}

	@Override
	public String managedBy() {
		return RESTName.INGEST;
	}

}
