package ch.dlcm.model.oais;

import java.net.URI;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;

import javax.persistence.CascadeType;
import javax.persistence.Entity;
import javax.persistence.EntityListeners;
import javax.persistence.JoinColumn;
import javax.persistence.OneToMany;
import javax.persistence.PrePersist;
import javax.persistence.Table;

import org.springframework.data.jpa.domain.Specification;
import org.springframework.hateoas.mvc.ControllerLinkBuilder;

import com.fasterxml.jackson.annotation.JsonIgnore;

import ch.dlcm.criteria.SubmissionInfoPackageSpecification;
import ch.dlcm.exception.DLCMException;
import ch.dlcm.listener.HistoryListener;
import ch.dlcm.listener.PackageListener;
import ch.dlcm.model.Status.StatusValue;
import ch.dlcm.rest.RESTName;
import ch.dlcm.rest.RESTResource;
import ch.dlcm.rest.RESTSubResourceInterface;

@Entity
@Table(name = "sip")
@EntityListeners({HistoryListener.class, PackageListener.class})
public class SubmissionInfoPackage extends RESTResource<SubmissionInfoPackage> implements RESTSubResourceInterface, InfoPackageInterface {

	private RepresentationInfo info = new RepresentationInfo();

	@JsonIgnore
	@OneToMany(cascade = CascadeType.ALL, orphanRemoval = true)
	@JoinColumn(name = "packageId", referencedColumnName = "dlcmId")
	private List<SipDataFile> dataFiles = new ArrayList<>();

	private String depositId;
	private String aipId;
	
	@JsonIgnore
	private String aip;

	@Override
	public boolean equals(final Object other) {
		if (!(other instanceof SubmissionInfoPackage)) {
			return false;
		}
		SubmissionInfoPackage castOther = (SubmissionInfoPackage) other;
		return Objects.equals(info, castOther.info) && Objects.equals(dataFiles, castOther.dataFiles)
				&& Objects.equals(depositId, castOther.depositId) && Objects.equals(aipId, castOther.aipId)
				&& Objects.equals(aip, castOther.aip);
	}

	@Override
	public int hashCode() {
		return Objects.hash(info, dataFiles, depositId, aipId, aip);
	}

	public RepresentationInfo getInfo() {
		return info;
	}

	public void setInfo(RepresentationInfo info) {
		this.info = info;
		updateTime();
	}

	@SuppressWarnings("unchecked")
	public List<SipDataFile> getDataFiles() {
		return dataFiles;
	}

	public void setDataFiles(List<SipDataFile> dataFiles) {
		this.dataFiles = dataFiles;
		updateTime();
	}

	public String getAip() {
		return aip;
	}

	public void setAip(String aip) {
		this.aip = aip;
		updateTime();
	}

	public String getDepositId() {
		return depositId;
	}

	public void setDepositId(String depositId) {
		this.depositId = depositId;
		updateTime();
	}

	public String getAipId() {
		return aipId;
	}

	public void setAipId(String aipId) {
		this.aipId = aipId;
		updateTime();
	}

	@JsonIgnore
	public URI getAipUri() throws DLCMException {
		if ("".equals(aip)) {
			throw new DLCMException("AIP package not available");
		}
		return URI.create(this.aip);
	}

	@Override
	public Specification<SubmissionInfoPackage> getSpecification() {
		return new SubmissionInfoPackageSpecification(this);
	}

	@Override
	@PrePersist
	public void init() {
		info.init();
	}

	@Override
	public boolean update(RESTResource<SubmissionInfoPackage> item) {
		SubmissionInfoPackage t = (SubmissionInfoPackage)item;
		if (this.getInfo().update(t.getInfo())) {
			this.updateTime();
			return true;
		}
		return false;
	}

	@Override
	public <T> boolean addItem(T t) {
		if (t instanceof SipDataFile)
			return this.addDataFile((SipDataFile) t);
		return false;
	}

	private boolean addDataFile(SipDataFile t) {
		updateTime();
		return this.dataFiles.add(t);
	}

	@Override
	public void addLinks(ControllerLinkBuilder clb, boolean mainRes, boolean subResOnly) {
		super.addLinks(clb, mainRes, subResOnly);
		if (mainRes) {
			this.add(clb.slash(getDlcmId()).slash(RESTName.DATAFILE).withRel(RESTName.DATAFILE));
			if (this.getInfo().getStatus().getValue() == StatusValue.Completed) {
				this.add(clb.slash(getDlcmId()).slash(RESTName.AIP).withRel(RESTName.AIP));
			}
			if (this.getInfo().getStatus().getValue() != StatusValue.InPreparation) {
				this.add(clb.slash(getDlcmId()).slash(RESTName.DL).withRel(RESTName.DL));
			}
			if (this.getInfo().getStatus().getValue() != StatusValue.Completed) {
				this.add(clb.slash(getDlcmId()).slash(RESTName.UL).withRel(RESTName.UL));
				this.add(clb.slash(getDlcmId()).slash(RESTName.RESUME).withRel(RESTName.RESUME));
			}
			this.add(clb.slash(getDlcmId()).slash(RESTName.HISTORY).withRel(RESTName.HISTORY));
		}
	}

	@Override
	public <T> boolean removeItem(T t) {
		if (t instanceof SipDataFile)
			return this.removeDataFile((SipDataFile) t);
		return false;
	}

	private boolean removeDataFile(SipDataFile t) {
		updateTime();
		return this.dataFiles.remove(t);
	}

	public boolean isAllDataFileReady() {
	  if (this.getDataFiles().isEmpty()) {
      return false;
    }
		for (SipDataFile df : this.getDataFiles()) {
			if (df.getDataFile().getStatus().getValue() != StatusValue.Ready) {
				return false;
			}
		}
		return true;
	}

	@Override
	public URI getDownloadUri() throws DLCMException {
		if (this.getInfo().getStatus().getValue() == StatusValue.InPreparation) {
			throw new DLCMException("Package not available");
		}
		if (this.getDataFiles().isEmpty()) {
			throw new DLCMException("Package not accessible");
		}
		return this.getDataFiles().get(0).getDataFile().getFinalData();
	}

	@Override
	public String managedBy() {
		return RESTName.INGEST;
	}
}
