package ch.dlcm.tool;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.StandardCopyOption;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.ArrayList;
import java.util.List;
import java.util.stream.Collectors;

import org.apache.log4j.Logger;
import org.springframework.http.MediaType;
import org.springframework.util.FileCopyUtils;
import org.springframework.util.FileSystemUtils;

import ch.dlcm.model.DataFileChecksum;
import ch.dlcm.model.DataFileChecksum.ChecksumAlgorithm;

public class FileTool {
	private static final Logger log = Logger.getLogger(FileTool.class);

	public static File getFileFromUri(URI fileuri) {
		return new File(fileuri.getPath());
	}

	public static String getFileUri(String file) {
		return getFileUri(new File(file));
	}

	public static String getFileUri(File file) {
		return file.toURI().toString();
	}

	public static boolean checkDirectory(File file) {
		if (file.isDirectory()) {
			return true;
		}
		return file.mkdirs();
	}

	public static List<Path> readDirectory(String dir) {
		try {
			return Files.list(Paths.get(dir)).filter(Files::isRegularFile).collect(Collectors.toList());
		} catch (IOException e) {
			log.error(e);
		}
		return new ArrayList<>();
	}
	
	public static boolean deleteDirectory(File dir) {
	  if (FileSystemUtils.deleteRecursively(dir)) {
	    return true;
	  }
	  return false;
	}
	
	public static boolean copyFile(File src, File dest) throws IOException {
	  if (FileCopyUtils.copy(src, dest) > 0) {
	    return true;
	  }
	  return false;
  }
	
	public static boolean moveFile(File src, File dest) throws IOException {
    if (FileTool.copyFile(src, dest)) {
      if (FileTool.deleteDirectory(src)) {
        return true;
      }
    }
    return false;
  }
	
	public static MediaType getContentType(URI fileUri) throws URISyntaxException {
		File f = getFileFromUri(fileUri);
		try {
			String fileType = Files.probeContentType(f.toPath());
			if (fileType == null) {
				return MediaType.APPLICATION_OCTET_STREAM;
			}
			return MediaType.valueOf(fileType);
		} catch (IOException ioException) {
			log.error(ioException.getMessage(), ioException);
			return MediaType.APPLICATION_OCTET_STREAM;
		}
	}
	
	public static boolean copyFile(Path source, Path destination) {
		try {
			if (!FileTool.checkDirectory(destination.getParent().toFile())) {
				log.error("Cannot create directory: "+destination.getParent().toString());
				return false;
			}
			Files.copy(source, destination, StandardCopyOption.COPY_ATTRIBUTES);
		} catch (IOException e) {
			log.error(e.getMessage(), e);
			return false;
		}
		return true;
	}

	public static boolean deleteFile(URI fileUri) {
		try {
			File f = getFileFromUri(fileUri);
			if (f.exists()) {
				Files.delete(f.toPath());
				log.info("File " + fileUri + " deleted");
			}
		} catch (IOException e) {
			log.error(e.getMessage(), e);
			return false;
		}
		return true;
	}
	
	public static boolean copyFile(String source, String target) {
		File srcFile = new File(source);
		File tgtFile = new File(target);
		if (tgtFile.exists()) {
			log.error("File "+target+" already exists");
			return false;
		}
		try {
			if (FileCopyUtils.copy(srcFile, tgtFile) > 0) {
				return true;
			}
		} catch (IOException e) {
			log.error(e.getMessage(), e);
			return false;
		}
		return true;
	}


	public static long getSize(URI fileUri) throws URISyntaxException {
		File f = getFileFromUri(fileUri);
		try {
			return Files.size(f.toPath());
		} catch (IOException e) {
			log.error(e.getMessage(), e);
			return 0;
		}
	}

	public static String[] computeChecksum(URI fileUri, String[] checksums)
			throws URISyntaxException, IOException, NoSuchAlgorithmException {
		File f = getFileFromUri(fileUri);
		MessageDigest[] digests = new MessageDigest[checksums.length];
		for (int i = 0; i < checksums.length; i++) {
			digests[i] = getMessageDigest(ChecksumAlgorithm.valueOf(checksums[i]));
		}
		return getFileChecksum(digests, f);
	}

	public static DataFileChecksum computeChecksum(String file, DataFileChecksum checksum)
			throws NoSuchAlgorithmException, IOException, URISyntaxException {
		if (checksum == null) {
			checksum = new DataFileChecksum();
		}
		File f = new File(new URI(file).getPath());
		MessageDigest digest = getMessageDigest(checksum.getChecksumAlgo());

		checksum.setChecksum(getFileChecksum(digest, f));
		return checksum;
	}

	private static MessageDigest getMessageDigest(ChecksumAlgorithm checksumAlgo) throws NoSuchAlgorithmException {
		switch (checksumAlgo) {
		case MD5:
			return MessageDigest.getInstance("MD5");
		case SHA1:
			return MessageDigest.getInstance("SHA-1");
		case SHA256:
			return MessageDigest.getInstance("SHA-256");
		default:
			throw new NoSuchAlgorithmException();
		}
	}

	public static boolean checkChecksum(String file, DataFileChecksum checksum)
			throws NoSuchAlgorithmException, IOException, URISyntaxException {
		File f = new File(new URI(file).getPath());
		String newChecksum;
		switch (checksum.getChecksumAlgo()) {
		case MD5:
			newChecksum = getFileChecksum(MessageDigest.getInstance("MD5"), f);
			break;
		case SHA1:
			newChecksum = getFileChecksum(MessageDigest.getInstance("SHA-1"), f);
			break;
		case SHA256:
			newChecksum = getFileChecksum(MessageDigest.getInstance("SHA-256"), f);
			break;
		default:
			throw new NoSuchAlgorithmException();
		}

		return newChecksum.equals(checksum.getChecksum());
	}

	private static String getFileChecksum(MessageDigest digest, File file) throws IOException {
		FileInputStream fis = new FileInputStream(file);
		byte[] byteArray = new byte[1024];
		int bytesCount;

		// Read file data and update in message digest
		while ((bytesCount = fis.read(byteArray)) != -1) {
			digest.update(byteArray, 0, bytesCount);
		}
		fis.close();

		// Get the hash's bytes
		byte[] bytes = digest.digest();

		// Convert it to hexadecimal format
		StringBuilder sb = new StringBuilder();
		for (byte aByte : bytes) {
			sb.append(Integer.toString((aByte & 0xff) + 0x100, 16).substring(1));
		}

		return sb.toString();
	}

	private static String[] getFileChecksum(MessageDigest[] digests, File file) throws IOException {
		FileInputStream fis = new FileInputStream(file);
		byte[] byteArray = new byte[1024];
		int bytesCount;

		// Read file data and update in message digest
		while ((bytesCount = fis.read(byteArray)) != -1) {
			for (MessageDigest digest : digests) {
				digest.update(byteArray, 0, bytesCount);
			}
		}
		fis.close();

		String[] checksums = new String[digests.length];
		for (int i = 0; i < digests.length; i++) {
			byte[] bytes = digests[i].digest();

			// Convert it to hexadecimal format
			StringBuilder sb = new StringBuilder();
			for (byte aByte : bytes) {
				sb.append(Integer.toString((aByte & 0xff) + 0x100, 16).substring(1));
			}
			checksums[i] = sb.toString();
		}

		return checksums;
	}

}
