package ch.dlcm.tool;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.StringReader;
import java.net.URL;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.List;

import javax.xml.XMLConstants;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParserFactory;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.sax.SAXSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;
import javax.xml.validation.Schema;
import javax.xml.validation.SchemaFactory;
import javax.xml.validation.Validator;

import org.apache.log4j.Logger;
import org.json.JSONObject;
import org.json.XML;
import org.w3c.dom.Document;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;

import ch.dlcm.model.ArchiveMetadata;

public class XMLTool {
	private static final Logger log = Logger.getLogger(XMLTool.class);

	public static final String SCHEMA_LANGUAGE = "http://java.sun.com/xml/jaxp/properties/schemaLanguage";
	public static final String XML_VALIDATION = "http://xml.org/sax/features/validation";
	public static final String SCHEMA_VALIDATION = "http://apache.org/xml/features/validation/schema";
	public static final String SCHEMA_FULL_VALIDATION = "http://apache.org/xml/features/validation/schema-full-checking";
	public static final String SCHEMA_SOURCE = "http://java.sun.com/xml/jaxp/properties/schemaSource";

	public static ArchiveMetadata toArchiveMetadata(String index, String id, Path path) {
		ArchiveMetadata am = new ArchiveMetadata();
		am.setDlcmId(id);
		am.setType("metadata");
		am.setIndex(index);
		am.setMetadata(XML2JSON(path).toString());
		return am;
	}

	public static JSONObject XML2JSON(Path file) {
		String xml;
		try {
			xml = new String(Files.readAllBytes(file));
			return XML.toJSONObject(xml);
		} catch (IOException e) {
			log.error(e);
			return new JSONObject("{\"file\":\"" + file.toString() + "\", \"error\":\"" + e.getMessage() + "\"}");
		}
	}

	public static String XML2String(Path file) {
		try {
			return new String(Files.readAllBytes(file));
		} catch (IOException e) {
			log.error(e);
			return "<error \"file\"=\"" + file.toString() + "\">" + e.getMessage() + "</error>";
		}
	}

	public static boolean transformFile(Path xml, Path xsl, Path newXml) {

		try {
			StreamSource source = new StreamSource(Files.newInputStream(xml));
			StreamSource stylesource = new StreamSource(Files.newInputStream(xsl));

			TransformerFactory factory = TransformerFactory.newInstance();
			Transformer transformer;

			transformer = factory.newTransformer(stylesource);

			StreamResult result = new StreamResult(newXml.toFile());
			result.setSystemId(java.net.URLDecoder.decode(result.getSystemId(), "UTF-8"));
			transformer.transform(source, result);
		} catch (TransformerException | IOException e) {
			log.error(e);
			return false;
		}
		return true;
	}

	public static boolean validate(String xsd, String xml) {

		DocumentBuilder parser;
		Document doc;

		try {
			// parse an XML document into a DOM tree
			parser = DocumentBuilderFactory.newInstance().newDocumentBuilder();
			parser.setErrorHandler(new XMLError());

			doc = parser.parse(new InputSource(new StringReader(xml)));

			// create a SchemaFactory capable of understanding WXS schemas
			SchemaFactory factory = SchemaFactory.newInstance(XMLConstants.W3C_XML_SCHEMA_NS_URI);

			// load a WXS schema, represented by a Schema instance
			Schema schema = factory.newSchema(new SAXSource(new InputSource(new StringReader(xsd))));

			// create a Validator instance, which can be used to validate an
			// instance document
			Validator validator = schema.newValidator();

			// validate the DOM tree
			validator.validate(new DOMSource(doc));

		} catch (ParserConfigurationException | SAXException | IOException e) {
			log.error(e);
			return false;
		}
		return true;
	}

	public static boolean transform(String xml, String xsl, String newXml) {
		StreamSource source = new StreamSource(new StringReader(xml));
		StreamSource stylesource = new StreamSource(new StringReader(xsl));

		TransformerFactory factory = TransformerFactory.newInstance();
		Transformer transformer;
		try {
			transformer = factory.newTransformer(stylesource);
			File file = new File(newXml);
			StreamResult result = new StreamResult(file);
			transformer.transform(source, result);
		} catch (TransformerException e) {
			log.error(e);
			return false;
		}

		return true;
	}

	public static boolean wellformed(Path xml) {

		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		factory.setValidating(false);
		factory.setNamespaceAware(true);

		DocumentBuilder parser;
		try(InputStream inputStream = Files.newInputStream(xml)) {
			parser = factory.newDocumentBuilder();
			parser.setErrorHandler(new XMLError());
			parser.parse(new InputSource(new InputStreamReader(inputStream, "UTF-8")));
		} catch (ParserConfigurationException | SAXException | IOException e) {
			log.error(e);
			return false;
		}

		return true;
	}

	public static boolean validate(List<URL> xsds, Path xml) {
		return validate(xsds, xml, true);
	}
	
	public static boolean validate(List<URL> xsds, Path xml, boolean domParser) {

		SchemaFactory schemaFactory = SchemaFactory.newInstance(XMLConstants.W3C_XML_SCHEMA_NS_URI);
		Source[] sources;
		Schema schema;

		// Schema factory
		try {
			sources = getSchemaSources(xsds);
			schemaFactory.setFeature(SCHEMA_FULL_VALIDATION, true);
			schemaFactory.setResourceResolver(new SchemaResourceResolver());
			schema = schemaFactory.newSchema(sources);
		} catch (IOException | SAXException e) {
			log.error("{XML Schema} "+e);
			return false;
		}
					
		try {
			if (domParser) {
				parseWithDom(schema,xml);
			}else {
				parseWithSax(schema, xml);
			}
		} catch (ParserConfigurationException | SAXException | IOException e) {
			log.error("{XML Document} "+e);
			return false;
		}

		return true;
	}
	
	private static void parseWithDom(Schema schema, Path xml) throws ParserConfigurationException, SAXException, IOException {
		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		// Doc Factory
		factory.setSchema(schema);
		factory.setValidating(false);
		factory.setNamespaceAware(true);
		// factory.setXIncludeAware(true);
		// factory.setFeature(XML_VALIDATION, true);
		// factory.setFeature(SCHEMA_VALIDATION, true);
		// factory.setFeature(SCHEMA_FULL_VALIDATION, true);

		// Parse document
		DocumentBuilder parser = factory.newDocumentBuilder();
		parser.setErrorHandler(new XMLError());
		parser.parse(Files.newInputStream(xml));
	}
	
	private static void parseWithSax(Schema schema, Path xml) throws ParserConfigurationException, SAXException, IOException {
		SAXParserFactory factory = SAXParserFactory.newInstance();
		// Doc Factory
		factory.setSchema(schema);
		factory.setValidating(false);
		factory.setNamespaceAware(true);

		// Parse document
        XMLReader reader = factory.newSAXParser().getXMLReader();
        reader.setErrorHandler(new XMLError());
        reader.parse(new InputSource(Files.newInputStream(xml)));
	}

	private static Source[] getSchemaSources(List<URL> xsds) throws IOException {
		Source[] sources = new Source[xsds.size()];
		int i = 0;
		for (URL xsd : xsds) {
			sources[i] = new StreamSource(xsd.openStream());
			i++;
		}
		return sources;
	}
}
