package ch.dlcm.tool;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.net.URI;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.List;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;
import java.util.zip.ZipInputStream;
import java.util.zip.ZipOutputStream;

import org.apache.log4j.Logger;

public class ZipTool {
  private static final Logger log = Logger.getLogger(ZipTool.class);

  String archive;
  List<String> fileList;

  public ZipTool(URI a) {
    this.archive = a.getRawPath();
  }

  public ZipTool(String a) {
    this.archive = a;
  }

  public String getArchive() {
    return archive;
  }

  public boolean zipFiles(String source) {
    byte[] buffer = new byte[1024];
    int count = 0;
    try (FileOutputStream fileOut = new FileOutputStream(archive); ZipOutputStream zipOut = new ZipOutputStream(fileOut)) {
      resetFileList();
      generateFileList(source, new File(source));
      for (String file : this.fileList) {
        count++;
        File f = new File(source + "/" + file);
        ZipEntry zipEntry = new ZipEntry(file);
        zipOut.putNextEntry(zipEntry);
        if (f.isFile()) {
          try (FileInputStream fileIn = new FileInputStream(f)) {
            int len;
            while ((len = fileIn.read(buffer)) > 0) {
              zipOut.write(buffer, 0, len);
            }
          }
        }
        zipOut.closeEntry();
      }
    } catch (FileNotFoundException e) {
      log.error("Cannot create Zip package (" + archive + ")", e);
      return false;
    } catch (IOException e) {
      log.error("Problem in creating Zip package (" + source + ".zip)", e);
      return false;
    }
    log.info("Zip package ('" + source + "' in '" + this.archive + "') created with " + count + " entries");
    return true;
  }

  public boolean unzipFiles(String destination) {
    int count = 0;
    try (ZipInputStream zis = new ZipInputStream(new FileInputStream(this.archive))) {
      ZipEntry ze = zis.getNextEntry();
      while (ze != null) {
          saveFile(zis, ze, destination);
          count++;
          ze = zis.getNextEntry();
      }
      zis.closeEntry();
      zis.close();
    } catch (IOException e) {
      log.error("Problem in unzipping package (" + this.archive + ")", e);
      return false;
    }
    log.info("Unzip package ('" + this.archive + "') in '" + destination + "': " + count + " entries extracted");
    return true;
  }

  private void saveFile(ZipInputStream zipFile, ZipEntry zipEntry,String destination) {
    byte[] buffer = new byte[1024];
    File newFile = new File(destination + "/" + zipEntry.getName());
    if (zipEntry.isDirectory()) {
      FileTool.checkDirectory(newFile);
    } else {
        if (!FileTool.checkDirectory(newFile.getParentFile())) {
          log.error("Cannot create parent directory (" + newFile.getAbsolutePath() + ")");
          return;
        }
        try(FileOutputStream fos = new FileOutputStream(newFile)){
          int len;
          while ((len = zipFile.read(buffer)) > 0) {
            fos.write(buffer, 0, len);
          }
        } catch (IOException e) {
          log.error(e);
        }
    }
  }

  public boolean unzipFiles(String destination, List<String> list) {
    final int[] count = { 0 };
    try (ZipFile zipFile = new ZipFile(this.archive)) {
      zipFile.stream().filter(ze -> isFileCandidate(list, ze)).forEach(ze -> {
        saveFile(zipFile, ze, destination);
        count[0]++;
      });
    } catch (IOException e) {
      log.error("Problem in unzipping package (" + this.archive + ")", e);
      return false;
    }
    log.info("Unzip package ('" + this.archive + "') in '" + destination + "': " + count[0] + " entries extracted");
    return true;
  }

  private boolean isFileCandidate(List<String> list, ZipEntry ze) {
    if (list == null) {
      return true;
    }
    for (String f : list) {
      if (ze.getName().startsWith(f)) {
        return true;
      }
    }
    return false;
  }

  private void saveFile(ZipFile zipFile, ZipEntry zipEntry, String destination) {
    File newFile = new File(destination + "/" + zipEntry.getName());
    if (zipEntry.isDirectory()) {
      FileTool.checkDirectory(newFile);
    } else {
        if (!FileTool.checkDirectory(newFile.getParentFile())) {
          log.error("Cannot create parent directory (" + newFile.getAbsolutePath() + ")");
          return;
        }
        try(BufferedInputStream zis = new BufferedInputStream(zipFile.getInputStream(zipEntry));
            FileOutputStream fos = new FileOutputStream(newFile)){
           while (zis.available() > 0) {
              fos.write(zis.read());
           }
        } catch (IOException e) {
          log.error(e);
        }
    }
  }

  public int getEntryNumber() {
    try (ZipFile zipFile = new ZipFile(this.archive)) {
      return zipFile.size();
    } catch (IOException e) {
      log.error(e);
      return 0;
    }
  }

  public boolean checkZip() {
    try (ZipFile zipFile = new ZipFile(this.archive)) {
      Enumeration<? extends ZipEntry> entries = zipFile.entries();
      while (entries.hasMoreElements()) {
        ZipEntry zipEntry = entries.nextElement();
        zipFile.getInputStream(zipEntry);
        zipEntry.getCrc();
        zipEntry.getCompressedSize();
        zipEntry.getName();
      }
      return true;
    } catch (IOException e) {
      log.error(e);
      return false;
    }
  }

  private void resetFileList() {
    if (fileList == null) {
      fileList = new ArrayList<>();
    } else {
      fileList.clear();
    }
  }

  private void generateFileList(String src, File file) {
    if (file.isFile()) {
      fileList.add(generateZipEntry(src, file.getAbsoluteFile().toString()));
    }

    if (file.isDirectory()) {
      for (String filename : file.list()) {
        generateFileList(src, new File(file, filename));
      }
    }

  }

  private String generateZipEntry(String dir, String file) {
    return file.substring(dir.length() + 1, file.length());
  }

}
