package ch.dlcm.test.tool;

import static org.assertj.core.api.Assertions.fail;

import java.io.IOException;
import java.net.URISyntaxException;
import java.net.URL;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.Logger;
import org.hamcrest.core.IsEqual;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.ErrorCollector;
import org.junit.runner.RunWith;
import org.springframework.core.io.ClassPathResource;
import org.springframework.test.context.junit4.SpringJUnit4ClassRunner;

import ch.dlcm.tool.FileTool;
import ch.dlcm.tool.SchemaResourceResolver;
import ch.dlcm.tool.XMLTool;

@RunWith(SpringJUnit4ClassRunner.class)
public class XMLTest {
	private static final Logger log = Logger.getLogger(XMLTest.class);

	@Rule
	public ErrorCollector collector = new ErrorCollector();

	private static String PREFIX = "test/test";
	private static String DLCM = "dlcm/dlcm";

	private enum XML {
		SCHEMA(0), GOOD(1), GOOD_WO_SCHEMA(2), GOOD_WRONG_SCHEMA(3), NOT_VALID(4), NOT_WELLFORMED(5);
		private final int value;

		private XML(int value) {
			this.value = value;
		}

		public int getValue() {
			return value;
		}
	}

	@Test
	public void wellFormTest() {
		List<Path> fileList = loadFilesTest(PREFIX);
		collector.checkThat(true, IsEqual.equalTo((XMLTool.wellformed(fileList.get(XML.SCHEMA.getValue())))));
		collector.checkThat(true, IsEqual.equalTo((XMLTool.wellformed(fileList.get(XML.GOOD.getValue())))));
		collector.checkThat(true, IsEqual.equalTo((XMLTool.wellformed(fileList.get(XML.NOT_VALID.getValue())))));
		collector.checkThat(false, IsEqual.equalTo((XMLTool.wellformed(fileList.get(XML.NOT_WELLFORMED.getValue())))));
	}

	@Test
	public void validationTest() {
		List<Path> fileList = loadFilesTest(PREFIX);
		List<URL> xsdList = loadSchemaTest(PREFIX);
		collector.checkThat(true, IsEqual.equalTo((XMLTool.validate(xsdList, fileList.get(XML.GOOD.getValue())))));
		collector.checkThat(true, IsEqual.equalTo((XMLTool.validate(xsdList, fileList.get(XML.GOOD_WO_SCHEMA.getValue())))));
		collector.checkThat(true, IsEqual.equalTo((XMLTool.validate(xsdList, fileList.get(XML.GOOD_WRONG_SCHEMA.getValue())))));
		collector.checkThat(false, IsEqual.equalTo((XMLTool.validate(xsdList, fileList.get(XML.NOT_VALID.getValue())))));
		collector.checkThat(false, IsEqual.equalTo((XMLTool.validate(xsdList, fileList.get(XML.NOT_WELLFORMED.getValue())))));
	}
	
	@Test
	public void validationTestStandards() throws IOException {
		String[] dirList = {"DataCite-AllIn1-4.0", "DataCite-4.0", "METS-1.11","PREMIS-3.0"};
		int count=0;
		for (String dir: dirList) {
			log.info(dir);
			List<Path> xmlList = new ArrayList<>();
			List<URL> xsdList = new ArrayList<>();
			for (Path p: FileTool.readDirectory(getPath(dir).toString())) {
				if (p.toString().endsWith(".xml")) {
					xmlList.add(p);
				}
			}
			for (Path p: FileTool.readDirectory(getPath(SchemaResourceResolver.SCHEMA_HOME).toString())) {
				String f = p.getFileName().toString();
				if (f.startsWith(dir) && f.endsWith(".xsd")) {
					log.info("*Schema:"+f);
					xsdList.add(p.toUri().toURL());
				}
			}
			
			for (Path xml: xmlList) {
				count++;
				log.info("-"+xml.getFileName().toString());
				collector.checkThat(xml.getFileName().toString(), true, IsEqual.equalTo((XMLTool.validate(xsdList, xml))));
			}
		}	
		log.info(count+" executed tests");
	}


	@Test
	public void wellFormTestDLCM() {
		List<Path> fileList = loadFilesTest(DLCM);
		collector.checkThat(true, IsEqual.equalTo((XMLTool.wellformed(fileList.get(XML.SCHEMA.getValue())))));
		collector.checkThat(true, IsEqual.equalTo((XMLTool.wellformed(fileList.get(XML.GOOD.getValue())))));
		collector.checkThat(true, IsEqual.equalTo((XMLTool.wellformed(fileList.get(XML.NOT_VALID.getValue())))));
		collector.checkThat(false, IsEqual.equalTo((XMLTool.wellformed(fileList.get(XML.NOT_WELLFORMED.getValue())))));
	}

	@Test
	public void wellFormTestDLCMSchemas() throws URISyntaxException {
		List<URL> xsdList = loadSchemaTest(DLCM);
		for (URL xsd : xsdList) {
			collector.checkThat(true, IsEqual.equalTo((XMLTool.wellformed(Paths.get(xsd.toURI())))));
		}
	}

	//@Test
	public void validationTestDLCM() throws IOException {
		List<Path> fileList = loadFilesTest(DLCM);
		List<URL> xsdList = loadSchemaTest(DLCM);

		collector.checkThat(true, IsEqual.equalTo((XMLTool.validate(xsdList, fileList.get(XML.GOOD.getValue())))));
		collector.checkThat(true, IsEqual.equalTo((XMLTool.validate(xsdList, getPath("dlcm/dlcm.xml")))));
		collector.checkThat(true, IsEqual.equalTo((XMLTool.validate(xsdList, getPath("dlcm/dlcm_datacite.xml")))));
		collector.checkThat(true, IsEqual.equalTo((XMLTool.validate(xsdList, fileList.get(XML.GOOD.getValue())))));
		collector.checkThat(true, IsEqual.equalTo((XMLTool.validate(xsdList, fileList.get(XML.GOOD_WO_SCHEMA.getValue())))));
		collector.checkThat(true, IsEqual.equalTo((XMLTool.validate(xsdList, fileList.get(XML.GOOD_WRONG_SCHEMA.getValue())))));
		collector.checkThat(false, IsEqual.equalTo((XMLTool.validate(xsdList, fileList.get(XML.NOT_VALID.getValue())))));
		collector.checkThat(false, IsEqual.equalTo((XMLTool.validate(xsdList, fileList.get(XML.NOT_WELLFORMED.getValue())))));
	}

	private List<Path> loadFilesTest(String prefix) {
		List<Path> list = new ArrayList<>();
		try {
			list.add(getPath(prefix + ".xsd")); // 0 = XML Schema
			list.add(getPath(prefix + "_w_schema.xml")); // 1 = Good XML
			list.add(getPath(prefix + "_wo_schema.xml")); // 2 = Good XML without schema
			list.add(getPath(prefix + "_w_wrong_schema.xml")); // 3 = Good XML without schema
			list.add(getPath(prefix + "_not_valid.xml")); // 4 = Not Valid XML
			list.add(getPath(prefix + "_not_well_formed.xml")); // 5 = Not well-formed XML
		} catch (IOException e) {
			fail(e.getMessage());
		}

		return list;
	}

	private List<URL> loadSchemaTest(String prefix) {
		List<URL> list = new ArrayList<>();
		try {
			list.add(getPath(prefix + ".xsd").toUri().toURL());
		} catch (IOException e) {
			fail(e.getMessage());
		}

		return list;
	}

	private Path getPath(String f) throws IOException {
		return new ClassPathResource(f).getFile().toPath();
	}

}