package ch.dlcm.test.tool;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;

import java.io.IOException;
import java.net.URI;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.Logger;
import org.junit.AssumptionViolatedException;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.Stopwatch;
import org.junit.runner.Description;
import org.junit.runner.RunWith;
import org.springframework.core.io.ClassPathResource;
import org.springframework.test.context.junit4.SpringJUnit4ClassRunner;

import ch.dlcm.tool.FileTool;
import ch.dlcm.tool.ZipTool;

@RunWith(SpringJUnit4ClassRunner.class)
public class ZipTest {
  private static final Logger log = Logger.getLogger(ZipTest.class);

  @Rule
  public final Stopwatch stopwatch = new Stopwatch() {
      protected void succeeded(long nanos, Description description) {
          log.info(description.getMethodName() + " succeeded, duration=" + displayDuration(nanos));
      }

      private String displayDuration(long nanos) {
        double seconds = (double)nanos / 1000000000.0;
        return seconds+"s";
      }
      protected void failed(long nanos, Throwable e, Description description) {
        log.info(description.getMethodName() + " failed, duration=" + displayDuration(nanos));
      }
      protected void skipped(long nanos, AssumptionViolatedException e,
              Description description) {
        log.info(description.getMethodName() + " skipped,, duration=" + displayDuration(nanos));
      }

      protected void finished(long nanos, Description description) {
        log.info(description.getMethodName() + " finished, duration=" + displayDuration(nanos));
      }

  };

  private final static String ZIP_DIR = "dlcm";
  private final static String ZIP_ARCHIVE = "test.zip";
  private final static int ZIP_FILE_NUMBER = 16;
  private final static int ZIP_XSD_NUMBER = 9;
  private final static String UNZIP_DIR = "unzip";

  @Test
  public void zipArchiveTest() throws IOException {
    URI zipDir = new ClassPathResource(ZIP_DIR).getURI();
    ZipTool zipTool = new ZipTool(ZIP_ARCHIVE);
    assertNotNull(zipDir);
    assertTrue(zipTool.zipFiles(zipDir.getRawPath()));
    assertTrue(zipTool.checkZip());
    assertEquals(ZIP_FILE_NUMBER, zipTool.getEntryNumber());
    assertTrue(FileTool.deleteFile(URI.create(ZIP_ARCHIVE)));
  }

  @Test
  public void unzipArchiveTest() throws IOException {
    URI zipDir = new ClassPathResource(ZIP_DIR).getURI();
    ZipTool zipTool = new ZipTool(ZIP_ARCHIVE);
    assertTrue(zipTool.zipFiles(zipDir.getRawPath()));
    assertTrue(zipTool.unzipFiles(UNZIP_DIR));
    List<Path> fileList = FileTool.readDirectory(UNZIP_DIR);
    fileList.addAll(FileTool.readDirectory(UNZIP_DIR+"/include"));
    assertEquals(ZIP_FILE_NUMBER, fileList.size());
    assertTrue(FileTool.deleteFile(URI.create(ZIP_ARCHIVE)));
    assertTrue(FileTool.deleteDirectory(Paths.get(UNZIP_DIR).toFile()));
  }

  @Test
  public void partialUnzipArchiveTest() throws IOException {
    URI zipDir = new ClassPathResource(ZIP_DIR).getURI();
    ZipTool zipTool = new ZipTool(ZIP_ARCHIVE);
    List<String> extractList = new ArrayList<>();
    extractList.add("dlcm.xsd");
    extractList.add("include/datacite-");
    assertTrue(zipTool.zipFiles(zipDir.getRawPath()));
    assertTrue(zipTool.unzipFiles(UNZIP_DIR, extractList));
    List<Path> fileList = FileTool.readDirectory(UNZIP_DIR);
    fileList.addAll(FileTool.readDirectory(UNZIP_DIR+"/include"));
    assertEquals(ZIP_XSD_NUMBER, fileList.size());
    assertTrue(FileTool.deleteFile(URI.create(ZIP_ARCHIVE)));
    assertTrue(FileTool.deleteDirectory(Paths.get(UNZIP_DIR).toFile()));
  }
  
  @Test
  public void unzipArchiveTest1() throws IOException{
    URI zipDir = new ClassPathResource(ZIP_DIR).getURI();
    ZipTool zipTool = new ZipTool(ZIP_ARCHIVE);
    assertTrue(zipTool.zipFiles(zipDir.getRawPath()));
    assertTrue(zipTool.unzipFiles(UNZIP_DIR));
    assertTrue(FileTool.deleteFile(URI.create(ZIP_ARCHIVE)));
    assertTrue(FileTool.deleteDirectory(Paths.get(UNZIP_DIR).toFile()));
  }

  @Test
  public void unzipArchiveTest2() throws IOException {
    URI zipDir = new ClassPathResource(ZIP_DIR).getURI();
    ZipTool zipTool = new ZipTool(ZIP_ARCHIVE);
    assertTrue(zipTool.zipFiles(zipDir.getRawPath()));
    assertTrue(zipTool.unzipFiles(UNZIP_DIR, null));
    assertTrue(FileTool.deleteFile(URI.create(ZIP_ARCHIVE)));
    assertTrue(FileTool.deleteDirectory(Paths.get(UNZIP_DIR).toFile()));
  }


}