package ch.dlcm.preparation.archivematica;

import java.util.Base64;
import java.util.List;

import javax.annotation.PostConstruct;

import org.apache.log4j.Logger;
import org.springframework.context.annotation.Profile;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpHeaders;
import org.springframework.http.MediaType;
import org.springframework.stereotype.Service;
import org.springframework.util.LinkedMultiValueMap;
import org.springframework.util.MultiValueMap;
import org.springframework.web.client.RestClientException;
import org.springframework.web.client.RestTemplate;

import ch.dlcm.model.Status.StatusValue;
import ch.dlcm.preparation.PreparationService;
import ch.dlcm.service.DLCMService;

@Profile("pre-archivematica")
@Service
public class ArchivematicaPreparationService extends DLCMService implements PreparationService {
    private static final Logger log = Logger.getLogger(ArchivematicaPreparationService.class);

	private String preparationPath;
	private String preparationId;
	private String baseUrl;
	private String userName;
	private String apiKey;

	RestTemplate restClt;
	
	@PostConstruct
	@Override
	protected void loadConfig(){
		this.preparationPath = this.dlcmConfig.getPreingestPreparationLocation();
		this.preparationId = this.dlcmConfig.getPreingestPreparationId();
		this.baseUrl = this.dlcmConfig.getPreingestPreparationUrl();
		this.userName = this.dlcmConfig.getPreingestPreparationUser();
		this.apiKey = this.dlcmConfig.getPreingestPreparationApiKey();
		restClt = new RestTemplate();
	}

	@Override
	public String submitPackage(String id, String name) {
		final String url = baseUrl + "/api/transfer/start_transfer/";
		final String transferName = this.getTransferName(id,name);
		
		// Create header
		MultiValueMap<String, String> headers = getRequestHeader(MediaType.APPLICATION_FORM_URLENCODED_VALUE);
		// Set Request
		MultiValueMap<String, String> params = new LinkedMultiValueMap<>();
		params.add("name", transferName);
		params.add("type", "standard");
		params.add("accession", "");
		params.add("row_ids[]", "");
		params.add("paths[]", Base64.getEncoder().encodeToString(getLocation(id).getBytes()));
		HttpEntity<MultiValueMap<String, String>> request = new HttpEntity<>(params, headers);
		// Send request
		TransferResult res = restClt.postForObject(url, request, TransferResult.class);
		if (res == null){
			return null;
		}
		log.info(res.getMessage() + "(" + res.getPath() + ")");
		String[] list = res.getPath().split("/");
		return this.approveTransfer(list[list.length-1]);
	}
	
	private String getTransferName(String id, String name) {
		String tranferName = "DCLM " + id ;
		return this.normalize(tranferName);
	}

	private String getLocation(String id) {
		return preparationId + ":" + preparationPath + "/" + id +"/";
	}

	private String normalize(String name) {
		return name.replace(" ", "_").replace(".", "_");
	}

	private String approveTransfer(String transferName) {
		final String url = baseUrl + "/api/transfer/approve/";

		// Create header
		MultiValueMap<String, String> headers = getRequestHeader(MediaType.APPLICATION_FORM_URLENCODED_VALUE);
		// Set Request
		MultiValueMap<String, String> params = new LinkedMultiValueMap<>();
		params.add("type", "standard");
		params.add("directory", transferName);
		HttpEntity<MultiValueMap<String, String>> request = new HttpEntity<>(params, headers);
		TransferApproval res = restClt.postForObject(url, request, TransferApproval.class);
		if (res == null){
			return null;
		}
		log.info(res.getMessage() + "(" + res.getUuid() + ")");
		return res.getUuid();
	}

	@Override
	public StatusValue getStatus(String id) {
		final String url = this.getUrlParameters(baseUrl + "/api/transfer/status/" + id + "/");
		TransferStatus t;

		try {
			t = restClt.getForObject(url, TransferStatus.class);
		} catch (Exception e) {
			log.error(e.getMessage(), e);
			return null;
		}
		if ("FAILED".equals(t.getStatus())){
			log.error(t.toString());
			return StatusValue.InError;
		}else if ("COMPLETE".equals(t.getStatus())){
			return StatusValue.Completed;
		}else if ("USER_INPUT".equals(t.getStatus())){
			return StatusValue.Paused;
		}
		return StatusValue.InProgress;
	}

	@Override
	public List<?> listPackages() {
		final String url = this.getUrlParameters(baseUrl + "/api/transfer/unapproved/");

		TransferList transferList;
		try {
			transferList = restClt.getForObject(url, TransferList.class);
		} catch (RestClientException e) {
			log.error(e.getMessage(), e);
			return null;
		}
		return transferList.getResults();
	}

	private String getUrlParameters(String baseURL) {
		return baseURL+"?username="+this.userName+"&api_key="+this.apiKey;
	}
	
	private MultiValueMap<String, String> getRequestHeader(String mimeType) {
		MultiValueMap<String, String> headers = new LinkedMultiValueMap<>();
		headers.add(HttpHeaders.AUTHORIZATION, "ApiKey " + userName + ":" + apiKey);
		if (mimeType != null){
			headers.add(HttpHeaders.CONTENT_TYPE, mimeType);
		}
		return headers;
	}

	@Override
	public String getPackageUri(String id) {
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public String sendPackage(String id, String name) {
		return this.sendSIP(id, name);
	}
}
