package ch.dlcm.preparation.bagit;

import java.io.IOException;
import java.io.InputStream;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.security.NoSuchAlgorithmException;
import java.time.OffsetDateTime;
import java.time.ZoneOffset;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.List;

import javax.annotation.PostConstruct;

import org.apache.log4j.Logger;
import org.springframework.context.annotation.Profile;
import org.springframework.core.io.ClassPathResource;
import org.springframework.stereotype.Service;

import ch.dlcm.model.Status.StatusValue;
import ch.dlcm.preparation.PreparationService;
import ch.dlcm.service.DLCMService;
import ch.dlcm.tool.FileTool;
import ch.dlcm.tool.StringTool;
import ch.dlcm.tool.ZipTool;
import gov.loc.repository.bagit.conformance.BagProfileChecker;
import gov.loc.repository.bagit.creator.BagCreator;
import gov.loc.repository.bagit.domain.Bag;
import gov.loc.repository.bagit.exceptions.CorruptChecksumException;
import gov.loc.repository.bagit.exceptions.FileNotInPayloadDirectoryException;
import gov.loc.repository.bagit.exceptions.InvalidBagitFileFormatException;
import gov.loc.repository.bagit.exceptions.MaliciousPathException;
import gov.loc.repository.bagit.exceptions.MissingBagitFileException;
import gov.loc.repository.bagit.exceptions.MissingPayloadDirectoryException;
import gov.loc.repository.bagit.exceptions.MissingPayloadManifestException;
import gov.loc.repository.bagit.exceptions.UnparsableVersionException;
import gov.loc.repository.bagit.exceptions.UnsupportedAlgorithmException;
import gov.loc.repository.bagit.exceptions.VerificationException;
import gov.loc.repository.bagit.exceptions.conformance.BagitVersionIsNotAcceptableException;
import gov.loc.repository.bagit.exceptions.conformance.FetchFileNotAllowedException;
import gov.loc.repository.bagit.exceptions.conformance.MetatdataValueIsNotAcceptableException;
import gov.loc.repository.bagit.exceptions.conformance.RequiredManifestNotPresentException;
import gov.loc.repository.bagit.exceptions.conformance.RequiredMetadataFieldNotPresentException;
import gov.loc.repository.bagit.exceptions.conformance.RequiredTagFileNotPresentException;
import gov.loc.repository.bagit.hash.StandardSupportedAlgorithms;
import gov.loc.repository.bagit.hash.SupportedAlgorithm;
import gov.loc.repository.bagit.reader.BagReader;
import gov.loc.repository.bagit.verify.BagVerifier;
import gov.loc.repository.bagit.writer.BagWriter;

@Profile("pre-bagit")
@Service
public class BagItPreparationService extends DLCMService implements PreparationService {
    private static final Logger log = Logger.getLogger(BagItPreparationService.class);
	private String preparationPath;
	
	public final String PROFILE = "BagIt-Profile-DLCM.json";

	@PostConstruct
	@Override
	protected void loadConfig(){	
		this.preparationPath = this.dlcmConfig.getPreingestPreparationLocation();
	}
	
	@Override
	public String submitPackage(String id, String name) {
		List<SupportedAlgorithm> algorithms = new ArrayList<>();
		algorithms.add(StandardSupportedAlgorithms.MD5);
		algorithms.add(StandardSupportedAlgorithms.SHA1);
		algorithms.add(StandardSupportedAlgorithms.SHA256);
		boolean includeHiddenFiles = false;
		Path packageLoc = Paths.get(preparationPath+"/"+id);
		try {
			// Create BagIt
			Bag bag = BagCreator.bagInPlace(packageLoc, algorithms, includeHiddenFiles);
			bag.getMetadata().add("Bagging-Date", StringTool.formatDate(OffsetDateTime.now(ZoneOffset.UTC),DateTimeFormatter.ISO_LOCAL_DATE));
			bag.getMetadata().add("Payload-Oxum", "");
			bag.getMetadata().add("Source-Organization", "University of Geneva");
			bag.getMetadata().add("Organization-Address", "Geneva");
			bag.getMetadata().add("DLCM-Id", id);
			bag.getMetadata().add("DLCM-Name", name);
			BagWriter.write(bag, packageLoc);
			String payloadOxum = bag.getMetadata().get("Payload-Oxum").get(0);
			bag.getMetadata().add("Bag-Size", StringTool.formatSmartSize(new Integer(payloadOxum.substring(0, payloadOxum.indexOf('.')))));
			BagWriter.write(bag, packageLoc);
			// Zip BagIt
			ZipTool zipTool = new ZipTool(packageLoc.toString()+".zip");
			if (!zipTool.zipFiles(packageLoc.toString())){
				log.error("Cannot zip BagIt");
				return null;
			}
		} catch (NoSuchAlgorithmException | IOException e) {
			log.error(e.getMessage(), e);
			return null;
		}
		return id;
	}

	private InputStream getBagItProfile() throws IOException {
		return new ClassPathResource(this.PROFILE).getInputStream();
	}
	@Override
	public StatusValue getStatus(String id) {
		BagVerifier bagVerifier = new BagVerifier();
		BagReader bagReader = new BagReader();
		boolean ignoreHiddenFiles = true;
		try {
			Bag bag = bagReader.read(Paths.get(preparationPath+"/"+id));
			if (bag == null) {
				log.error("Cannot find BagIt " + id);
				return StatusValue.InError;
			}
			bagVerifier.isComplete(bag, ignoreHiddenFiles);
			bagVerifier.isValid(bag, ignoreHiddenFiles);
			BagProfileChecker.bagConformsToProfile(this.getBagItProfile(), bag);
		} catch (IOException | MissingPayloadManifestException | MissingBagitFileException
				| MissingPayloadDirectoryException | FileNotInPayloadDirectoryException | InterruptedException
				| MaliciousPathException | UnsupportedAlgorithmException | InvalidBagitFileFormatException
				| CorruptChecksumException | VerificationException | FetchFileNotAllowedException | RequiredMetadataFieldNotPresentException 
				| MetatdataValueIsNotAcceptableException | RequiredManifestNotPresentException | BagitVersionIsNotAcceptableException
				| RequiredTagFileNotPresentException | UnparsableVersionException e) {
			log.error(e);
			return StatusValue.InError;
		}
		return StatusValue.Completed;
	}

	@Override
	public String getPackageUri(String id) {
		return FileTool.getFileUri(preparationPath + "/" + id + ".zip");
	}

	@Override
	public String sendPackage(String id, String name) {
		return this.sendSIP(id, name);
	}

	@Override
	public List<?> listPackages() {
		throw new UnsupportedOperationException();
	}

	
}
