package ch.dlcm.test.executor;

import ch.dlcm.exception.DLCMRuntimeException;
import jersey.repackaged.com.google.common.collect.Lists;
import org.hamcrest.Matcher;
import org.springframework.http.HttpStatus;
import org.springframework.http.MediaType;
import org.springframework.restdocs.hypermedia.HypermediaDocumentation;
import org.springframework.restdocs.hypermedia.LinkDescriptor;
import org.springframework.restdocs.mockmvc.MockMvcRestDocumentation;
import org.springframework.restdocs.mockmvc.RestDocumentationRequestBuilders;
import org.springframework.restdocs.payload.FieldDescriptor;
import org.springframework.restdocs.payload.PayloadDocumentation;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.test.web.servlet.ResultActions;
import org.springframework.test.web.servlet.request.MockHttpServletRequestBuilder;
import org.springframework.test.web.servlet.result.MockMvcResultHandlers;
import org.springframework.test.web.servlet.result.MockMvcResultMatchers;
import org.springframework.web.bind.annotation.RequestMethod;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import static org.springframework.restdocs.mockmvc.RestDocumentationRequestBuilders.get;
import static org.springframework.test.web.servlet.result.MockMvcResultHandlers.print;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.status;


/**
 * Class executor that gives the 4 cycles for make Spring Rest test and documentation. Spring has five documentation generator factory:
 * payload, http, header, request and hypermedia. This class covers the payload and hypermedia for instance.
 */
public class DLCMExecutorTestDoc {


    /**
     *  The GET http mock request. Usually the first cycle of the tests. By defaut, it auto-test the "content type" and the "http status code" expected by the client.
     */
    public static ResultActions doPerformRequestCycle(MockMvc mockMvc, String url, MediaType mediaType, HttpStatus status) throws Exception {
        return DLCMExecutorTestDoc.doPerformGetRequestCycle(mockMvc, url, mediaType, status);
    }

    public static ResultActions doPerformGetRequestCycle(MockMvc mockMvc, String url, MediaType mediaType, HttpStatus status) throws Exception {
        return new DLCMExecutorTestDoc().doPerformAnyMethodRequestCycle(mockMvc, url, mediaType, status, RequestMethod.GET, null);
    }

    /**
     *  The http mock request. Usually the first cycle of the tests. By defaut, it auto-test the "content type" and the "http status code" expected by the client.
     * @param mockMvc the mockmvp generated by the client in setupbefore
     * @param url the endpoint url
     * @param mediaType the content type spected for this request
     */
    public static ResultActions doPerformPostRequestCycle(MockMvc mockMvc, String url, MediaType mediaType, HttpStatus status, String content) throws Exception {
        return new DLCMExecutorTestDoc().doPerformAnyMethodRequestCycle(mockMvc, url, mediaType, status,  RequestMethod.POST, content);
    }

    private ResultActions doPerformAnyMethodRequestCycle(MockMvc mockMvc, String url, MediaType mediaType, HttpStatus status, RequestMethod method, String content) throws Exception {
        MockHttpServletRequestBuilder mockRequestBuilder = null;
        switch (method){
            case GET:
                mockRequestBuilder = RestDocumentationRequestBuilders.get(url);
                break;
            case POST:
                mockRequestBuilder = RestDocumentationRequestBuilders.post(url).content(content).contentType(mediaType);
                break;
            case PUT:
                mockRequestBuilder = RestDocumentationRequestBuilders.put(url).content(content).contentType(mediaType);
                break;
            default:
                break;
        }
        return performAndTestRequest(mockMvc, mockRequestBuilder, mediaType, status);
    }

    private ResultActions performAndTestRequest(MockMvc mockMvc, MockHttpServletRequestBuilder mockHttpServletRequestBuilder,MediaType mediaType, HttpStatus status) throws Exception {
        return mockMvc.perform(mockHttpServletRequestBuilder)
                .andDo(MockMvcResultHandlers.print())
                .andExpect(MockMvcResultMatchers.content().contentType(mediaType))
                .andExpect(MockMvcResultMatchers.status().is(status.value()));
    }

    /**
     *  The rest integration test cycle. Usually the second cycle of the tests
     * @param resultActions The request execution. Usually started in the 1st cycle
     * @param expectedMap a map containing the pair [xpath=Matcher]
     */
    @SuppressWarnings("unchecked")
    public static ResultActions doTestsExpectedCycle(ResultActions resultActions, Map<String, Matcher> expectedMap){
        expectedMap.forEach( (key, expected) -> {
            try {
                resultActions.andExpect(MockMvcResultMatchers.jsonPath(key, expected));
            } catch (Throwable e) {
                throw new DLCMRuntimeException("Error in doTestsExpectedCycle. [URI]=" + resultActions.andReturn().getRequest().getRequestURI(), e);
            }
        });
        return resultActions;
    }

    /**
     *  Creation of the fields snippets. Usually the 1st cycle of the rest doc's generation
     * @param resultActions The request execution. Usually started in the 1st cycle
     * @param restName the endpoint name
     * @param fieldsMap a map containing the pair [xpath=fieldDescriptor]
     */
    public static ResultActions doDocumentationFieldsCycle(ResultActions resultActions, String restName, Map<String, String> fieldsMap ) throws Exception {
        final List<FieldDescriptor> fieldDescriptors = new ArrayList<>();
        fieldsMap.forEach( (key, field) ->
            fieldDescriptors.add(PayloadDocumentation.fieldWithPath(key).description(field))
        );
        resultActions.andDo(MockMvcRestDocumentation.document(restName, PayloadDocumentation.responseFields(Lists.reverse(fieldDescriptors))));
        return resultActions;
    }

    /**
     * @param resultActions The request execution. Usually started in the 1st cycle
     * @param restName the endpoint name
     * @param linksMap a map containing the pair [xpath=linkDescriptor]
     */
    public static ResultActions doDocumentationLinksCycle(ResultActions resultActions, String restName, Map<String, String> linksMap ) throws Exception {
        final List<LinkDescriptor> linkDescriptors = new ArrayList<>();
        linksMap.forEach((key, descriptor) ->
                linkDescriptors.add(HypermediaDocumentation.linkWithRel(key).description(descriptor))
        );
        resultActions.andDo(MockMvcRestDocumentation.document(restName, HypermediaDocumentation.links(HypermediaDocumentation.halLinks(), linkDescriptors)));
        return resultActions;
    }

    /** Test the 4xx error code for any invalid url.
     * PS: Can't doc errors, Spring rest-docs does not support error codes.
     * https://stackoverflow.com/questions/36559367/how-to-document-http-error-codes-using-spring-restdocs
     */
    public static ResultActions doErrorUrlRestLink(MockMvc mockMvc, String url) throws Exception {
        return mockMvc.perform(get(url))
                .andDo(print())
                .andExpect(status().is4xxClientError());
    }


}
