package ch.dlcm.tools;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.List;
import java.util.function.Predicate;
import java.util.stream.Collectors;

import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.core.ParameterizedTypeReference;
import org.springframework.core.io.FileSystemResource;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpHeaders;
import org.springframework.http.HttpMethod;
import org.springframework.http.MediaType;
import org.springframework.http.ResponseEntity;
import org.springframework.util.LinkedMultiValueMap;
import org.springframework.web.client.RestTemplate;

import ch.dlcm.model.DataFileInterface;
import ch.dlcm.model.Deposit;
import ch.dlcm.model.oais.AipDataFile;
import ch.dlcm.model.oais.ArchivalInfoPackage;
import ch.dlcm.model.oais.SipDataFile;
import ch.dlcm.model.oais.SubmissionInfoPackage;
import ch.dlcm.rest.RESTName;
import ch.dlcm.tool.RestClientTool;

public abstract class Tools {
	private static final Logger log = Logger.getLogger(Tools.class);

	@Value("${dlcm.module.preingest}")
	protected String preIngestUrl;
	@Value("${dlcm.module.ingest}")
	protected String ingestUrl;
	@Value("${dlcm.module.archival-storage}")
	protected String archivalStorageUrl;

	protected List<Path> scanFolder(String root, Predicate<? super Path> t) {
		try {
			return Files.list(Paths.get(root)).filter(t).collect(Collectors.toList());
		} catch (IOException e) {
			log.error(e);
		}
		return new ArrayList<>();
	}

	protected List<Path> scanAll(String root) {
		List<Path> list = new ArrayList<>();
		for (Path d : scanFolder(root, Files::isDirectory)) {
			list.addAll(scanAll(d.toString()));
		}
		list.addAll(scanFolder(root, Files::isRegularFile));
		return list;
	}
	
	
	protected List<Path> scanByExtension(String root, String extension) {
		List<Path> list = new ArrayList<>();
		list.addAll(scanFolder(root, path -> path.toString().endsWith("."+extension)));
		return list;
	}

	protected String createDeposit(String name) {
		try {
		  RestTemplate restClt = RestClientTool.getClient(this.preIngestUrl);
	    Deposit deposit = new Deposit();
	    deposit.setTitle(name);
			// Send Deposit object
			Deposit resDeposit = restClt.postForObject(this.preIngestUrl + "/" + RESTName.DEPOSIT, deposit,
					Deposit.class);
			if (resDeposit.getDlcmId() != null) {
				log.info("Deposit created: ID=" + resDeposit.getDlcmId());
				return resDeposit.getDlcmId();
			} else {
				log.error("Cannot create deposit (name=" + name + ")");
			}
		} catch (Exception e) {
			log.error("Cannot send deposit (name=" + name + "): " + e.getMessage(), e);
		}
		return null;
	}

	protected <T extends DataFileInterface> String addDataFile(String url, Path df, String dataCat, String dataType,
			String relLocation, ParameterizedTypeReference<T[]> responseType) {
		LinkedMultiValueMap<String, Object> map = new LinkedMultiValueMap<>();
		map.add("file", new FileSystemResource(df.toFile()));
		if (dataCat != null && !dataCat.isEmpty()){
			map.add("category", dataCat);
		}
		if (dataType != null && !dataType.isEmpty()){
			map.add("type", dataType);
		}
		if (relLocation != null && !relLocation.isEmpty()) {
			map.add("folder", relLocation);
		}
		HttpHeaders headers = new HttpHeaders();
		headers.setContentType(MediaType.MULTIPART_FORM_DATA);
		HttpEntity<LinkedMultiValueMap<String, Object>> requestEntity = new HttpEntity<>(map, headers);

		T[] list;
		try {
		  RestTemplate restClt = RestClientTool.getClient(url);
			ResponseEntity<T[]> responseEntity = restClt.exchange(url, HttpMethod.POST, requestEntity, responseType);
			list = responseEntity.getBody();
		} catch (Exception e) {
			log.error(e);
			return null;
		}
		if (list.length != 1) {
			return null;
		}
		return list[0].getDlcmId();
	}
	
	public String createSIP(Path df) {
		try {
		  RestTemplate restClt = RestClientTool.getClient(this.ingestUrl);
	    // Create SIP
	    SubmissionInfoPackage sip = new SubmissionInfoPackage();
	    sip.getInfo().setName(df.getFileName().toString());
	    sip.getInfo().setDescription("Created by DLCM Tools");
			// Send SIP object
			SubmissionInfoPackage resSip = restClt.postForObject(this.ingestUrl + "/" + RESTName.SIP, sip,
					SubmissionInfoPackage.class);
			if (resSip.getDlcmId() != null) {
				log.info("SIP created: ID=" + resSip.getDlcmId());
				if (this.addDataFile(this.ingestUrl + "/" + RESTName.SIP + "/" + resSip.getDlcmId() + "/" + RESTName.UL,
						df, null, null, null, new ParameterizedTypeReference<SipDataFile[]>() {}) != null) {
					return resSip.getDlcmId();
				} else {
					log.error("Cannot add package for SIP " + resSip.getDlcmId());
				}
			} else {
				log.error("Cannot create SIP for " + df.toString());
			}
		} catch (Exception e) {
			log.error("Cannot send AIP (" + df.toString() + "): " + e.getMessage(), e);
		}
		return null;
	}
	
	public String createAIP(Path df) {
		try {
		  RestTemplate restClt = RestClientTool.getClient(this.archivalStorageUrl);
	    // Create AIP
	    ArchivalInfoPackage aip = new ArchivalInfoPackage();
	    aip.getInfo().setName(df.getFileName().toString());
	    aip.getInfo().setDescription("Created by DLCM Tools");
			// Send AIP object
			ArchivalInfoPackage resAip = restClt.postForObject(this.archivalStorageUrl + "/" + RESTName.AIP, aip,
					ArchivalInfoPackage.class);
			if (resAip.getDlcmId() != null) {
				log.info("AIP created: ID=" + resAip.getDlcmId());
				if (this.addDataFile(this.archivalStorageUrl + "/" + RESTName.AIP + "/" + resAip.getDlcmId() + "/" + RESTName.UL,
						df, null, null, null, new ParameterizedTypeReference<AipDataFile[]>() {}) != null) {
					return resAip.getDlcmId();
				} else {
					log.error("Cannot add package for AIP " + resAip.getDlcmId());
				}
			} else {
				log.error("Cannot create AIP for " + df.toString());
			}
		} catch (Exception e) {
			log.error("Cannot send AIP (" + df.toString() + "): " + e.getMessage(), e);
		}
		return null;
	}

}
